/***************************************************************************
  iPIC3D.cpp  -  Main file for 3D simulation
  -------------------
 ************************************************************************** */


// Topology
#include "processtopology/VirtualTopology3D.h"
#include "processtopology/VCtopology3D.h"
// input
#include "inputoutput/CollectiveIO.h"
#include "inputoutput/Collective.h"
// this is tow write VTK files
#include "inputoutput/SerialIO.h"
// grid
#include "grids/Grid.h"
#include "grids/Grid3DCU.h"
// fields
#include "fields/Field.h"
#include "fields/EMfields3D.h"


#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

using namespace std;
using std::cerr;
using std::endl;
using std::ofstream;


int main(int argc, char **argv) {
  // read input file
  Collective *col = new Collective(argc, argv); // Every proc loads the parameters of simulation from class Collective
  bool verbose = col->getVerbose();
  int restart_cycle = col->getRestartOutputCycle();
  string SaveDirName = col->getSaveDirName();
  const int restart = col->getRestart_status();
  const int first_cycle = col->getLast_cycle() + 1; // get the last cycle from the restart
  // initialize topology: here serial
  VCtopology3D *vct = new VCtopology3D();
  vct->setup_vctopology();
  // Print the initial settings to stdout and a file
  vct->Print();
  col->Print();
  col->save();
  
  // Create the local grid
  Grid3DCU *grid = new Grid3DCU(col, vct);  // Create the local grid
  EMfields3D *EMf = new EMfields3D(col, grid);  // Create Electromagnetic Fields Object
  EMf->initGEMdoublePeriodicMHD(vct,grid);
  // EMf->initRhoTest2(vct,grid);
  // energies file
  double Benergy = 0.0, TOTenergy = 0.0, Ke = 0.0, CurrTime;
  string cq = SaveDirName + "/ConservedQuantities.txt";
  ofstream my_file(cq.c_str());
  my_file.close();
  


  // *******************************************//
  // **** Start the Simulation! ***//
  // *******************************************//
  for (int cycle = first_cycle; cycle < (col->getNcycles() + first_cycle); cycle++) {
      //  output
      cout << endl;
      cout << "**************************************" << endl;
      // MHD SOLVER
      EMf->calculateFieldMHD(grid,vct,cycle);
      cout << "Total Energy: " << EMf->getTotenergyMHD() << endl;
      cout << "Magnetic Field Energy: " << EMf->getBenergyMHD() << endl;
      cout << "Kinetic Energy: " << EMf->getKeMHD() << endl;
      // write the conserved quantities (MHD VERSION)
      if (cycle % col->getDiagnosticsOutputCycle() == 0) {
          CurrTime = EMf->getMHDtime();
          Benergy = EMf->getBenergyMHD();
          TOTenergy = EMf->getTotenergyMHD(); // in iPic3D, this is Ke + E + B energies
          Ke = EMf->getKeMHD();  // in iPic3D, this is TOTEnergy
          ofstream my_file(cq.c_str(), fstream::app);
          my_file << cycle << "\t" << CurrTime <<"\t" << TOTenergy << "\t" << Benergy << "\t" << Ke << endl;
          my_file.close();
      }
      // write VTK file
      if (cycle % col->getFieldOutputCycle() == 0){
          VTK_Write_Vectors(cycle,grid,EMf);
          VTK_Write_Scalars(cycle,grid,EMf);
          VTK_Write_Ohms(cycle,grid,EMf,col);

      }
  }
  return (0);

}
