#ifndef SerialIO_H
#define SerialIO_H



#include <iostream>
#include <sstream>
#include <string>
#include <fstream>

#include "../grids/Grid.h"
#include "../fields/Field.h"
//#include "../utility/Alloc.h"

using std::string;
using std::stringstream;
using std::ofstream;
using std::cout;
using std::endl;


inline void VTK_Write_Scalars(int cycle, Grid *grid,Field* field){
    int nxc = grid->getNXC();
    int nyc = grid->getNYC();
    int nzc = grid->getNZC();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    double ***scalar= newArr3(double,nxc,nyc,nzc);
    scalar = field->getRHO0();
    
    // density
    string filename = "rho";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./results/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    
    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "Density from MHD" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file << "SCALARS rho float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            if (fabs(scalar[i][j][2])< 1E-8)
                scalar[i][j][2] = 0;
            my_file << scalar[i][j][2] <<  endl;
	    }
    }
    my_file.close();
    
    // U energy
    scalar = field->getUC();
    filename = "U";
    temp = "./results/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Energy from MHD" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file2 << "SCALARS U float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            if (fabs(scalar[i][j][2])< 1E-8)
                scalar[i][j][2] = 0;
            my_file2 << scalar[i][j][2] <<  endl;
	    }
    }
    my_file2.close();
    
    // grid resistivity
    scalar = field->getgrid_eta();
    filename = "eta_resistivity";
    temp = "./results/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Resistivity on the grid" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file1 << "SCALARS eta float" << endl;
    my_file1 << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            if (fabs(scalar[i][j][2])< 1E-16)
                scalar[i][j][2] = 0;
            my_file1 << scalar[i][j][2] <<  endl;
	    }
    }
    my_file1.close();
    
    
}


inline void VTK_Write_Vectors(int cycle, Grid *grid,Field* field){
    // stream file to be opened
    string filename;
    string temp;
    stringstream cc;
    cc << cycle;
    
    
    int nxc = grid->getNXC();
    int nyc = grid->getNYC();
    int nzc = grid->getNZC();
    
    double dx = grid->getDX();
    double dy = grid->getDY();
    
    double ***vector_x= newArr3(double,nxc,nyc,nzc);
    double ***vector_y= newArr3(double,nxc,nyc,nzc);
    double ***vector_z= newArr3(double,nxc,nyc,nzc);
   
    vector_x = field->getBxC();
    vector_y = field->getByC();
    vector_z = field->getBzC();
   
    
    filename = "B";
    temp = "./results/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "B from MHD simulation" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file2 << "VECTORS B float" << endl;
    
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            if (fabs(vector_x[i][j][2])< 1E-8)
                vector_x[i][j][2] = 0;
            if (fabs(vector_y[i][j][2])< 1E-8)
                vector_y[i][j][2] = 0;
            if (fabs(vector_z[i][j][2])< 1E-8)
                vector_z[i][j][2] = 0;
            my_file2 << vector_x[i][j][2] << " " << vector_y[i][j][2] <<  " " << vector_z[i][j][2] <<  endl;
	    }
    }
    my_file2.close();
    
//    // write momentum: rho*u
    vector_x = field->getpxC();
    vector_y = field->getpyC();
    vector_z = field->getpzC();
    
    
    filename = "p";
    temp = "./results/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "Momentum from MHD simulation" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file1 << "VECTORS p-momentum float" << endl;
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            if (fabs(vector_x[i][j][2])< 1E-8)
                vector_x[i][j][2] = 0;
            if (fabs(vector_y[i][j][2])< 1E-8)
                vector_y[i][j][2] = 0;
            if (fabs(vector_z[i][j][2])< 1E-8)
                vector_z[i][j][2] = 0;
            my_file1 << vector_x[i][j][2] << " " << vector_y[i][j][2] <<  " " << vector_z[i][j][2] <<  endl;
        }
    }
    
    my_file1.close();
    // deallocate
    //delArr3(vector_x,nxc,nyc);
    //delArr3(vector_y,nxc,nyc);
    //delArr3(vector_z,nxc,nyc);
    
}



inline void VTK_Write_Ohms(int cycle,Grid *grid,Field* field,Collective *col){

    string filename = "VeXB";
    string temp;
    stringstream cc;
    cc << cycle;
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    
    int nxc = grid->getNXC();
    int nyc = grid->getNYC();
    int nzc = grid->getNZC();
    
    double dx = grid->getDX();
    double dy = grid->getDY();

    double temp_variable1, temp_variable2, temp_variable3, temp_variable4;

    // temp_variable3 is a simple plug for time=0 case, when moments are not collected yet. Passed as an argument from iPIC3D
    // if(cycle==0)
    //    temp_variable3=0.0;
    // else
    //   temp_variable3=1.0;
    
    // Va = col->getB0x(); // Works for GEM-like reconnection only. If need general Parsek units, just put Va=1
    
    
    ofstream my_file(temp.c_str());
    my_file << "# vtk DataFile Version 1.0" << endl;
    my_file << "E convective z from MHD" << endl;
    my_file << "ASCII" << endl;
    my_file << "DATASET STRUCTURED_POINTS"  << endl;
    my_file << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file << "SCALARS Econv float" << endl;
    my_file << "LOOKUP_TABLE default" << endl;
    
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
        
            if (fabs((field->getRHO0(i,j,2)))< 1E-10)
                temp_variable1 = 0.0;
            else
                temp_variable1= - ((field->getpxC(i,j,2))*(field->getByC(i,j,2))-(field->getpyC(i,j,2))*(field->getBxC(i,j,2)))/(field->getRHO0(i,j,2));
                my_file << temp_variable1 <<  endl;
	    }
    }
    
    my_file.close();


    filename = "etaJ";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file1(temp.c_str());
    my_file1 << "# vtk DataFile Version 1.0" << endl;
    my_file1 << "etaJ from MHD" << endl;
    my_file1 << "ASCII" << endl;
    my_file1 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file1 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file1 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file1 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file1 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file1 << "SCALARS etaJ float" << endl;
    my_file1 << "LOOKUP_TABLE default" << endl;
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
        
            temp_variable1=((((field->getByC(i+1,j,2)) - (field->getByC(i-1,j,2)))/(2*dx))-(((field->getBxC(i,j+1,2)) - (field->getBxC(i,j-1,2)))/(2*dx)))*(field->getgrid_eta(i,j,2));
            my_file1 <<  temp_variable1 << endl;
            
	    }
    }
    my_file1.close();
    
    
     
    filename = "Esum";
    temp = "./data_Ohm/" + filename + "_"+ cc.str() ;
    temp += ".vtk";
    cout << "Opening file: " << temp << endl;
    ofstream my_file2(temp.c_str());
    my_file2 << "# vtk DataFile Version 1.0" << endl;
    my_file2 << "Esum from MHD" << endl;
    my_file2 << "ASCII" << endl;
    my_file2 << "DATASET STRUCTURED_POINTS"  << endl;
    my_file2 << "DIMENSIONS " << (nxc-4) << " " << (nyc-4) << " " << 1 << endl;
    my_file2 << "ORIGIN " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
    my_file2 << "SPACING " << dx << " " << dy << " " << 1.0 << endl;
    my_file2 << "POINT_DATA " << (nxc-4)*(nyc-4) << endl;
    my_file2 << "SCALARS Esum float" << endl;
    my_file2 << "LOOKUP_TABLE default" << endl;
    
    for (int j=2; j < nyc-2; j++){
        for (int i=2; i < nxc-2; i++){
            
            if (fabs((field->getRHO0(i,j,2)))< 1E-10)
                temp_variable1 = 0.0;
            else
                temp_variable1= - ((field->getpxC(i,j,2))*(field->getByC(i,j,2))-(field->getpyC(i,j,2))*(field->getBxC(i,j,2)))/(field->getRHO0(i,j,2));
            
            temp_variable1+=((((field->getByC(i+1,j,2)) - (field->getByC(i-1,j,2)))/(2*dx))-(((field->getBxC(i,j+1,2)) - (field->getBxC(i,j-1,2)))/(2*dx)))*(field->getgrid_eta(i,j,2));
            
            my_file2 <<  temp_variable1 << endl;
            
	    }
    }
    my_file2.close();



}



#endif
