#ifndef VCtopology3D_H
#define VCtopology3D_H

#include "VirtualTopology3D.h"
#include "../utility/Alloc.h"
#include <iostream>



using std::cout;
using std::endl;



class VCtopology3D:public VirtualTopology3D {
public:
  /** constructor: Define topology parameters: dimension, domain decomposition,... */
  VCtopology3D();
  /** destructor */
  ~VCtopology3D();
  /** Find the neighbors in the new communicator  */
  void setup_vctopology();
  /** Print topology info */
  void Print();
  /** Print the mapping of topology */
  void PrintMapping();
  /** get XLEN */
  int getXLEN();
  /** get YLEN */
  int getYLEN();
  /** get ZLEN */
  int getZLEN();
  /** get nprocs */
  int getNprocs();
  /** get periodicity on boundaries - DIRECTION X*/
  bool getPERIODICX();
  /** get periodicity on boundaries - DIRECTION Y*/
  bool getPERIODICY();
  /** get periodicity on boundaries - DIRECTION Z*/
  bool getPERIODICZ();
  /** get the cartesian rank of the process */
  int getCartesian_rank();
  /** get the cartesian rank of XLEFT neighbor */
  int getXleft_neighbor();
  /** get the cartesian rank of XRIGHT neighbor */
  int getXright_neighbor();
  /** get the cartesian rank of YLEFT neighbor */
  int getYleft_neighbor();
  /** get the cartesian rank of YRIGHT neighbor */
  int getYright_neighbor();
  /** get the cartesian rank of ZLEFT neighbor */
  int getZleft_neighbor();
  /** get the cartesian rank of ZRIGHT neighbor */
  int getZright_neighbor();
  /** get the coordinates in dir direction of process*/
  int getCoordinates(int dir);
  /** get Periodicity condition in dir direction */
  int getPeriods(int dir);
  /** if cVERBOSE == true, print to the screen all the comunication */
  bool getcVERBOSE();

private:
  /** number of subdomains - Direction X */
  int XLEN;
  /** number of subdomains - Direction Y */
  int YLEN;
  /** number of subdomains - Direction Z */
  int ZLEN;
  /** nprocs = number of processors */
  int nprocs;
  /** periodicity on boundaries - DIRECTION X*/
  bool PERIODICX;
  /** periodicity on boundaries - DIRECTION Y*/
  bool PERIODICY;
  /** periodicity on boundaries - DIRECTION Z*/
  bool PERIODICZ;
  /** cartesian rank */
  int cartesian_rank;
  /** cartesian rank of XLEFT neighbor */
  int xleft_neighbor;
  /** cartesian rank of XRIGHT neighbor */
  int xright_neighbor;
  /** cartesian rank of YLEFT neighbor */
  int yleft_neighbor;
  /** cartesian rank of YRIGHT neighbor */
  int yright_neighbor;
  /** cartesian rank of ZRIGHT neighbor */
  int zleft_neighbor;
  /** cartesian rank of ZLEFT neighbor */
  int zright_neighbor;

  /** if cVERBOSE == true, print to the screen all the comunication */
  bool cVERBOSE;
};

/** DEFINE THE Topology HERE, setting XLEN,YLEN,ZLEN */
inline VCtopology3D::VCtopology3D() {
  // *******************************************
  // *******************************************
  // change these values to change the topology
  XLEN = 1;
  YLEN = 1;
  ZLEN = 1;
  nprocs = 1;
  // here you have to set the topology for the fields
  PERIODICX = false;
  PERIODICY = true;
  PERIODICZ = true;
  // *******************************************
  // *******************************************
  cVERBOSE = false;             // communication verbose ?

}





/** Within CART_COMM, processes find about their new rank numbers, their cartesian coordinates,
  and their neighbors  */
inline void VCtopology3D::setup_vctopology() {
    xleft_neighbor  = 0;
    xright_neighbor = 0;
    yleft_neighbor  = 0;
    yright_neighbor = 0;
    zleft_neighbor  = 0;
    zright_neighbor = 0;
}
/** destructor */
inline VCtopology3D::~VCtopology3D() {

}
/** print topology info */
inline void VCtopology3D::Print() {
  
}
/** print cartesian rank of neighbors and coordinate of process */
inline void VCtopology3D::PrintMapping() {
  
}
/** get XLEN */
inline int VCtopology3D::getXLEN() {
  return (1);
}
/** get YLEN */
inline int VCtopology3D::getYLEN() {
  return (1);
}
/** get ZLEN */
inline int VCtopology3D::getZLEN() {
  return (1);
}
/** get nprocs */
inline int VCtopology3D::getNprocs() {
  return (1);
}
/** get periodicity on boundaries - DIRECTION X*/
inline bool VCtopology3D::getPERIODICX() {
  return (PERIODICX);
}
/** get periodicity on boundaries - DIRECTION Y*/
inline bool VCtopology3D::getPERIODICY() {
  return (PERIODICY);
}
/** get periodicity on boundaries - DIRECTION Z*/
inline bool VCtopology3D::getPERIODICZ() {
  return (PERIODICZ);
}
/** get the cartesian rank of the process */
inline int VCtopology3D::getCartesian_rank() {
  return (0);
}
/** get the cartesian rank of XLEFT neighbor */
inline int VCtopology3D::getXleft_neighbor() {
  return (xleft_neighbor);
}
/** get the cartesian rank of XRIGHT neighbor */
inline int VCtopology3D::getXright_neighbor() {
  return (xright_neighbor);
}
/** get the cartesian rank of YLEFT neighbor */
inline int VCtopology3D::getYleft_neighbor() {
  return (yleft_neighbor);
}
/** get the cartesian rank of YRIGHT neighbor */
inline int VCtopology3D::getYright_neighbor() {
  return (yright_neighbor);
}
/** get the cartesian rank of ZLEFT neighbor */
inline int VCtopology3D::getZleft_neighbor() {
  return (zleft_neighbor);
}
/** get the cartesian rank of ZRIGHT neighbor */
inline int VCtopology3D::getZright_neighbor() {
  return (zright_neighbor);
}
/** if cVERBOSE == true, print to the screen all the comunication */
inline bool VCtopology3D::getcVERBOSE() {
  return (cVERBOSE);
}
/** get the coordinates in dir direction of process*/
inline int VCtopology3D::getCoordinates(int dir) {
  return (0);
}
/** get Periodicity condition in dir direction */
inline int VCtopology3D::getPeriods(int dir) {
  return (0);
}

#endif
