#ifndef EMfields3D_H
#define EMfields3D_H


#include <iostream>

#include <math.h>



#include "../utility/Alloc.h"
#include "../mathlib/Basic.h"
#include "../utility/TransArraySpace3D.h"
#include "../solvers/CG.h"
#include "../solvers/GMRES.h"


using std::cout;
using std::cerr;
using std::endl;



#define cut_MHD(W,W0) (((W)<(W0))?(W0):(W))
#define ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm) (((U)-((0.5)*((px)*(px) + (py)*(py) + (pz)*(pz))/(ro))-(((Bx)*(Bx) +(By)*(By) +(Bz)*(Bz))*(((double)(1.0))/(mu/((double)(4.0)))/((double)(8.0)))))*(gamm-((double)(1.0))))
#define min3_MHD(a,b,c) (min(min((a),(b)),(c)))
#define mod_MHD(a) (((a)<0.0?(-(a)):(a)))
#define CN 0.2
#define sgn_MHD(a) (((a)<0.0?(-1.0):(1.0)))

#define minmod_MHD(w1,w2) (sgn_MHD((w1))*max(0.0,min(mod_MHD((w1)),(w2)*sgn_MHD((w1)))))
#define minmod3_MHD(w1,w2,w3) (sgn_MHD((w1))*max(0.0,min3_MHD(mod_MHD((w1)),(w2)*sgn_MHD((w1)),(w3)*sgn_MHD((w1)))))
#define avr_MHD(W1,W2) (0.5*((W1)+(W2)))


#define TVDPROP 2

#if TVDPROP==3
#define F_dWq(dW1,dW2) (minmod_MHD((dW1),(dW2)))
#elif TVDPROP==2
#define F_dWq(dW1,dW2) (minmod3_MHD( (2.0)*(dW1),(dW2)*(2.0),(0.5)*((dW1)+(dW2))))
#elif TVDPROP==1
#define F_dWq(dW1,dW2) (sgn_MHD(dW2)*max3(0.0,min(2.0*(mod_MHD(dW2)),sgn_MHD(dW2)*(dW1)),min(mod_MHD(dW2),2.0*sgn_MHD(dW2)*(dW1))))
//#define F_dWq(W1,W2,W3) dont tested, sorry
#endif



#define BEGIN_x_small  x=1; while (x<nxc-1) {
#define END_x_small  x++; }

#define BEGIN_y_small  y=1; while (y<nyc-1) {
#define END_y_small  y++; }

#define BEGIN_z_small  z=1; while (z<nzc-1) {
#define END_z_small  z++; }

#define BEGIN_small  BEGIN_x_small BEGIN_y_small BEGIN_z_small
#define END_small  END_z_small END_y_small END_x_small

#define BEGIN_x_small2  x=2; while (x<nxc-2) {

#define BEGIN_y_small2  y=2; while (y<nyc-2) {

#define BEGIN_z_small2  z=2; while (z<nzc-2) {


#define BEGIN_x_full  x=0; while (x<nxc) {
#define END_x_full  x++; }

#define BEGIN_y_full  y=0; while (y<nyc) {
#define END_y_full  y++; }

#define BEGIN_z_full  z=0; while (z<nzc) {
#define END_z_full  z++; }

#define BEGIN_full BEGIN_x_full  BEGIN_y_full  BEGIN_z_full
#define END_full  END_z_full  END_y_full  END_x_full

#define ASSIGN_DELTA   dxplus=dx;dxminus=dx;dyplus=dy;dyminus=dy;dzplus=dz;dzminus=dz;
#define ASSIGN_DELTA_X dxplus=dx;dxminus=dx;
#define ASSIGN_DELTA_Y dyplus=dy;dyminus=dy;

#define ASSIGN_DELTA_Z dzplus=dz;dzminus=dz;
//#define ALPHA_FRICTION 0.1
#define ALPHA_FRICTION ((MHDtime)<20.0?((10.0*cos(3.14*MHDtime/40.0))+0.1):0.1)

#define nrotxB ((nFDy(mBz,dyplus,dyminus)) - (nFDz(mBy,dzplus,dzminus)))
#define nrotyB ((nFDz(mBx,dzplus,dzminus)) - (nFDx(mBz,dxplus,dxminus)))
#define nrotzB ((nFDx(mBy,dxplus,dxminus)) - (nFDy(mBx,dyplus,dyminus)))



#define nFDx(A,dplus,dminus)  (((((A[x+1][y][z]) - (A[x][y][z]))*((dminus)/(dplus)))+(((A[x][y][z]) - (A[x-1][y][z]))*((dplus)/(dminus))))/((dminus)+(dplus)))
#define nFDy(A,dplus,dminus)  (((((A[x][y+1][z]) - (A[x][y][z]))*((dminus)/(dplus)))+(((A[x][y][z]) - (A[x][y-1][z]))*((dplus)/(dminus))))/((dminus)+(dplus)))
#define nFDz(A,dplus,dminus)  (((((A[x][y][z+1]) - (A[x][y][z]))*((dminus)/(dplus)))+(((A[x][y][z]) - (A[x][y][z-1]))*((dplus)/(dminus))))/((dminus)+(dplus)))

#define nFD(AR,AC,AL,dplus,dminus) (((((AR) - (AC))*((dminus)/(dplus)))+(((AC) - (AL))*((dplus)/(dminus))))/((dminus)+(dplus)))


#define SET_DELTA double dxplus,dxminus,dyplus,dyminus,dzplus,dzminus;


#define F_W1(W,dt,dq,G1,G2,Q) ((W)-(0.5)*(dt)*((G2)-(G1))*(dq)+(1/6.0)*(dt)*(Q))
// F_W2 with the artificial viscosity.
#define F_W2(W,dt,dq,GL1,GR1,GL2,GR2,WL1,WR1,WL2,WR2,cq1,cq2,Q) ((W)-0.5*(dt)*((GL2)+(GR2)-(GL1)-(GR1)-(cq2)*((WR2)-(WL2))+(cq1)*((WR1)-(WL1)))*(dq)+(1/3.0)*(dt)*(Q))

//Bxc2[x][y][z]=F_W2(Bxc0[x][y][z],dt,invdx,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,Q__Bxc[x][y][z]);


#define Gxm_U(ro,px,py,pz,gamm,Bx,By,Bz,U) (((( (U)+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm)) )*( (px)/ (ro)  ))+((0.5/(mu))*(  ((Bx)*(Bx) +(By)*(By) +(Bz)*(Bz))*( (px)/ (ro)  ) )) - ((1.0/(mu))*( (Bx)*( (Bx)*(px) +(By)*(py) +(Bz)*(pz))/(ro) )  )))
#define Gmx_Bx(ro,px,py,pz,gamm,Bx,By,Bz,U) (0.0)
#define Gmx_By(ro,px,py,pz,gamm,Bx,By,Bz,U) ( ((By)*(px)-(Bx)*(py))/(ro))
#define Gmx_Bz(ro,px,py,pz,gamm,Bx,By,Bz,U) ( ((Bz)*(px)-(Bx)*(pz))/(ro))
#define Gmx_px(ro,px,py,pz,gamm,Bx,By,Bz,U) (((px)*(px)/(ro))+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm) )+ (((By)*(By)+(Bz)*(Bz)-(Bx)*(Bx))*(1.0/8.0/(mu/4.0))))
#define Gmx_py(ro,px,py,pz,gamm,Bx,By,Bz,U) (((px)*(py)/(ro)) - ((Bx)*(By)*(1.0/(mu))))
#define Gmx_pz(ro,px,py,pz,gamm,Bx,By,Bz,U) (((px)*(pz)/(ro)) - ((Bx)*(Bz)*(1.0/(mu))))
#define Gmx_ro(px) ((px))


#define Gmy_ro(py) ((py))
#define Gmz_ro(pz) ((pz))

#define Gym_U(ro,px,py,pz,gamm,Bx,By,Bz,U)  ((( (U)+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm)) )*( (py)/ (ro)  ))+((0.5/(mu))*(  ((Bx)*(Bx) +(By)*(By) +(Bz)*(Bz))*( (py)/ (ro)  ) )) - ((1.0/(mu))*( (By)*( (Bx)*(px) +(By)*(py) +(Bz)*(pz))/(ro) ))  )
#define Gzm_U(ro,px,py,pz,gamm,Bx,By,Bz,U)  ((( (U)+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm)) )*( (pz)/ (ro)  ))+((0.5/(mu))*(  ((Bx)*(Bx) +(By)*(By) +(Bz)*(Bz))*( (pz)/ (ro)  ) )) - ((1.0/(mu))*( (Bz)*( (Bx)*(px) +(By)*(py) +(Bz)*(pz))/(ro) ))  )

#define Gmy_Bx(ro,px,py,pz,gamm,Bx,By,Bz,U)  ( ((Bx)*(py)-(By)*(px))/(ro))
#define Gmz_Bx(ro,px,py,pz,gamm,Bx,By,Bz,U)  ( ((Bx)*(pz)-(Bz)*(px))/(ro))

#define Gmy_By(ro,px,py,pz,gamm,Bx,By,Bz,U) (0.0)
#define Gmz_By(ro,px,py,pz,gamm,Bx,By,Bz,U) ( ((By)*(pz)-(Bz)*(py))/(ro))

#define Gmy_Bz(ro,px,py,pz,gamm,Bx,By,Bz,U)   ( ((Bz)*(py)-(By)*(pz))/(ro))
#define Gmz_Bz(ro,px,py,pz,gamm,Bx,By,Bz,U) (0.0)

#define Gmy_px(ro,px,py,pz,gamm,Bx,By,Bz,U) (((px)*(py)/(ro)) - ((Bx)*(By)*(1.0/(mu))))
#define Gmz_px(ro,px,py,pz,gamm,Bx,By,Bz,U) (((px)*(pz)/(ro)) - ((Bx)*(Bz)*(1.0/(mu))))

#define Gmy_py(ro,px,py,pz,gamm,Bx,By,Bz,U) (((py)*(py)/(ro))+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm) )+ (((Bx)*(Bx)+(Bz)*(Bz)-(By)*(By))*(1.0/(mu*2.0))))
#define Gmz_py(ro,px,py,pz,gamm,Bx,By,Bz,U) (((pz)*(py)/(ro)) - ((Bz)*(By)*(1.0/(mu))))

#define Gmy_pz(ro,px,py,pz,gamm,Bx,By,Bz,U) (((py)*(pz)/(ro)) - ((By)*(Bz)*(1.0/(mu))))
#define Gmz_pz(ro,px,py,pz,gamm,Bx,By,Bz,U) (((pz)*(pz)/(ro))+(ppp_MHD(ro,px,py,pz,Bx,By,Bz,U,gamm) )+ (((Bx)*(Bx)+(By)*(By)-(Bz)*(Bz))*(1.0/(mu*2.0))))




/**
*  Electromagnetic fields and sources defined for each local grid, and for an implicit maxwell's solver 
*   
* @date May 2008
* @par Copyright:
* (C) 2008 KUL
* @author Stefano Markidis, Giovanni Lapenta.
* @version 3.0
*/

class EMfields3D : public Field{
  public:
  /** constructor */
  EMfields3D(CollectiveIO *col,Grid *grid);
  /** destructor */
  ~EMfields3D();

 /** initialize the electromagnetic fields with constant values */
  void init(VirtualTopology3D *vct, Grid *grid);
  /** init beam */
  void initBEAM(VirtualTopology3D *vct, Grid *grid, double x_center, double y_center, double z_center, double radius);
  /** initialize GEM challenge */
  void initGEM(VirtualTopology3D *vct, Grid *grid);
  /** initialize GEM challenge with no Perturbation */
  void initGEMnoPert(VirtualTopology3D *vct, Grid *grid);
    
    
  void initGEMdoublePeriodicMHD(VirtualTopology3D *vct, Grid *grid);
  /**  Init Force Free (JxB=0) */
  void initForceFree(VirtualTopology3D *vct, Grid *grid);
  /**  Init 3D null point(s) */
  void initNullPoint(VirtualTopology3D *vct, Grid *grid);
  /**  Init 3D null point(s) following Galsgaard & Nordlund (1997) */
  void initNullGalsgaard(VirtualTopology3D *vct, Grid *grid);
  /** initialized with rotated magnetic field */
  void initEM_rotate(VirtualTopology3D *vct, Grid *grid, double B, double theta);
  /** add a perturbattion to charge density */
  void AddPerturbationRho(double deltaBoB, double kx, double ky, double Bx_mod, double By_mod, double Bz_mod, double ne_mod, double ne_phase, double ni_mod, double ni_phase, double B0, Grid *grid);
  /** add a perturbattion to the EM field */
  void AddPerturbation(double deltaBoB, double kx, double ky, double Ex_mod, double Ex_phase, double Ey_mod, double Ey_phase, double Ez_mod, double Ez_phase, double Bx_mod, double Bx_phase, double By_mod, double By_phase, double Bz_mod, double Bz_phase, double B0, Grid *grid);
  
  
  /** Calculate Electric field using the implicit Maxwell solver */
  void calculateE(Grid *grid, VirtualTopology3D *vct);
  /** Image of Poisson Solver (for SOLVER)*/
  void PoissonImage(double *image, double *vector, Grid *grid, VirtualTopology3D *vct);
  /** Image of Maxwell Solver (for Solver) */
  void MaxwellImage(double *im, double *vector, Grid *grid,VirtualTopology3D *vct);     
  /** Maxwell source term (for SOLVER) */
  void MaxwellSource(double *bkrylov, Grid *grid, VirtualTopology3D *vct);
  /** Calculate Magnetic field with the implicit solver: calculate B defined on nodes
  With E(n+ theta) computed, the magnetic field is evaluated from Faraday's law */
  void calculateB(Grid *grid, VirtualTopology3D *vct);
  /** fix B on the boundary for gem challange*/
  void fixBgem(Grid *grid, VirtualTopology3D *vct);
  /** fix B on the boundary for gem challange*/
  void fixBforcefree(Grid *grid, VirtualTopology3D *vct);
  
  /** Calculate the three components of Pi(implicit pressure) cross image vector */
  void PIdot(double ***PIdotX, double ***PIdotY, double ***PIdotZ, double ***vectX, double ***vectY, double ***vectZ, int ns, Grid *grid);
  /** Calculate the three components of mu (implicit permeattivity) cross image vector */
  void MUdot(double ***MUdotX, double ***MUdotY, double ***MUdotZ, double ***vectX, double ***vectY, double ***vectZ, Grid *grid);
  /** Calculate rho hat, Jx hat, Jy hat, Jz hat */
  void calculateHatFunctions(Grid *grid, VirtualTopology3D *vct);
 
    
  // MHD_update
	void calculateFieldMHD(Grid *grid, VirtualTopology3D *vct,int step_cycle);
    int calc_Lx(int ord,Grid *grid, VirtualTopology3D *vct);
    int calc_Ly(int ord,Grid *grid, VirtualTopology3D *vct);
    int calc_Lz(int ord,Grid *grid, VirtualTopology3D *vct);

    void PoissonImageMHD(double *image, double *vector, Grid *grid, VirtualTopology3D *vct);
    void Poisson_correction_MHD(Grid *grid, VirtualTopology3D *vct);
    void Boundarize_fi(int nx, int ny, int nz, double ***mfi, VirtualTopology3D *vct);
    void Boundarize(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, VirtualTopology3D *vct);
    
    /**  Init 3D null point(s). MHD version */
    void initNullPoint_MHD(VirtualTopology3D *vct, Grid *grid);
    void initRhoTest(VirtualTopology3D *vct, Grid *grid);
    void initRhoTest2(VirtualTopology3D *vct, Grid *grid);
    
    double mu;
    double C_g;
    
    void calc_dt(Grid *grid, VirtualTopology3D *vct);
    double F_cq(double Bxd,double Byd,double Bzd,double Bqd,double rod,double gd,double Pd,double vqd);
    void put_resistivity(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid);
    void put_Sources_fluxes(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid);

    double*** getBxC();
    double*** getByC();
    double*** getBzC();
    
    double*** getpxC();
    double*** getpyC();
    double*** getpzC();
    
    double*** getUC();
    double*** getRHO0();
    double*** getgrid_eta();
    int getStopFlag();
    
     double &getBxC(int indexX, int indexY, int indexZ) const;
     double &getByC(int indexX, int indexY, int indexZ) const;
     double &getBzC(int indexX, int indexY, int indexZ) const;
     double &getpxC(int indexX, int indexY, int indexZ) const;
     double &getpyC(int indexX, int indexY, int indexZ) const;
     double &getpzC(int indexX, int indexY, int indexZ) const;
     double &getRHO0(int indexX, int indexY, int indexZ) const;
     double &getgrid_eta(int indexX, int indexY, int indexZ) const;
    
    //get current time
    double getMHDtime();
    
    /** get the magnetic field energy*/
    double getBenergyMHD(void);

    /** get the total energy*/
    double getTotenergyMHD(void);
    double getKeMHD(void);
    

    void Fix_P(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid);

    

  /** communicate ghost for densities and interp rho from node to center */
  void interpDensitiesN2C(VirtualTopology3D *vct, Grid *grid);
  /** set to 0 all the densities fields */
  void setZeroDensities();
  /** Sum rhon over species */
  void sumOverSpecies(VirtualTopology3D *vct);
  /** Sum current over different species */
  void sumOverSpeciesJ();
  /** Smoothing after the interpolation**/
  void smooth(double value ,double ***vector,bool type, Grid *grid, VirtualTopology3D *vct);
  /** SPECIES: Smoothing after the interpolation for species fields**/
  void smooth(double value ,double ****vector,int is, bool type, Grid *grid, VirtualTopology3D *vct);
  /** smooth the electric field */
  void smoothE(double value, VirtualTopology3D *vct);
  
  /** communicate ghost for grid -> Particles interpolation */
  void communicateGhostP2G(int ns, int bcFaceXright, int bcFaceXleft, int bcFaceYright, int bcFaceYleft, VirtualTopology3D *vct);
    /** add an amount of charge density to charge density field at node X,Y,Z */
  void addRho(double weight[][2][2], int X, int Y,int Z, int is);
  /** add an amount of current density - direction X to current density field at node X,Y,Z */
  void addJx(double weight[][2][2], int X, int Y, int Z,int is);
  /** add an amount of current density - direction Y to current density field at node X,Y,Z */
  void addJy(double weight[][2][2], int X, int Y, int Z, int is);
  /** add an amount of current density - direction Z to current density field at node X,Y,Z */
  void addJz(double weight[][2][2], int X, int Y, int Z, int is);
 
  /** add an amount of pressure density - direction XX to current density field at node X,Y,Z */
  void addPxx(double weight[][2][2], int X, int Y, int Z, int is);
  /** add an amount of pressure density - direction XY to current density field at node X,Y,Z */
  void addPxy(double weight[][2][2], int X, int Y, int Z,int is);
  /** add an amount of pressure density - direction XZ to current density field at node X,Y,Z */
  void addPxz(double weight[][2][2], int X, int Y, int Z,int is);
  /** add an amount of pressure density - direction YY to current density field at node X,Y,Z */
  void addPyy(double weight[][2][2], int X, int Y, int Z, int is);
  /** add an amount of pressure density - direction YZ to current density field at node X,Y,Z */
  void addPyz(double weight[][2][2], int X, int Y, int Z, int is);
  /** add an amount of pressure density - direction ZZ to current density field at node X,Y,Z */
  void addPzz(double weight[][2][2], int X, int Y, int Z,int is);
  
   /** adjust densities on boundaries that are not periodic */
  void adjustNonPeriodicDensities(int is, VirtualTopology3D *vct);

  
  /** Perfect conductor boundary conditions LEFT wall */
  void perfectConductorLeft(double ***imageX, double ***imageY, double ***imageZ,double ***vectorX, double ***vectorY,double ***vectorZ,int dir,Grid *grid);
  /** Perfect conductor boundary conditions RIGHT wall */
  void perfectConductorRight(double ***imageX, double ***imageY, double ***imageZ,double ***vectorX, double ***vectorY, double ***vectorZ,int dir,Grid *grid);
  /** Perfect conductor boundary conditions for source LEFT wall*/
  void perfectConductorLeftS(double ***vectorX, double ***vectorY, double ***vectorZ,int dir);
  /** Perfect conductor boundary conditions for source RIGHT wall*/
  void perfectConductorRightS(double ***vectorX, double ***vectorY, double ***vectorZ,int dir);
  
  /** Calculate the sysceptibility tensor on the boundary */
  void sustensorRightY(double** susxy,double** susyy,double** suszy);
  void sustensorLeftY(double** susxy,double** susyy,double** suszy);
   
  /** get Potential array */
  double*** getPHI();
  /** get Electric Field component X defined on node(indexX,indexY,indexZ) */
  double &getEx(int indexX, int indexY, int indexZ) const;
  /** get Electric field X component array */
  double*** getEx() ;
   /** get Electric Field component Y defined on node(indexX,indexY,indexZ) */
  double &getEy(int indexX, int indexY, int indexZ) const;
  /** get Electric field Y component array */
  double*** getEy() ;
  /** get Electric Field component Z defined on node(indexX,indexY,indexZ) */
  double &getEz(int indexX, int indexY, int indexZ) const;
  /** get Electric field Z component array */
  double*** getEz() ;
  /** get Magnetic Field component X defined on node(indexX,indexY,indexZ) */
  double &getBx(int indexX, int indexY, int indexZ) const;
  /** get Magnetic field X component array */
  double*** getBx();
  /** get Magnetic Field component Y defined on node(indexX,indexY,indexZ) */
  double &getBy(int indexX, int indexY, int indexZ) const;
  /** get Magnetic field Y component array */
  double*** getBy();
  /** get Magnetic Field component Z defined on node(indexX,indexY,indexZ) */
  double &getBz(int indexX, int indexY, int indexZ) const;
  /** get Magnetic field Z component array */
  double*** getBz() ;
  /** get density on cell(indexX,indexY,indexZ)  */
  double &getRHOc(int indexX, int indexY, int indexZ) const;
  /** get density array on center cell */
  double*** getRHOc() ;
  /** get density on nodes(indexX,indexY,indexZ)  */
  double &getRHOn(int indexX, int indexY, int indexZ) const;
  /** get density array on nodes  */
  double*** getRHOn() ;
  /** SPECIES: get density on nodes(indexX,indexY,indexZ)*/
  double &getRHOns(int indexX, int indexY,int indexZ,int is) const;
  /** SPECIES: get density on center cell(indexX,indexY,indexZ)*/
  double &getRHOcs(int indexX, int indexY,int indexZ,int is) const;
  /** SPECIES: get density array on nodes*/
  double**** getRHOns() ;



  /** get pressure tensor XX for species */
  double**** getpXXsn() ;
  /** get pressure tensor XY for species */
  double**** getpXYsn() ;
  /** get pressure tensor XZ for species*/
  double**** getpXZsn() ;
  /** get pressure tensor YY for species */
  double**** getpYYsn() ;
  /** get pressure tensor YZ for species */
  double**** getpYZsn() ;
  /** get pressure tensor ZZ for species */
  double**** getpZZsn() ;
    
  /** get Jx(X,Y,Z)  */
  double &getJx(int indexX, int indexY, int indexZ) const;
  /** get current -Direction X */
  double*** getJx();
  /** get Jxs(X,Y,Z,is)  */
  double &getJxs(int indexX, int indexY, int indexZ, int is) const;
  /** SPECIES: get current -Direction X */
  double**** getJxs();
  /** get Jy(X,Y,Z)  */
  double &getJy(int indexX, int indexY, int indexZ) const;
  /** get current -Direction Y */
  double*** getJy();
  /** get Jys(X,Y,Z,is)  */
  double &getJys(int indexX, int indexY, int indexZ, int is) const;
  /** SPECIES: get current -Direction Y */
  double**** getJys();
  /** get Jz(X,Y,Z)  */
  double &getJz(int indexX, int indexY, int indexZ) const;
  /** get current -Direction Z */
  double*** getJz();
  /** get Jzs(X,Y,Z,is)  */
  double &getJzs(int indexX, int indexY, int indexZ, int is) const;
  /** SPECIES: get current -Direction Z */
  double**** getJzs();
  /** get the electric field energy */
  double getEenergy();
  /** get the magnetic field energy */
  double getBenergy();  
  /** get bulk kinetic energy */
  double getBulkEnergy(int is);

  
  /** print electromagnetic fields info */
  void print(void) const;
  
// **********************************
// VARIABLES
// ********************************** 
private:
  /** light speed */
  double c;
  /* 4*PI for normalization */
  double FourPI;
  /** time step */
  double dt;
  /** decentering parameter */
  double th;
  /** Smoothing value*/
  double Smooth;
  /** delt = c*th*dt */
  double delt;
  /** number of particles species */
  int ns;
  /** GEM challenge parameters */
  double B0x, B0y, B0z, delta;
  /** charge to mass ratio array for different species */
  double *qom;
  /** thermal velocity for different species */
  double *uth, *vth, *wth;
  

  // KEEP IN MEMORY GUARD CELLS ARE INCLUDED
  /** number of cells - X direction, including + 2 (guard cells) */
  int nxc;
  /** number of nodes - X direction, including + 2 extra nodes for guard cells */
  int nxn;
  /** number of cell - Y direction, including + 2 (guard cells) */
  int nyc;
  /** number of nodes - Y direction, including + 2 extra nodes for guard cells */
  int nyn;
  /** number of cell - Z direction, including + 2 (guard cells) */
  int nzc;
  /** number of nodes - Z direction, including + 2 extra nodes for guard cells */
  int nzn;
  /** local grid boundaries coordinate  */
  double xStart, xEnd, yStart, yEnd, zStart, zEnd;
  /** grid spacing */
  double dx, dy, dz, invVOL;
  /** simulation box length - X direction   */
  double Lx;
  /** simulation box length - Y direction   */
  double Ly;
  /** simulation box length - Z direction   */
  double Lz;
  

  /** PHI: electric potential (indexX, indexY, indexZ), defined on central points between nodes*/
  double ***PHI;
  /** Ex: electric field X-component (indexX, indexY, indexZ), defined on nodes */
  double ***Ex;
  /** Exth: implicit electric field X-component (indexX, indexY, indexZ), defined on nodes */
  double ***Exth;
  /** Ey: electric field Y-component (indexX, indexY, indexZ), defined on nodes */
  double ***Ey;
  /** Eyth: implicit electric field Y-component (indexX, indexY, indexZ), defined on nodes */
  double ***Eyth;
  /** Ez: electric field Z-component (indexX, indexY, indexZ, #species), defined on nodes */
  double ***Ez;
  /** Ezth: implicit electric field Z-component (indexX, indexY, indexZ), defined on nodes */
  double ***Ezth;
    
// MHD_update
  /** Bxc: magnetic field X-component (indexX, indexY, indexZ), defined on central points between nodes*/
//  double ***Bxc;
  /** Byc: magnetic field Y-component (indexX, indexY, indexZ), defined on central points between nodes*/
//  double ***Byc;
  /** Bzc: magnetic field Z-component (indexX, indexY, indexZ), defined on central points between nodes*/
//  double ***Bzc;
  /** Bxn: magnetic field X-component (indexX, indexY, indexZ), defined on nodes*/
  double ***Bxn;
  /** Byn: magnetic field Y-component (indexX, indexY, indexZ), defined on nodes*/
  double ***Byn;
  /** Bzn: magnetic field Z-component (indexX, indexY, indexZ), defined on nodes*/
  double ***Bzn;
 
  // *************************************
  // TEMPORARY ARRAY
  // *************************************
   /**some temporary arrays (for calculate hat functions)*/
  double*** tempXC;
  double*** tempYC;
  double*** tempZC;
  double*** tempXN;
  double*** tempYN;
  double*** tempZN;
  /** other temporary arrays (in MaxwellSource) */			  
  double*** tempC; 
  double*** tempX;
  double*** tempY;
  double*** tempZ;
  double*** temp2X;
  double*** temp2Y;
  double*** temp2Z;
  /** and some for MaxwellImage */
  double*** imageX; 
  double*** imageY; 
  double*** imageZ;
  double*** Dx;
  double*** Dy;
  double*** Dz;
  double*** vectX;
  double*** vectY;
  double*** vectZ;
  double*** divC;
 
    //All new declarations in EMfields.h, MHD_update
    string Case;
    double invdx;
	double invdy;
	double invdz;

    
	double***  Bxc;
	double***  Byc;
	double***  Bzc;
    
	double***  Uc;
    
	double***  pxc;
	double***  pyc;
	double***  pzc;
    
	double***  rhoc;
    
    // Flux limiters
    
	double***  dBxc;
	double***  dByc;
	double***  dBzc;
    
	double***  dUc;
    
	double***  dpxc;
	double***  dpyc;
	double***  dpzc;
    
	double***  drhoc;
    
    //Sources
    
	double***  Q__Bxc;
	double***  Q__Byc;
	double***  Q__Bzc;
    
	double***  Q__Uc;
    
	double***  Q__pxc;
	double***  Q__pyc;
	double***  Q__pzc;
    
	double***  Q__rhoc;
    
    //Keep the copy during the timestep
    
	double***  Bxc1;
	double***  Byc1;
	double***  Bzc1;
    
	double***  Uc1;
    
	double***  pxc1;
	double***  pyc1;
	double***  pzc1;
    
	double***  rhoc1;
    
    
	double***  Bxc0;
	double***  Byc0;
	double***  Bzc0;
    
	double***  Uc0;
    
	double***  pxc0;
	double***  pyc0;
	double***  pzc0;
    
	double***  rhoc0;
    
    
    
	double***  Bxc2;
	double***  Byc2;
	double***  Bzc2;
    
	double***  Uc2;
    
	double***  pxc2;
	double***  pyc2;
	double***  pzc2;
    
	double***  rhoc2;
    
    //Store the copy of initial variables
	double***	roStored;
	double***	pxStored;
	double***	pyStored;
	double***	pzStored;
	double***	BxStored;
	double***	ByStored;
	double***	BzStored;
	double***	UStored;
    
	double***	grid_eta;
    double***       grid_dx;
    double***       grid_dy;
    double***       grid_dz;
    double MHDtime; 
    int stopFlag;
    
    
    
    
    
    
  // *********************************************************************************
  // *************                SOURCES                                           **
  // *********************************************************************************

  /** Charge density, defined on central points of the cell */

  //MHD_UPDATE, SEE DEFINITION BELOW
  //  double*** rhoc;
  /** Charge density, defined on nodes */
  double*** rhon;
  /** Implicit charge density, defined on central points of the cell */
  double***  rhoh;
  /** SPECIES: charge density for each species, defined on nodes */
  double****  rhons;
  /** SPECIES: charge density for each species, defined on central points of the cell */
  double**** rhocs;
  /** Current density component-X, defined on nodes */
  double***  Jx;
  /** Current density component-Y, defined on nodes */
  double***  Jy;
  /** Current density component-Z, defined on nodes */
  double***  Jz;
  /** Implicit current density X-component, defined on nodes */
  double***  Jxh;
  /** Implicit current density Y-component, defined on nodes */
  double***  Jyh;
  /** Implicit current density Z-component, defined on nodes */
  double***  Jzh;
  /** SPECIES: current density component-X for species, defined on nodes */
  double**** Jxs;
  /** SPECIES: current density component-Y for species, defined on nodes */
  double****  Jys;
  /** SPECIES: current density component-Z for species, defined on nodes */
  double****  Jzs;
  
  /** SPECIES: pressure tensor component-XX, defined on nodes */
  double**** pXXsn;
  /** SPECIES: pressure tensor component-XY, defined on nodes */
  double ****pXYsn;
  /** SPECIES: pressure tensor component-XZ, defined on nodes */
  double ****pXZsn;
  /** SPECIES: pressure tensor component-XZ, defined on nodes */
  double ****pYYsn;
  /** SPECIES: pressure tensor component-YZ, defined on nodes */
  double ****pYZsn;
  /** SPECIES: pressure tensor component-ZZ, defined on nodes */
  double ****pZZsn;
  

  /** Field Boundary Condition
   0 =  Dirichlet Boundary Condition: specifies the value to take on the boundary of the domain
   1 =  Neumann Boundary Condition: specifies the value of derivative to take on the boundary of the domain
   2 =  Periodic condition
  */

  /** Boundary Condition Electrostatic Potential: FaceXright */
  int bcPHIfaceXright;
  /** Boundary Condition Electrostatic Potential:FaceXleft */
  int bcPHIfaceXleft;
  /** Boundary Condition Electrostatic Potential:FaceYright */
  int bcPHIfaceYright;
  /** Boundary Condition Electrostatic Potential:FaceYleft */
  int bcPHIfaceYleft;
  /** Boundary Condition Electrostatic Potential:FaceZright */
  int bcPHIfaceZright;
  /** Boundary Condition Electrostatic Potential:FaceZleft */
  int bcPHIfaceZleft;

  /** Boundary condition for electric field
  0 = perfect conductor
  1 = magnetic mirror
  */
  /** Boundary Condition EM Field: FaceXright */
  int bcEMfaceXright;
  /** Boundary Condition EM Field: FaceXleft */
  int bcEMfaceXleft;
  /** Boundary Condition EM Field: FaceYright */
  int bcEMfaceYright;
  /** Boundary Condition EM Field: FaceYleft */
  int bcEMfaceYleft;
  /** Boundary Condition EM Field: FaceZright */
  int bcEMfaceZright;
  /** Boundary Condition EM Field: FaceZleft */
  int bcEMfaceZleft;

  
  /** GEM Challenge background ion */
  double *rhoINIT; 
  /** Drift of the species */
  bool *DriftSpecies;
  
  /** boolean for divergence cleaning */
  bool PoissonCorrection;
  /** RESTART BOOLEAN */
  int restart1 ; 
  /** String with the directory for the restart file */
  string RestartDirName;
  
   /** CG tolerance criterium for stopping iterations */
  double CGtol;
  /**  GMRES tolerance criterium for stopping iterations */
  double GMREStol;
    
};

/** constructor */
inline EMfields3D::EMfields3D(CollectiveIO *col,Grid *grid){
  nxc = grid->getNXC();
  nxn = grid->getNXN();
  nyc = grid->getNYC();
  nyn = grid->getNYN();
  nzc = grid->getNZC();
  nzn = grid->getNZN();
  dx = grid->getDX();
  dy = grid ->getDY();
  dz = grid->getDZ();
  invVOL = grid->getInvVOL();
  //MHD_update
    
	invdx = 1.0/dx;
	invdy = 1.0/dy;
	invdz = 1.0/dz;

    mu=1.0;
    C_g=1.66666;
    
    
    Case = col->getCase();
    
  xStart = grid->getXstart();
  xEnd   = grid->getXend();
  yStart = grid->getYstart();
  yEnd   = grid->getYend();
  zStart = grid->getZstart();
  zEnd   = grid->getZend();
  Lx = col->getLx();
  Ly = col->getLy();
  Lz = col->getLz();
  ns  = col->getNs();
  c = col->getC();
  dt = col->getDt();
  th = col->getTh();
  delt = c*th*dt;
  PoissonCorrection = true;
  CGtol = col->getCGtol();
  GMREStol = col->getGMREStol();
  qom = new double[ns];
  uth = new double[ns];
  vth = new double[ns];
  wth = new double[ns];
  for (int i=0; i < ns;i++) {
    qom[i] = col->getQOM(i);
    uth[i] = col->getUth(i);
    vth[i] = col->getVth(i);
    wth[i] = col->getWth(i);
  }
  // boundary conditions: PHI and EM fields
  bcPHIfaceXright  = col->getBcPHIfaceXright();
  bcPHIfaceXleft   = col->getBcPHIfaceXleft();
  bcPHIfaceYright  = col->getBcPHIfaceYright();
  bcPHIfaceYleft   = col->getBcPHIfaceYleft();
  bcPHIfaceZright  = col->getBcPHIfaceZright();
  bcPHIfaceZleft   = col->getBcPHIfaceZleft();

  bcEMfaceXright   = col->getBcEMfaceXright();
  bcEMfaceXleft    = col->getBcEMfaceXleft();
  bcEMfaceYright   = col->getBcEMfaceYright();
  bcEMfaceYleft    = col->getBcEMfaceYleft();
  bcEMfaceZright   = col->getBcEMfaceZright();
  bcEMfaceZleft    = col->getBcEMfaceZleft();
  // GEM challenge parameters
  B0x = col->getB0x();
  B0y = col->getB0y();
  B0z = col->getB0z();
  delta = col->getDelta();
  Smooth = col->getSmooth();
  // get the density background for the gem Challange
  rhoINIT = new double[ns];
  DriftSpecies = new bool[ns];
  for (int i=0; i < ns;i++){
	    rhoINIT[i] = col->getRHOinit(i);
	    if ((fabs(col->getW0(i))!=0) || (fabs(col->getU0(i))!=0)) // GEM and LHDI
		  DriftSpecies[i] = true;
		else
		  DriftSpecies[i] = false;
  }
  /** parameters for GEM challenge */
  FourPI = 16 * atan(1.0);
  /** Restart */
  restart1 = col->getRestart_status();
  RestartDirName = col->getRestartDirName();
  // arrays allocation: nodes
  Ex = newArr3(double, nxn, nyn, nzn);
  Ey = newArr3(double, nxn, nyn, nzn);
  Ez = newArr3(double, nxn, nyn, nzn);
  Exth = newArr3(double, nxn, nyn, nzn);
  Eyth = newArr3(double, nxn, nyn, nzn);
  Ezth = newArr3(double, nxn, nyn, nzn);
  Bxn = newArr3(double, nxn, nyn, nzn);
  Byn = newArr3(double, nxn, nyn, nzn);
  Bzn = newArr3(double, nxn, nyn, nzn);
  rhon = newArr3(double, nxn, nyn, nzn);
  Jx = newArr3(double, nxn, nyn, nzn);
  Jy = newArr3(double, nxn, nyn, nzn);
  Jz = newArr3(double, nxn, nyn, nzn);
  Jxh = newArr3(double, nxn, nyn, nzn);
  Jyh = newArr3(double, nxn, nyn, nzn);
  Jzh = newArr3(double, nxn, nyn, nzn);
  // involving species
  rhons = newArr4(double, ns, nxn, nyn, nzn);
  rhocs = newArr4(double, ns, nxc, nyc, nzc);
  Jxs = newArr4(double, ns, nxn, nyn, nzn);
  Jys = newArr4(double, ns, nxn, nyn, nzn);
  Jzs = newArr4(double, ns, nxn, nyn, nzn);
  pXXsn = newArr4(double, ns, nxn, nyn, nzn);
  pXYsn = newArr4(double, ns, nxn, nyn, nzn);
  pXZsn = newArr4(double, ns, nxn, nyn, nzn);
  pYYsn = newArr4(double, ns, nxn, nyn, nzn);
  pYZsn = newArr4(double, ns, nxn, nyn, nzn);
  pZZsn = newArr4(double, ns, nxn, nyn, nzn);
  // arrays allocation: central points 
  PHI = newArr3(double, nxc, nyc, nzc);
// See declarations below (MHD_update)
//  Bxc = newArr3(double, nxc, nyc, nzc);
//  Byc = newArr3(double, nxc, nyc, nzc);
//  Bzc = newArr3(double, nxc, nyc, nzc);
//  rhoc = newArr3(double, nxc, nyc, nzc);
  rhoh = newArr3(double, nxc, nyc, nzc);

  // temporary arrays
  tempXC = newArr3(double, nxc, nyc, nzc);
  tempYC = newArr3(double, nxc, nyc, nzc);
  tempZC = newArr3(double, nxc, nyc, nzc);

  tempXN = newArr3(double, nxn, nyn, nzn);
  tempYN = newArr3(double, nxn, nyn, nzn);
  tempZN = newArr3(double, nxn, nyn, nzn);
  tempC = newArr3(double, nxc, nyc, nzc);
  tempX = newArr3(double, nxn, nyn, nzn);
  tempY = newArr3(double, nxn, nyn, nzn);
  tempZ = newArr3(double, nxn, nyn, nzn);
  temp2X = newArr3(double, nxn, nyn, nzn);
  temp2Y = newArr3(double, nxn, nyn, nzn);
  temp2Z = newArr3(double, nxn, nyn, nzn);
  imageX = newArr3(double, nxn, nyn, nzn);
  imageY = newArr3(double, nxn, nyn, nzn);
  imageZ = newArr3(double, nxn, nyn, nzn);
  Dx = newArr3(double, nxn, nyn, nzn);
  Dy = newArr3(double, nxn, nyn, nzn);
  Dz = newArr3(double, nxn, nyn, nzn);
  vectX = newArr3(double, nxn, nyn, nzn);
  vectY = newArr3(double, nxn, nyn, nzn);
  vectZ = newArr3(double, nxn, nyn, nzn);
  divC = newArr3(double, nxc, nyc, nzc);
    
    
    
    //main variables MHD_update
    Bxc= newArr3(double,nxc,nyc,nzc);
    Byc= newArr3(double,nxc,nyc,nzc);
    Bzc= newArr3(double,nxc,nyc,nzc);
    
    Uc= newArr3(double,nxc,nyc,nzc);
    
    pxc= newArr3(double,nxc,nyc,nzc);
    pyc= newArr3(double,nxc,nyc,nzc);
    pzc= newArr3(double,nxc,nyc,nzc);
    
    rhoc= newArr3(double,nxc,nyc,nzc);
    
    grid_eta= newArr3(double,nxc,nyc,nzc);
    
    // Flux limiters
    dBxc= newArr3(double,nxc,nyc,nzc);
    dByc= newArr3(double,nxc,nyc,nzc);
    dBzc= newArr3(double,nxc,nyc,nzc);
    
    dUc= newArr3(double,nxc,nyc,nzc);
    
    dpxc= newArr3(double,nxc,nyc,nzc);
    dpyc= newArr3(double,nxc,nyc,nzc);
    dpzc= newArr3(double,nxc,nyc,nzc);
    
    drhoc= newArr3(double,nxc,nyc,nzc);
    
    
    // Sources
    Q__Bxc= newArr3(double,nxc,nyc,nzc);
    Q__Byc= newArr3(double,nxc,nyc,nzc);
    Q__Bzc= newArr3(double,nxc,nyc,nzc);
    
    Q__Uc= newArr3(double,nxc,nyc,nzc);
    Q__pxc= newArr3(double,nxc,nyc,nzc);
    Q__pyc= newArr3(double,nxc,nyc,nzc);
    Q__pzc= newArr3(double,nxc,nyc,nzc);
    
    Q__rhoc= newArr3(double,nxc,nyc,nzc);
    
    //Keep the copy
    Bxc1= newArr3(double,nxc,nyc,nzc);
    Byc1= newArr3(double,nxc,nyc,nzc);
    Bzc1= newArr3(double,nxc,nyc,nzc);
    
    Uc1= newArr3(double,nxc,nyc,nzc);
    pxc1= newArr3(double,nxc,nyc,nzc);
    pyc1= newArr3(double,nxc,nyc,nzc);
    pzc1= newArr3(double,nxc,nyc,nzc);
    
    rhoc1= newArr3(double,nxc,nyc,nzc);
    
    
    // ANOTHER 2 COPIES!!!
    
    Bxc2= newArr3(double,nxc,nyc,nzc);
    Byc2= newArr3(double,nxc,nyc,nzc);
    Bzc2= newArr3(double,nxc,nyc,nzc);
    
    Uc2= newArr3(double,nxc,nyc,nzc);
    pxc2= newArr3(double,nxc,nyc,nzc);
    pyc2= newArr3(double,nxc,nyc,nzc);
    pzc2= newArr3(double,nxc,nyc,nzc);
    
    rhoc2= newArr3(double,nxc,nyc,nzc);
    
    Bxc0= newArr3(double,nxc,nyc,nzc);
    Byc0= newArr3(double,nxc,nyc,nzc);
    Bzc0= newArr3(double,nxc,nyc,nzc);
    
    Uc0= newArr3(double,nxc,nyc,nzc);
    pxc0= newArr3(double,nxc,nyc,nzc);
    pyc0= newArr3(double,nxc,nyc,nzc);
    pzc0= newArr3(double,nxc,nyc,nzc);
    rhoc0= newArr3(double,nxc,nyc,nzc);
    
    
    //Store the copy
    roStored= newArr3(double,nxc,nyc,nzc);
    pxStored= newArr3(double,nxc,nyc,nzc);
    pyStored= newArr3(double,nxc,nyc,nzc);
    pzStored= newArr3(double,nxc,nyc,nzc);
    BxStored= newArr3(double,nxc,nyc,nzc);
    ByStored= newArr3(double,nxc,nyc,nzc);
    BzStored= newArr3(double,nxc,nyc,nzc);
    UStored= newArr3(double,nxc,nyc,nzc);
    
    
    
    
}


//inline void EMfields3D::PoissonImageMHD(double *image, double *vector, Grid *grid, VirtualTopology3D *vct){
//    // allocate  2 two dimensional service vectors
//    double ***temp = newArr3(double,nxc,nyc,nzc);
//    double ***im  = newArr3(double,nxc,nyc,nzc);
//    eqValue (0.0, image,(nxc-4)*(nyc-4)*(nzc-4));
//    eqValue (0.0, temp,nxc,nyc,nzc);
//    eqValue (0.0, im,nxc,nyc,nzc);
//    // move from krylov space to physical space and communicate ghost cells
//    solver2phys25D(temp,vector,nxc,nyc,nzc);
//    // MPI_Barrier(MPI_COMM_WORLD);
//    //communicateCenter(nxc,nyc,temp,vct);
//    communicateCenter25D(nxc,nyc,nzc,temp,vct);
//	// adjust boundarie before calculating the lap
//    Boundarize_fi(nxc,nyc,nzc,temp,vct);
//    
//	// calculate the laplacian
//    grid->lapC2C25D(im,temp,vct);
//    phys2solver25D(image,im,nxc,nyc,nzc);
//    // deallocate temporary array and objects
//    delArr3(temp,nxc,nyc);
//    delArr3(im,nxc,nyc);
//}

//
//inline void EMfields3D::Poisson_correction_MHD(Grid *grid, VirtualTopology3D *vct){
//    int x,y,z;
//    
//    //Need (copy-paste from calculateFields) :
//    //EMfields.h
//    //cout<<"start allocation in Poisson_correction_MHD"<<endl;
//    double ***divB = newArr3(double,nxc,nyc,nzc);
//    double ***divBPhi = newArr3(double,nxc,nyc,nzc);
//    double ***gradPHIX = newArr3(double,nxc,nyc,nzn);
//    double ***gradPHIY = newArr3(double,nxc,nyc,nzc);
//    double ***gradPHIZ = newArr3(double,nxc,nyc,nzc);
//    
//    
//	double *xkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
//    double *bkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
//    //	double ***PHI = newArr3(double,nxc,nyc,nzc);
//    
//    
//    //eqValue (0.0, xkrylov, 3*(nxn-2)*(nyn-2));
//    //eqValue (0.0, bkrylov, 3*(nxn-2)*(nyn-2));
//    //cout<<"start eqvalue in Poisson_correction_MHD"<<endl;
//    
//    eqValue (0.0, divBPhi,nxc,nyc,nzc);
//    eqValue (0.0, divB,nxc,nyc,nzc);
//	
//    //  ?      eqValue (0.0, tempC,nxc,nyc);
//    eqValue (0.0, gradPHIX,nxc,nyc,nzc);
//    eqValue (0.0, gradPHIY,nxc,nyc,nzc);
//    eqValue (0.0, gradPHIZ,nxc,nyc,nzc);
//    //cout<<"eqvalue done in Poisson_correction_MHD"<<endl;
////    
////    //divC2C25D
//    grid->divC2CMHD(divB,Bxc0,Byc0,Bzc0,vct);
////    //                scale(tempC,rhoc,-FourPI,nxc,nyc);
////    //                sum(divE,tempC,nxc,nyc);
////    //cout<<"divC2C25D in Poisson_correction_MHD"<<endl;
////    
//     phys2solverMHD(bkrylovPoisson,divB,nxc,nyc,nzc); // bkrylovPoisson is defined only in active centers
////    //cout<<"phys2solver25D in Poisson_correction_MHD"<<endl;
////    
//     eqValue(0.0,xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4));
////    //cout<<"eqvalue xkrylovPoisson in Poisson_correction_MHD"<<endl;
////    
////    //?                if (vct->getPERIODICX() && vct->getPERIODICY())
////    //?                        GMRES(&Field::PoissonImage, xkrylovPoisson, (nxc-2)*(nyc-2),bkrylovPoisson,20,200,GMREStol, grid, vct, this);
////    //?                else
////    //CG(xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4),bkrylovPoisson, 10000, CGtol, &Field::PoissonImageMHD, grid, vct, this);
////    //cout<<"CG in Poisson_correction_MHD"<<endl;
////    
////
//      solver2physMHD(divBPhi,xkrylovPoisson,nxc,nyc,nzc);
//    BEGIN_full
//    if (!(divBPhi[x][y][z]==divBPhi[x][y][z])) cout<< "in divBphi NaN"<<"x : "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<vct->getCartesian_rank()<<endl;
//    END_full
//    
////    //cout<<"solver2phys in Poisson_correction_MHD"<<endl;
////
//     // MPI_Barrier(MPI_COMM_WORLD);
//
//     communicateCenterMHD(nxc,nyc,nzc,divBPhi,vct);
//    cout<<" divBphi "<<divBPhi[10][10][10]<<endl;
//    cout<<" divB "<<divB[10][10][10]<<endl;
//    cout<<" divBphi "<<divBPhi[10][20][10]<<endl;
//    cout<<" gradPHIX "<<gradPHIX[10][10][10]<<endl;
//    cout<<" gradPHIY "<<gradPHIY[10][10][10]<<endl;
//    cout<<" gradPHIZ "<<gradPHIZ[10][10][10]<<endl;
//    cout<<" xkrylovPoisson "<<xkrylovPoisson[20]<<endl;
//    cout<<" bkrylovPoisson "<<bkrylovPoisson[20]<<endl;
//    
//    //cout<<"communicatecenter divBPhi in Poisson_correction_MHD"<<endl;
//    
//    Boundarize_fi(nxc,nyc,nzc,divBPhi,vct);
//    //cout<<"Boundarize in Poisson_correction_MHD"<<endl;
//    
//    grid->gradC2CMHD(gradPHIX,gradPHIY,gradPHIZ,divBPhi,vct);
//    //cout<<"gradc2c25d in Poisson_correction_MHD"<<endl;
//    BEGIN_full
//    Uc0[x][y][z]+=-(Bxc0[x][y][z]*gradPHIX[x][y][z])-(Byc0[x][y][z]*gradPHIY[x][y][z])-(Bzc0[x][y][z]*gradPHIZ[x][y][z])+(0.5*((gradPHIX[x][y][z]*gradPHIX[x][y][z])+(gradPHIY[x][y][z]*gradPHIY[x][y][z]) + (gradPHIZ[x][y][z]*gradPHIZ[x][y][z])      )  );
//    END_full
//    
//    
//    
//    
//    sub(Bxc0,gradPHIX,nxc,nyc,nzc);
//    sub(Byc0,gradPHIY,nxc,nyc,nzc);
//    sub(Bzc0,gradPHIZ,nxc,nyc,nzc);
//    //cout<<"sub in Poisson_correction_MHD"<<endl;
//    
//    delArr3(gradPHIX,nxc,nyc);
//    delArr3(gradPHIY,nxc,nyc);
//    delArr3(gradPHIZ,nxc,nyc);
//    delArr3(divB,nxc,nyc);
//    delArr3(divBPhi,nxc,nyc);
//    
//    delete[] xkrylovPoisson;
//    delete[] bkrylovPoisson;
//    //cout<<"delete is done in Poisson_correction_MHD"<<endl;
//    
//    
//}
//
//inline void EMfields3D::PoissonImageMHD(double *image, double *vector, Grid *grid, VirtualTopology3D *vct){
//    // cout<<"tes t1"<<endl;
//    // allocate  2 two dimensional service vectors
//    double ***temp = newArr3(double,nxc,nyc,nzc);
//    double ***im  = newArr3(double,nxc,nyc,nzc);
//    // cout<<"tes t1"<<endl;
//    
//    eqValue (0.0, image,(nxc-4)*(nyc-4)*(nzc-4));
//    eqValue (0.0, temp,nxc,nyc,nzc);
//    eqValue (0.0, im,nxc,nyc,nzc);
//    // move from krylov space to physical space and communicate ghost cells
//    //cout<<"tes t2"<<endl;
//    
//    solver2physMHD(temp,vector,nxc,nyc,nzc);
//    //cout<<"tes t3"<<endl;
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    //communicateCenter(nxc,nyc,temp,vct);
//    //cout<<"tes t4"<<endl;
//    
//    communicateCenter(nxc,nyc,nzc,temp,vct);
//	// adjust boundarie before calculating the lap
//    //cout<<"tes t5"<<endl;
//    
//    Boundarize_fi(nxc,nyc,nzc,temp,vct);
//    //cout<<"tes t6"<<endl;
//    
//	// calculate the laplacian
//    // cout<<"tes t7"<<endl;
//    
//    grid->lapC2CMHD(im,temp,vct);
//    // cout<<"tes t8: "<<im[10][2][2]<<endl;
//    phys2solverMHD(image,im,nxc,nyc,nzc);
//    // deallocate temporary array and objects
//    // delArr3(temp,nxc,nyc);
//    // delArr3(im,nxc,nyc);
//    // cout<<"tes t9"<<endl;
//    
//}
//
//inline void EMfields3D::Poisson_correction_MHD(Grid *grid, VirtualTopology3D *vct){
//    int x,y,z;
//    
//    //Need (copy-paste from calculateFields) :
//    //EMfields.h
//    //cout<<"start allocation in Poisson_correction_MHD"<<endl;
//    
//    double ***divB = newArr3(double,nxc,nyc,nzc);
//    double ***divBPhi = newArr3(double,nxc,nyc,nzc);
//    double ***gradPHIX = newArr3(double,nxc,nyc,nzn);
//    double ***gradPHIY = newArr3(double,nxc,nyc,nzc);
//    double ***gradPHIZ = newArr3(double,nxc,nyc,nzc);
//    
//    
//	double *xkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
//    double *bkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
//    //	double ***PHI = newArr3(double,nxc,nyc,nzc);
//    
//    
//    //eqValue (0.0, xkrylov, 3*(nxn-2)*(nyn-2));
//    //eqValue (0.0, bkrylov, 3*(nxn-2)*(nyn-2));
//    cout<<"start eqvalue in Poisson_correction_MHD"<<endl;
//    
//    eqValue (0.0, divBPhi,nxc,nyc,nzc);
//    eqValue (0.0, divB,nxc,nyc,nzc);
//	
//    //  ?      eqValue (0.0, tempC,nxc,nyc);
//    eqValue (0.0, gradPHIX,nxc,nyc,nzc);
//    eqValue (0.0, gradPHIY,nxc,nyc,nzc);
//    eqValue (0.0, gradPHIZ,nxc,nyc,nzc);
//    cout<<"eqvalue done in Poisson_correction_MHD"<<endl;
//    
//    //divC2CMHD
//    grid->divC2CMHD(divB,Bxc0,Byc0,Bzc0,vct);
//    //                scale(tempC,rhoc,-FourPI,nxc,nyc);
//    //                sum(divE,tempC,nxc,nyc);
//    cout<<"divC2CMHD in Poisson_correction_MHD: "<<divB[10][10][10]<<endl;
//    
//    phys2solverMHD(bkrylovPoisson,divB,nxc,nyc,nzc); // bkrylovPoisson is defined only in active centers
//    cout<<"phys2solverMHD in Poisson_correction_MHD"<<endl;
//    
//    eqValue(0.0,xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4));
//    cout<<"eqvalue xkrylovPoisson in Poisson_correction_MHD"<<endl;
//    
//    //?                if (vct->getPERIODICX() && vct->getPERIODICY())
//    //?                        GMRES(&Field::PoissonImage, xkrylovPoisson, (nxc-2)*(nyc-2),bkrylovPoisson,20,200,GMREStol, grid, vct, this);
//    //?                else
//    
//    CG(xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4),bkrylovPoisson, 10000, CGtol, &Field::PoissonImageMHD, grid, vct, this);
//    
//    cout<<"CG in Poisson_correction_MHD"<<endl;
//    
//    
//    solver2physMHD(divBPhi,xkrylovPoisson,nxc,nyc,nzc);
//    cout<<"solver2phys in Poisson_correction_MHD"<<endl;
//    
//    communicateCenter(nxc,nyc,nzc,divBPhi,vct);
//    cout<<"communicatecenter divBPhi in Poisson_correction_MHD"<<endl;
//    
//    Boundarize_fi(nxc,nyc,nzc,divBPhi,vct);
//    cout<<"Boundarize in Poisson_correction_MHD"<<endl;
//    
//    grid->gradC2CMHD(gradPHIX,gradPHIY,gradPHIZ,divBPhi,vct);
//    cout<<"gradC2CMHD in Poisson_correction_MHD"<<endl;
//    BEGIN_full
//    Uc0[x][y][z]+=-(Bxc0[x][y][z]*gradPHIX[x][y][z])-(Byc0[x][y][z]*gradPHIY[x][y][z])-(Bzc0[x][y][z]*gradPHIZ[x][y][z])+(0.5*((gradPHIX[x][y][z]*gradPHIX[x][y][z])+(gradPHIY[x][y][z]*gradPHIY[x][y][z]) + (gradPHIZ[x][y][z]*gradPHIZ[x][y][z])      )  );
//    END_full
//    
//    
//    
//    
//    sub(Bxc0,gradPHIX,nxc,nyc,nzc);
//    sub(Byc0,gradPHIY,nxc,nyc,nzc);
//    sub(Bzc0,gradPHIZ,nxc,nyc,nzc);
//    cout<<"sub in Poisson_correction_MHD"<<endl;
//    
//    delArr3(gradPHIX,nxc,nyc);    cout<<"delete 1 is done in Poisson_correction_MHD"<<endl;
//    
//    delArr3(gradPHIY,nxc,nyc);    cout<<"delete 2 is done in Poisson_correction_MHD"<<endl;
//    
//    delArr3(gradPHIZ,nxc,nyc);    cout<<"delete 3 is done in Poisson_correction_MHD"<<endl;
//    
//    delArr3(divB,nxc,nyc);    cout<<"delete 4 is done in Poisson_correction_MHD"<<endl;
//    
//    delArr3(divBPhi,nxc,nyc);    cout<<"delete 5 is done in Poisson_correction_MHD"<<endl;
//    
//    cout<<"delete 0 is done in Poisson_correction_MHD"<<endl;
//    
//    delete[] xkrylovPoisson;
//    delete[] bkrylovPoisson;
//    cout<<"delete is done in Poisson_correction_MHD"<<endl;
//    
//    
//}



// UPDATE:   MHD_update

inline void EMfields3D::PoissonImageMHD(double *image, double *vector, Grid *grid, VirtualTopology3D *vct){
    // allocate  2 two dimensional service vectors
    double ***temp = newArr3(double,nxc,nyc,nzc);
    double ***im  = newArr3(double,nxc,nyc,nzc);
    eqValue (0.0, image,(nxc-4)*(nyc-4)*(nzc-4));
    eqValue (0.0, temp,nxc,nyc,nzc);
    eqValue (0.0, im,nxc,nyc,nzc);
    // move from krylov space to physical space and communicate ghost cells
    solver2physMHD(temp,vector,nxc,nyc,nzc);
    //communicateCenter(nxc,nyc,temp,vct);
    communicateCenterMHD(nxc,nyc,nzc,temp,vct);
	// adjust boundarie before calculating the lap
    Boundarize_fi(nxc,nyc,nzc,temp,vct);
    
	// calculate the laplacian
    grid->lapC2CMHD(im,temp,vct);
    phys2solverMHD(image,im,nxc,nyc,nzc);
    // deallocate temporary array and objects
    delArr3(temp,nxc,nyc);
    delArr3(im,nxc,nyc);
}


inline void EMfields3D::Poisson_correction_MHD(Grid *grid, VirtualTopology3D *vct){
    int x,y,z;
    if (vct->getCartesian_rank() == 0)
        cout << "*** Poisson Correction for B ***" << endl;
    //Need (copy-paste from calculateFields) :
    //EMfields.h
    //cout<<"start allocation in Poisson_correction_MHD"<<endl;
    double ***divB = newArr3(double,nxc,nyc,nzc);
    double ***divBPhi = newArr3(double,nxc,nyc,nzc);
    double ***gradPHIX = newArr3(double,nxc,nyc,nzn);
    double ***gradPHIY = newArr3(double,nxc,nyc,nzc);
    double ***gradPHIZ = newArr3(double,nxc,nyc,nzc);
    
    
	double *xkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
    double *bkrylovPoisson = new double[(nxc-4)*(nyc-4)*(nzc-4)];
    //	double ***PHI = newArr3(double,nxc,nyc,nzc);
    
    
    //eqValue (0.0, xkrylov, 3*(nxn-2)*(nyn-2));
    //eqValue (0.0, bkrylov, 3*(nxn-2)*(nyn-2));
    //cout<<"start eqvalue in Poisson_correction_MHD"<<endl;
    
    eqValue (0.0, divBPhi,nxc,nyc,nzc);
    eqValue (0.0, divB,nxc,nyc,nzc);
	
    //  ?      eqValue (0.0, tempC,nxc,nyc);
    eqValue (0.0, gradPHIX,nxc,nyc,nzc);
    eqValue (0.0, gradPHIY,nxc,nyc,nzc);
    eqValue (0.0, gradPHIZ,nxc,nyc,nzc);
    //cout<<"eqvalue done in Poisson_correction_MHD"<<endl;
    
    //divC2C25D
    grid->divC2CMHD(divB,Bxc0,Byc0,Bzc0,vct);
    //                scale(tempC,rhoc,-FourPI,nxc,nyc);
    //                sum(divE,tempC,nxc,nyc);
    //cout<<"divC2C25D in Poisson_correction_MHD"<<endl;
    
    phys2solverMHD(bkrylovPoisson,divB,nxc,nyc,nzc); // bkrylovPoisson is defined only in active centers
    //cout<<"phys2solver25D in Poisson_correction_MHD"<<endl;
    
    eqValue(0.0,xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4));
    //cout<<"eqvalue xkrylovPoisson in Poisson_correction_MHD"<<endl;
    
    //?                if (vct->getPERIODICX() && vct->getPERIODICY())
    //?                        GMRES(&Field::PoissonImage, xkrylovPoisson, (nxc-2)*(nyc-2),bkrylovPoisson,20,200,GMREStol, grid, vct, this);
    //?                else
    CG(xkrylovPoisson,(nxc-4)*(nyc-4)*(nzc-4),bkrylovPoisson, 10000, CGtol, &Field::PoissonImageMHD, grid, vct, this);
    //cout<<"CG in Poisson_correction_MHD"<<endl;
    
    
    solver2physMHD(divBPhi,xkrylovPoisson,nxc,nyc,nzc);
    //cout<<"solver2phys in Poisson_correction_MHD"<<endl;
    
    communicateCenterMHD(nxc,nyc,nzc,divBPhi,vct);
    //cout<<"communicatecenter divBPhi in Poisson_correction_MHD"<<endl;
    
    Boundarize_fi(nxc,nyc,nzc,divBPhi,vct);
    //cout<<"Boundarize in Poisson_correction_MHD"<<endl;
    
    grid->gradC2CMHD(gradPHIX,gradPHIY,gradPHIZ,divBPhi,vct);
    //cout<<"gradc2c25d in Poisson_correction_MHD"<<endl;
    BEGIN_full
    Uc0[x][y][z]+=-(Bxc0[x][y][z]*gradPHIX[x][y][z])-(Byc0[x][y][z]*gradPHIY[x][y][z])-(Bzc0[x][y][z]*gradPHIZ[x][y][z])+(0.5*((gradPHIX[x][y][z]*gradPHIX[x][y][z])+(gradPHIY[x][y][z]*gradPHIY[x][y][z]) + (gradPHIZ[x][y][z]*gradPHIZ[x][y][z])      )  );
    END_full
    
    
    
    
    sub(Bxc0,gradPHIX,nxc,nyc,nzc);
    sub(Byc0,gradPHIY,nxc,nyc,nzc);
    sub(Bzc0,gradPHIZ,nxc,nyc,nzc);
    //cout<<"sub in Poisson_correction_MHD"<<endl;
    
    delArr3(gradPHIX,nxc,nyc);
    delArr3(gradPHIY,nxc,nyc);
    delArr3(gradPHIZ,nxc,nyc);
    delArr3(divB,nxc,nyc);
    delArr3(divBPhi,nxc,nyc);
    
    delete[] xkrylovPoisson;
    delete[] bkrylovPoisson;
    //cout<<"delete is done in Poisson_correction_MHD"<<endl;
    
    
}




/** Calculate Electric field with the implicit solver: the Maxwell solver method is called here */
inline void EMfields3D::calculateE(Grid * grid, VirtualTopology3D * vct) {
  if (vct->getCartesian_rank() == 0)
    cout << "*** E CALCULATION ***" << endl;
  double ***divE = newArr3(double, nxc, nyc, nzc);
  double ***gradPHIX = newArr3(double, nxn, nyn, nzn);
  double ***gradPHIY = newArr3(double, nxn, nyn, nzn);
  double ***gradPHIZ = newArr3(double, nxn, nyn, nzn);

  double *xkrylov = new double[3 * (nxn - 2) * (nyn - 2) * (nzn - 2)];  // 3 E components
  double *bkrylov = new double[3 * (nxn - 2) * (nyn - 2) * (nzn - 2)];  // 3 components

  double *xkrylovPoisson = new double[(nxc - 2) * (nyc - 2) * (nzc - 2)];
  double *bkrylovPoisson = new double[(nxc - 2) * (nyc - 2) * (nzc - 2)];
  // set to zero all the stuff 
  eqValue(0.0, xkrylov, 3 * (nxn - 2) * (nyn - 2) * (nzn - 2));
  eqValue(0.0, xkrylovPoisson, (nxc - 2) * (nyc - 2) * (nzc - 2));
  eqValue(0.0, bkrylov, 3 * (nxn - 2) * (nyn - 2) * (nzn - 2));
  eqValue(0.0, divE, nxc, nyc, nzc);
  eqValue(0.0, tempC, nxc, nyc, nzc);
  eqValue(0.0, gradPHIX, nxn, nyn, nzn);
  eqValue(0.0, gradPHIY, nxn, nyn, nzn);
  eqValue(0.0, gradPHIZ, nxn, nyn, nzn);
  // Adjust E calculating laplacian(PHI) = div(E) -4*PI*rho DIVERGENCE CLEANING
  if (PoissonCorrection) {
    if (vct->getCartesian_rank() == 0)
      cout << "*** DIVERGENCE CLEANING ***" << endl;
    grid->divN2C(divE, Ex, Ey, Ez);
    scale(tempC, rhoc, -FourPI, nxc, nyc, nzc);
    sum(divE, tempC, nxc, nyc, nzc);
    // move to krylov space 
    phys2solver(bkrylovPoisson, divE, nxc, nyc, nzc);
    // use conjugate gradient first
    if (!CG(xkrylovPoisson, (nxc - 2) * (nyc - 2) * (nzc - 2), bkrylovPoisson, 3000, CGtol, &Field::PoissonImage, grid, vct, this)) {
      if (vct->getCartesian_rank() == 0)
        cout << "CG not Converged. Trying with GMRes. Consider to increase the number of the CG iterations" << endl;
      eqValue(0.0, xkrylovPoisson, (nxc - 2) * (nyc - 2) * (nzc - 2));
      GMRES(&Field::PoissonImage, xkrylovPoisson, (nxc - 2) * (nyc - 2) * (nzc - 2), bkrylovPoisson, 20, 200, GMREStol, grid, vct, this);
    }
    solver2phys(PHI, xkrylovPoisson, nxc, nyc, nzc);
    communicateCenterBC(nxc, nyc, nzc, PHI, 2, 2, 2, 2, 2, 2, vct);
    // calculate the gradient
    grid->gradC2N(gradPHIX, gradPHIY, gradPHIZ, PHI);
    // sub
    sub(Ex, gradPHIX, nxn, nyn, nzn);
    sub(Ey, gradPHIY, nxn, nyn, nzn);
    sub(Ez, gradPHIZ, nxn, nyn, nzn);

  }                             // end of divergence cleaning 
  if (vct->getCartesian_rank() == 0)
    cout << "*** MAXWELL SOLVER ***" << endl;
  // prepare the source 
  MaxwellSource(bkrylov, grid, vct);
  phys2solver(xkrylov, Ex, Ey, Ez, nxn, nyn, nzn);
  // solver
  GMRES(&Field::MaxwellImage, xkrylov, 3 * (nxn - 2) * (nyn - 2) * (nzn - 2), bkrylov, 20, 200, GMREStol, grid, vct, this);
  // move from krylov space to physical space
  solver2phys(Exth, Eyth, Ezth, xkrylov, nxn, nyn, nzn);



  addscale(1 / th, -(1.0 - th) / th, Ex, Exth, nxn, nyn, nzn);
  addscale(1 / th, -(1.0 - th) / th, Ey, Eyth, nxn, nyn, nzn);
  addscale(1 / th, -(1.0 - th) / th, Ez, Ezth, nxn, nyn, nzn);

  // apply to smooth to electric field 3 times
  smoothE(Smooth, vct);
  smoothE(Smooth, vct);
  smoothE(Smooth, vct);


  // communicate so the interpolation can have good values
  communicateNodeBC(nxn, nyn, nzn, Exth, 1, 1, 1, 1, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Eyth, 1, 1, 2, 2, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Ezth, 1, 1, 1, 1, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Ex, 1, 1, 1, 1, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Ey, 1, 1, 2, 2, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Ez, 1, 1, 1, 1, 1, 1, vct);


  // deallocate temporary arrays
  delete[]xkrylov;
  delete[]bkrylov;
  delete[]xkrylovPoisson;
  delete[]bkrylovPoisson;
  delArr3(divE, nxc, nyc);
  delArr3(gradPHIX, nxn, nyn);
  delArr3(gradPHIY, nxn, nyn);
  delArr3(gradPHIZ, nxn, nyn);

}

/** Calculate sorgent for Maxwell solver */
inline void EMfields3D::MaxwellSource(double *bkrylov, Grid * grid, VirtualTopology3D * vct) {
  eqValue(0.0, tempC, nxc, nyc, nzc);
  eqValue(0.0, tempX, nxn, nyn, nzn);
  eqValue(0.0, tempY, nxn, nyn, nzn);
  eqValue(0.0, tempZ, nxn, nyn, nzn);
  eqValue(0.0, tempXN, nxn, nyn, nzn);
  eqValue(0.0, tempYN, nxn, nyn, nzn);
  eqValue(0.0, tempZN, nxn, nyn, nzn);
  eqValue(0.0, temp2X, nxn, nyn, nzn);
  eqValue(0.0, temp2Y, nxn, nyn, nzn);
  eqValue(0.0, temp2Z, nxn, nyn, nzn);
  // communicate
  communicateCenterBC(nxc, nyc, nzc, Bxc, 2, 2, 2, 2, 2, 2, vct);
  communicateCenterBC(nxc, nyc, nzc, Byc, 1, 1, 1, 1, 1, 1, vct);
  communicateCenterBC(nxc, nyc, nzc, Bzc, 2, 2, 2, 2, 2, 2, vct);
  // fixBforcefree(grid,vct);
  //fixBgem(grid, vct);
  // prepare curl of B for known term of Maxwell solver: for the source term
  grid->curlC2N(tempXN, tempYN, tempZN, Bxc, Byc, Bzc);
  scale(temp2X, Jxh, -FourPI / c, nxn, nyn, nzn);
  scale(temp2Y, Jyh, -FourPI / c, nxn, nyn, nzn);
  scale(temp2Z, Jzh, -FourPI / c, nxn, nyn, nzn);


  sum(temp2X, tempXN, nxn, nyn, nzn);
  sum(temp2Y, tempYN, nxn, nyn, nzn);
  sum(temp2Z, tempZN, nxn, nyn, nzn);
  scale(temp2X, delt, nxn, nyn, nzn);
  scale(temp2Y, delt, nxn, nyn, nzn);
  scale(temp2Z, delt, nxn, nyn, nzn);

  communicateCenterBC_P(nxc, nyc, nzc, rhoh, 2, 2, 2, 2, 2, 2, vct);
  grid->gradC2N(tempX, tempY, tempZ, rhoh);

  scale(tempX, -delt * delt * FourPI, nxn, nyn, nzn);
  scale(tempY, -delt * delt * FourPI, nxn, nyn, nzn);
  scale(tempZ, -delt * delt * FourPI, nxn, nyn, nzn);
  // sum E, past values
  sum(tempX, Ex, nxn, nyn, nzn);
  sum(tempY, Ey, nxn, nyn, nzn);
  sum(tempZ, Ez, nxn, nyn, nzn);
  // sum curl(B) + jhat part
  sum(tempX, temp2X, nxn, nyn, nzn);
  sum(tempY, temp2Y, nxn, nyn, nzn);
  sum(tempZ, temp2Z, nxn, nyn, nzn);

  // Boundary condition in the known term
  // boundary condition: Xleft
  if (vct->getXleft_neighbor() == -1 && bcEMfaceXleft == 0)  // perfect conductor
    perfectConductorLeftS(tempX, tempY, tempZ, 0);
  // boundary condition: Xright
  if (vct->getXright_neighbor() == -1 && bcEMfaceXright == 0)  // perfect conductor
    perfectConductorRightS(tempX, tempY, tempZ, 0);
  // boundary condition: Yleft
  if (vct->getYleft_neighbor() == -1 && bcEMfaceYleft == 0)  // perfect conductor
    perfectConductorLeftS(tempX, tempY, tempZ, 1);
  // boundary condition: Yright
  if (vct->getYright_neighbor() == -1 && bcEMfaceYright == 0)  // perfect conductor
    perfectConductorRightS(tempX, tempY, tempZ, 1);
  // boundary condition: Zleft
  if (vct->getZleft_neighbor() == -1 && bcEMfaceZleft == 0)  // perfect conductor
    perfectConductorLeftS(tempX, tempY, tempZ, 2);
  // boundary condition: Zright
  if (vct->getZright_neighbor() == -1 && bcEMfaceZright == 0)  // perfect conductor
    perfectConductorRightS(tempX, tempY, tempZ, 2);

  // physical space -> Krylov space
  phys2solver(bkrylov, tempX, tempY, tempZ, nxn, nyn, nzn);

}
/** Mapping of Maxwell image to give to solver */
inline void EMfields3D::MaxwellImage(double *im, double *vector, Grid * grid, VirtualTopology3D * vct) {
  eqValue(0.0, im, 3 * (nxn - 2) * (nyn - 2) * (nzn - 2));
  eqValue(0.0, imageX, nxn, nyn, nzn);
  eqValue(0.0, imageY, nxn, nyn, nzn);
  eqValue(0.0, imageZ, nxn, nyn, nzn);
  eqValue(0.0, tempX, nxn, nyn, nzn);
  eqValue(0.0, tempY, nxn, nyn, nzn);
  eqValue(0.0, tempZ, nxn, nyn, nzn);
  eqValue(0.0, Dx, nxn, nyn, nzn);
  eqValue(0.0, Dy, nxn, nyn, nzn);
  eqValue(0.0, Dz, nxn, nyn, nzn);
  // move from krylov space to physical space
  solver2phys(vectX, vectY, vectZ, vector, nxn, nyn, nzn);
  grid->lapN2N(imageX, vectX, vct);
  grid->lapN2N(imageY, vectY, vct);
  grid->lapN2N(imageZ, vectZ, vct);
  neg(imageX, nxn, nyn, nzn);
  neg(imageY, nxn, nyn, nzn);
  neg(imageZ, nxn, nyn, nzn);
  // grad(div(mu dot E(n + theta)) mu dot E(n + theta) = D
  MUdot(Dx, Dy, Dz, vectX, vectY, vectZ, grid);
  grid->divN2C(divC, Dx, Dy, Dz);
  // communicate you should put BC 
  // think about the Physics 
  // communicateCenterBC(nxc,nyc,nzc,divC,1,1,1,1,1,1,vct);
  communicateCenterBC(nxc, nyc, nzc, divC, 2, 2, 2, 2, 2, 2, vct);  // GO with Neumann, now then go with rho


  grid->gradC2N(tempX, tempY, tempZ, divC);

  // -lap(E(n +theta)) - grad(div(mu dot E(n + theta))
  sub(imageX, tempX, nxn, nyn, nzn);
  sub(imageY, tempY, nxn, nyn, nzn);
  sub(imageZ, tempZ, nxn, nyn, nzn);

  // scale delt*delt
  scale(imageX, delt * delt, nxn, nyn, nzn);
  scale(imageY, delt * delt, nxn, nyn, nzn);
  scale(imageZ, delt * delt, nxn, nyn, nzn);

  // -lap(E(n +theta)) - grad(div(mu dot E(n + theta)) + eps dot E(n + theta)
  sum(imageX, Dx, nxn, nyn, nzn);
  sum(imageY, Dy, nxn, nyn, nzn);
  sum(imageZ, Dz, nxn, nyn, nzn);
  sum(imageX, vectX, nxn, nyn, nzn);
  sum(imageY, vectY, nxn, nyn, nzn);
  sum(imageZ, vectZ, nxn, nyn, nzn);

  // boundary condition: Xleft
  if (vct->getXleft_neighbor() == -1 && bcEMfaceXleft == 0)  // perfect conductor
    perfectConductorLeft(imageX, imageY, imageZ, vectX, vectY, vectZ, 0, grid);
  // boundary condition: Xright
  if (vct->getXright_neighbor() == -1 && bcEMfaceXright == 0)  // perfect conductor
    perfectConductorRight(imageX, imageY, imageZ, vectX, vectY, vectZ, 0, grid);
  // boundary condition: Yleft
  if (vct->getYleft_neighbor() == -1 && bcEMfaceYleft == 0)  // perfect conductor
    perfectConductorLeft(imageX, imageY, imageZ, vectX, vectY, vectZ, 1, grid);
  // boundary condition: Yright
  if (vct->getYright_neighbor() == -1 && bcEMfaceYright == 0)  // perfect conductor
    perfectConductorRight(imageX, imageY, imageZ, vectX, vectY, vectZ, 1, grid);
  // boundary condition: Zleft
  if (vct->getZleft_neighbor() == -1 && bcEMfaceZleft == 0)  // perfect conductor
    perfectConductorLeft(imageX, imageY, imageZ, vectX, vectY, vectZ, 2, grid);
  // boundary condition: Zright
  if (vct->getZright_neighbor() == -1 && bcEMfaceZright == 0)  // perfect conductor
    perfectConductorRight(imageX, imageY, imageZ, vectX, vectY, vectZ, 2, grid);
  // move from physical space to krylov space
  phys2solver(im, imageX, imageY, imageZ, nxn, nyn, nzn);


}

/** Calculate PI dot (vectX, vectY, vectZ)*/
inline void EMfields3D::PIdot(double ***PIdotX, double ***PIdotY, double ***PIdotZ, double ***vectX, double ***vectY, double ***vectZ, int ns, Grid * grid) {
  double beta, edotb, omcx, omcy, omcz, denom;
  beta = .5 * qom[ns] * dt / c;
  for (int i = 1; i < nxn - 1; i++)
    for (int j = 1; j < nyn - 1; j++)
      for (int k = 1; k < nzn - 1; k++) {
        omcx = beta * Bxn[i][j][k];
        omcy = beta * Byn[i][j][k];
        omcz = beta * Bzn[i][j][k];
        edotb = vectX[i][j][k] * omcx + vectY[i][j][k] * omcy + vectZ[i][j][k] * omcz;
        denom = 1 / (1.0 + omcx * omcx + omcy * omcy + omcz * omcz);
        PIdotX[i][j][k] += (vectX[i][j][k] + (vectY[i][j][k] * omcz - vectZ[i][j][k] * omcy + edotb * omcx)) * denom;
        PIdotY[i][j][k] += (vectY[i][j][k] + (vectZ[i][j][k] * omcx - vectX[i][j][k] * omcz + edotb * omcy)) * denom;
        PIdotZ[i][j][k] += (vectZ[i][j][k] + (vectX[i][j][k] * omcy - vectY[i][j][k] * omcx + edotb * omcz)) * denom;
      }


}
/** Calculate MU dot (vectX, vectY, vectZ)*/
inline void EMfields3D::MUdot(double ***MUdotX, double ***MUdotY, double ***MUdotZ, double ***vectX, double ***vectY, double ***vectZ, Grid * grid) {
  double beta, edotb, omcx, omcy, omcz, denom;
  for (int i = 1; i < nxn - 1; i++)
    for (int j = 1; j < nyn - 1; j++)
      for (int k = 1; k < nzn - 1; k++) {
        MUdotX[i][j][k] = 0.0;
        MUdotY[i][j][k] = 0.0;
        MUdotZ[i][j][k] = 0.0;
      }
  for (int is = 0; is < ns; is++) {
    beta = .5 * qom[is] * dt / c;
    for (int i = 1; i < nxn - 1; i++)
      for (int j = 1; j < nyn - 1; j++)
        for (int k = 1; k < nzn - 1; k++) {
          omcx = beta * Bxn[i][j][k];
          omcy = beta * Byn[i][j][k];
          omcz = beta * Bzn[i][j][k];
          edotb = vectX[i][j][k] * omcx + vectY[i][j][k] * omcy + vectZ[i][j][k] * omcz;
          denom = FourPI / 2 * delt * dt / c * qom[is] * rhons[is][i][j][k] / (1.0 + omcx * omcx + omcy * omcy + omcz * omcz);
          MUdotX[i][j][k] += (vectX[i][j][k] + (vectY[i][j][k] * omcz - vectZ[i][j][k] * omcy + edotb * omcx)) * denom;
          MUdotY[i][j][k] += (vectY[i][j][k] + (vectZ[i][j][k] * omcx - vectX[i][j][k] * omcz + edotb * omcy)) * denom;
          MUdotZ[i][j][k] += (vectZ[i][j][k] + (vectX[i][j][k] * omcy - vectY[i][j][k] * omcx + edotb * omcz)) * denom;
        }

  }

}
/*Interpolation smoothing:
Smoothing (vector must already have ghost cells)
TO MAKE SMOOTH value as to be different from 1.0
type = 0 --> center based vector ; 
type = 1 --> node based vector   ;                     
*/
inline void  EMfields3D::smooth(double value,double ***vector, bool type, Grid *grid, VirtualTopology3D *vct){

  }
/* Interpolation smoothing: Smoothing (vector must already have ghost cells) TO MAKE SMOOTH value as to be different from 1.0 type = 0 --> center based vector ; type = 1 --> node based vector ; */
inline void EMfields3D::smoothE(double value, VirtualTopology3D * vct) {

}

/* SPECIES: Interpolation smoothing
TO MAKE SMOOTH value as to be different from 1.0
type = 0 --> center based vector  
type = 1 --> node based vector                        
*/
inline void  EMfields3D::smooth(double value,double ****vector,int is, bool type, Grid *grid, VirtualTopology3D *vct){
  cout << "Smoothing for Species not implemented in 3D" << endl; 
}

/** fix the B boundary when running gem*/
inline void EMfields3D::fixBgem(Grid * grid, VirtualTopology3D * vct) {
  if (vct->getYright_neighbor() == -1) {
    for (int i = 0; i < nxc; i++)
      for (int k = 0; k < nzc; k++) {
        Bxc[i][nyc - 1][k] = B0x * tanh((grid->getYC(i, nyc - 1, k) - Ly / 2) / delta);
        Bxc[i][nyc - 2][k] = Bxc[i][nyc - 1][k];
        Bxc[i][nyc - 3][k] = Bxc[i][nyc - 1][k];
        Byc[i][nyc - 1][k] = B0y;
        Bzc[i][nyc - 1][k] = B0z;
        Bzc[i][nyc - 2][k] = B0z;
        Bzc[i][nyc - 3][k] = B0z;
      }
  }
  if (vct->getYleft_neighbor() == -1) {
    for (int i = 0; i < nxc; i++)
      for (int k = 0; k < nzc; k++) {
        Bxc[i][0][k] = B0x * tanh((grid->getYC(i, 0, k) - Ly / 2) / delta);
        Bxc[i][1][k] = Bxc[i][0][k];
        Bxc[i][2][k] = Bxc[i][0][k];
        Byc[i][0][k] = B0y;
        Bzc[i][0][k] = B0z;
        Bzc[i][1][k] = B0z;
        Bzc[i][2][k] = B0z;
      }
  }

}

/** fix the B boundary when running forcefree*/
inline void EMfields3D::fixBforcefree(Grid * grid, VirtualTopology3D * vct) {
  if (vct->getYright_neighbor() == -1) {
    for (int i = 0; i < nxc; i++)
      for (int k = 0; k < nzc; k++) {
        Bxc[i][nyc - 1][k] = B0x * tanh((grid->getYC(i, nyc - 1, k) - Ly / 2) / delta);
        Byc[i][nyc - 1][k] = B0y;
        Bzc[i][nyc - 1][k] = B0z / cosh((grid->getYC(i, nyc - 1, k) - Ly / 2) / delta);;
        Bzc[i][nyc - 2][k] = B0z / cosh((grid->getYC(i, nyc - 2, k) - Ly / 2) / delta);;
        Bzc[i][nyc - 3][k] = B0z / cosh((grid->getYC(i, nyc - 3, k) - Ly / 2) / delta);
      }
  }
  if (vct->getYleft_neighbor() == -1) {
    for (int i = 0; i < nxc; i++)
      for (int k = 0; k < nzc; k++) {
        Bxc[i][0][k] = B0x * tanh((grid->getYC(i, 0, k) - Ly / 2) / delta);
        Byc[i][0][k] = B0y;
        Bzc[i][0][k] = B0z / cosh((grid->getYC(i, 0, k) - Ly / 2) / delta);
        Bzc[i][1][k] = B0z / cosh((grid->getYC(i, 1, k) - Ly / 2) / delta);
        Bzc[i][2][k] = B0z / cosh((grid->getYC(i, 2, k) - Ly / 2) / delta);
      }
  }

}


/** adjust densities on boundaries that are not periodic */
inline void EMfields3D::adjustNonPeriodicDensities(int is, VirtualTopology3D * vct) {
  //

}
/** Calculate Magnetic field with the implicit solver: calculate B defined on nodes
  With E(n+ theta) computed, the magnetic field is evaluated from Faraday's law */
inline void EMfields3D::calculateB(Grid * grid, VirtualTopology3D * vct) {
  if (vct->getCartesian_rank() == 0)
    cout << "*** B CALCULATION ***" << endl;
  // calculate the curl of Eth
  grid->curlN2C(tempXC, tempYC, tempZC, Exth, Eyth, Ezth);
  // update the magnetic field
  addscale(-c * dt, 1, Bxc, tempXC, nxc, nyc, nzc);
  addscale(-c * dt, 1, Byc, tempYC, nxc, nyc, nzc);
  addscale(-c * dt, 1, Bzc, tempZC, nxc, nyc, nzc);
  // communicate ghost 
  communicateCenterBC(nxc, nyc, nzc, Bxc, 2, 2, 2, 2, 2, 2, vct);
  communicateCenterBC(nxc, nyc, nzc, Byc, 1, 1, 1, 1, 1, 1, vct);
  communicateCenterBC(nxc, nyc, nzc, Bzc, 2, 2, 2, 2, 2, 2, vct);
  // fixBforcefree(grid,vct);
  //fixBgem(grid, vct);
  // interpolate C2N
  grid->interpC2N(Bxn, Bxc);
  grid->interpC2N(Byn, Byc);
  grid->interpC2N(Bzn, Bzc);


  communicateNodeBC(nxn, nyn, nzn, Bxn, 1, 1, 2, 2, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Byn, 1, 1, 1, 1, 1, 1, vct);
  communicateNodeBC(nxn, nyn, nzn, Bzn, 1, 1, 2, 2, 1, 1, vct);


}
/** initialize EM field with transverse electric waves 1D and rotate
  anticlockwise (theta degrees)*/
inline void EMfields3D::initEM_rotate(VirtualTopology3D * vct, Grid * grid, double B, double theta) {
  // initialize E and rhos on nodes
  for (int i = 0; i < nxn; i++)
    for (int j = 0; j < nyn; j++) {
      Ex[i][j][0] = 0.0;
      Ey[i][j][0] = 0.0;
      Ez[i][j][0] = 0.0;
      Bxn[i][j][0] = B * cos(theta * M_PI / 180);
      Byn[i][j][0] = B * sin(theta * M_PI / 180);
      Bzn[i][j][0] = 0.0;
      rhons[0][i][j][0] = 0.07957747154595; // electrons: species is now first index
      rhons[1][i][j][0] = 0.07957747154595; // protons: species is now first index
    }
  // initialize B on centers
  grid->interpN2C(Bxc, Bxn);
  grid->interpN2C(Byc, Byn);
  grid->interpN2C(Bzc, Bzn);


  for (int is = 0; is < ns; is++)
    grid->interpN2C(rhocs, is, rhons);

}
/**Add a periodic perturbation in rho exp i(kx - \omega t); deltaBoB is the ratio (Delta B / B0) **/
inline void EMfields3D::AddPerturbationRho(double deltaBoB, double kx, double ky, double Bx_mod, double By_mod, double Bz_mod, double ne_mod, double ne_phase, double ni_mod, double ni_phase, double B0, Grid * grid) {

  double alpha;
  alpha = deltaBoB * B0 / sqrt(Bx_mod * Bx_mod + By_mod * By_mod + Bz_mod * Bz_mod);

  ne_mod *= alpha;
  ni_mod *= alpha;
  // cout<<" ne="<<ne_mod<<" ni="<<ni_mod<<" alpha="<<alpha<<endl;
  for (int i = 0; i < nxn; i++)
    for (int j = 0; j < nyn; j++) {
      rhons[0][i][j][0] += ne_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + ne_phase);
      rhons[1][i][j][0] += ni_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + ni_phase);
    }

  for (int is = 0; is < ns; is++)
    grid->interpN2C(rhocs, is, rhons);
}


/**Add a periodic perturbation exp i(kx - \omega t); deltaBoB is the ratio (Delta B / B0) **/
inline void EMfields3D::AddPerturbation(double deltaBoB, double kx, double ky, double Ex_mod, double Ex_phase, double Ey_mod, double Ey_phase, double Ez_mod, double Ez_phase, double Bx_mod, double Bx_phase, double By_mod, double By_phase, double Bz_mod, double Bz_phase, double B0, Grid * grid) {

  double alpha;

  alpha = deltaBoB * B0 / sqrt(Bx_mod * Bx_mod + By_mod * By_mod + Bz_mod * Bz_mod);

  Ex_mod *= alpha;
  Ey_mod *= alpha;
  Ez_mod *= alpha;
  Bx_mod *= alpha;
  By_mod *= alpha;
  Bz_mod *= alpha;

  for (int i = 0; i < nxn; i++)
    for (int j = 0; j < nyn; j++) {
      Ex[i][j][0] += Ex_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + Ex_phase);
      Ey[i][j][0] += Ey_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + Ey_phase);
      Ez[i][j][0] += Ez_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + Ez_phase);
      Bxn[i][j][0] += Bx_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + Bx_phase);
      Byn[i][j][0] += By_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + By_phase);
      Bzn[i][j][0] += Bz_mod * cos(kx * grid->getXN(i, j, 0) + ky * grid->getYN(i, j, 0) + Bz_phase);

    }

  // initialize B on centers
  grid->interpN2C(Bxc, Bxn);
  grid->interpN2C(Byc, Byn);
  grid->interpN2C(Bzc, Bzn);


}


/** Calculate hat rho hat, Jx hat, Jy hat, Jz hat */
inline void EMfields3D::calculateHatFunctions(Grid * grid, VirtualTopology3D * vct) {
  // smoothing
  smooth(Smooth, rhoc, 0, grid, vct);
  // calculate j hat

  for (int is = 0; is < ns; is++) {
    grid->divSymmTensorN2C(tempXC, tempYC, tempZC, pXXsn, pXYsn, pXZsn, pYYsn, pYZsn, pZZsn, is);

    scale(tempXC, -dt / 2.0, nxc, nyc, nzc);
    scale(tempYC, -dt / 2.0, nxc, nyc, nzc);
    scale(tempZC, -dt / 2.0, nxc, nyc, nzc);
    // communicate before interpolating
    communicateCenterBC_P(nxc, nyc, nzc, tempXC, 2, 2, 2, 2, 2, 2, vct);
    communicateCenterBC_P(nxc, nyc, nzc, tempYC, 2, 2, 2, 2, 2, 2, vct);
    communicateCenterBC_P(nxc, nyc, nzc, tempZC, 2, 2, 2, 2, 2, 2, vct);

    grid->interpC2N(tempXN, tempXC);
    grid->interpC2N(tempYN, tempYC);
    grid->interpC2N(tempZN, tempZC);
    sum(tempXN, Jxs, nxn, nyn, nzn, is);
    sum(tempYN, Jys, nxn, nyn, nzn, is);
    sum(tempZN, Jzs, nxn, nyn, nzn, is);
    // PIDOT
    PIdot(Jxh, Jyh, Jzh, tempXN, tempYN, tempZN, is, grid);

  }
  // smooth j
  smooth(Smooth, Jxh, 1, grid, vct);
  smooth(Smooth, Jyh, 1, grid, vct);
  smooth(Smooth, Jzh, 1, grid, vct);

  // calculate rho hat = rho - (dt*theta)div(jhat)
  grid->divN2C(tempXC, Jxh, Jyh, Jzh);
  scale(tempXC, -dt * th, nxc, nyc, nzc);
  sum(tempXC, rhoc, nxc, nyc, nzc);
  eq(rhoh, tempXC, nxc, nyc, nzc);
  // communicate rhoh
  communicateCenterBC_P(nxc, nyc, nzc, rhoh, 2, 2, 2, 2, 2, 2, vct);
}
/** Image of Poisson Solver */
inline void EMfields3D::PoissonImage(double *image, double *vector, Grid * grid, VirtualTopology3D * vct) {
  // allocate 2 three dimensional service vectors
  double ***temp = newArr3(double, nxc, nyc, nzc);
  double ***im = newArr3(double, nxc, nyc, nzc);
  eqValue(0.0, image, (nxc - 2) * (nyc - 2) * (nzc - 2));
  eqValue(0.0, temp, nxc, nyc, nzc);
  eqValue(0.0, im, nxc, nyc, nzc);
  // move from krylov space to physical space and communicate ghost cells
  solver2phys(temp, vector, nxc, nyc, nzc);
  // calculate the laplacian
  grid->lapC2Cpoisson(im, temp, vct);
  // move from physical space to krylov space
  phys2solver(image, im, nxc, nyc, nzc);
  // deallocate temporary array and objects
  delArr3(temp, nxc, nyc);
  delArr3(im, nxc, nyc);
}
/** interpolate charge density and pressure density from node to center */
inline void EMfields3D::interpDensitiesN2C(VirtualTopology3D * vct, Grid * grid) {
  // do we need communication or not really?
  grid->interpN2C(rhoc, rhon);
}
/** communicate ghost for grid -> Particles interpolation */
inline void EMfields3D::communicateGhostP2G(int ns, int bcFaceXright, int bcFaceXleft, int bcFaceYright, int bcFaceYleft, VirtualTopology3D * vct) {
  // interpolate adding common nodes among processors
  communicateInterp(nxn, nyn, nzn, ns, rhons, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, Jxs, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, Jys, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, Jzs, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pXXsn, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pXYsn, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pXZsn, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pYYsn, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pYZsn, 0, 0, 0, 0, 0, 0, vct);
  communicateInterp(nxn, nyn, nzn, ns, pZZsn, 0, 0, 0, 0, 0, 0, vct);
  // calculate the correct densities on the boundaries
  adjustNonPeriodicDensities(ns, vct);
  // put the correct values on ghost cells

  communicateNode_P(nxn, nyn, nzn, rhons, ns, vct);
  communicateNode_P(nxn, nyn, nzn, Jxs, ns, vct);
  communicateNode_P(nxn, nyn, nzn, Jys, ns, vct);
  communicateNode_P(nxn, nyn, nzn, Jzs, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pXXsn, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pXYsn, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pXZsn, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pYYsn, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pYZsn, ns, vct);
  communicateNode_P(nxn, nyn, nzn, pZZsn, ns, vct);

}


/** add an amount of charge density to charge density field at node X,Y */
inline void EMfields3D::addRho(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        rhons[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of charge density to current density - direction X to current density field on the node*/
inline void EMfields3D::addJx(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        Jxs[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of current density - direction Y to current density field on the node */
inline void EMfields3D::addJy(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        Jys[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of current density - direction Z to current density field on the node */
inline void EMfields3D::addJz(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        Jzs[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction XX to current density field on the node */
inline void EMfields3D::addPxx(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        pXXsn[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction XY to current density field on the node*/
inline void EMfields3D::addPxy(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        pXYsn[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction XZ to current density field on the node */
inline void EMfields3D::addPxz(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        pXZsn[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction YY to current density field on the node*/
inline void EMfields3D::addPyy(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        pYYsn[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction YZ to current density field on the node */
inline void EMfields3D::addPyz(double weight[][2][2], int X, int Y, int Z, int is) {
  for (int i = 0; i < 2; i++)
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
        pYZsn[is][X - i][Y - j][Z - k] += weight[i][j][k] * invVOL;
}
/** add an amount of pressure density - direction ZZ to current density field on the node */
inline  void EMfields3D::addPzz(double weight[][2][2], int X, int Y, int Z, int is){
	for (int i=0; i < 2; i++)
      for (int j=0; j < 2; j++)
        for(int k=0; k < 2; k++)
            pZZsn[is][X -i][Y -j][Z - k]        += weight[i][j][k]*invVOL;
}



/** set to 0 all the densities fields */
inline  void EMfields3D::setZeroDensities(){
  for (register int i=0; i < nxn; i++)
	for (register int j=0; j < nyn; j++)
	          for (register int k=0; k < nzn; k++){
	    Jx[i][j][k]   = 0.0;
	    Jxh[i][j][k]  = 0.0;
	    Jy[i][j][k]   = 0.0;
	    Jyh[i][j][k]  = 0.0;
	    Jz[i][j][k]   = 0.0;
	    Jzh[i][j][k]  = 0.0;
		rhon[i][j][k] = 0.0;
	  }
for (register int i=0; i < nxc; i++)
	for (register int j=0; j < nyc; j++)
	 for (register int k=0; k < nzc; k++){
	   rhoc[i][j][k] = 0.0;
	   rhoh[i][j][k] = 0.0;
	}
for (register int kk=0; kk < ns; kk++)
 for (register int i=0; i < nxn; i++)
	for (register int j=0; j < nyn; j++)
	     for (register int k=0; k < nzn; k++){
		  rhons[kk][i][j][k] = 0.0;
		  Jxs[kk][i][j][k]   = 0.0;
		  Jys[kk][i][j][k]   = 0.0;
		  Jzs[kk][i][j][k]   = 0.0;
		  pXXsn[kk][i][j][k] = 0.0;
		  pXYsn[kk][i][j][k] = 0.0;
		  pXZsn[kk][i][j][k] = 0.0;
		  pYYsn[kk][i][j][k] = 0.0;
		  pYZsn[kk][i][j][k] = 0.0;
		  pZZsn[kk][i][j][k] = 0.0;
	  }  
       
}
/**SPECIES: Sum the charge density of different species on NODES*/
inline void EMfields3D::sumOverSpecies(VirtualTopology3D *vct){
	for (int is=0; is<ns; is++)
 	   for (register int i=0; i <nxn; i++)
  	      for (register int j=0; j <nyn; j++)
		    for (register int k=0; k <nzn; k++)
	           rhon[i][j][k]     +=rhons[is][i][j][k];
}

/**SPECIES: Sum current density for different species */
inline void EMfields3D::sumOverSpeciesJ(){
	for (int is=0; is < ns; is++)
		for (register int i=0; i <nxn; i++)
			for (register int j=0; j <nyn; j++)
			  for (register int k=0; k <nzn; k++) {
				Jx[i][j][k]     +=Jxs[is][i][j][k];
				Jy[i][j][k]     +=Jys[is][i][j][k];
				Jz[i][j][k]     +=Jzs[is][i][j][k];
			  }
}



/** initialize Magnetic and Electric Field with initial configuration */
inline void EMfields3D::init(VirtualTopology3D *vct, Grid *grid){
 if (restart1 ==0){
   for (int i=0; i < nxn; i++){
    for (int j=0; j < nyn; j++){
      for (int k=0; k < nzn; k++){
        for (int is=0; is < ns; is++){
		  rhons[is][i][j][k] = rhoINIT[is]/FourPI;
		}
		Ex[i][j][k] = 0.0;
        Ey[i][j][k] = 0.0;
        Ez[i][j][k] = 0.0;
        Bxn[i][j][k] = B0x;
        Byn[i][j][k] = B0y;
        Bzn[i][j][k] = B0z;
      }
    }
   }
   // initialize B on centers
   grid->interpN2C(Bxc,Bxn);
	grid->interpN2C(Byc,Byn);
	grid->interpN2C(Bzc,Bzn);
	for (int is=0 ; is<ns; is++)
		 grid->interpN2C(rhocs,is,rhons);
	} else {  // READING FROM RESTART
        cout << "RESTART not available" << endl;
  }
}

/**  initiliaze EM for GEM challange */
inline void EMfields3D::initGEM(VirtualTopology3D *vct, Grid *grid){
	// perturbation localized in X
	double pertX = 0.4;
	double xpert, ypert, exp_pert;
	if (restart1 ==0){
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "------------------------------------------" << endl;
			cout << "Initialize GEM Challenge with Pertubation" << endl; 
			cout << "------------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			cout << "Delta (current sheet thickness) = " << delta << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
				if (DriftSpecies[i])
					cout << " DRIFTING " << endl;
				else
					cout << " BACKGROUND " << endl;
			}
			cout << "-------------------------" << endl;
		}
		for (int i=0; i < nxn; i++)
			for (int j=0; j < nyn; j++)
				for (int k=0; k < nzn; k++){
					// initialize the density for species
					for (int is=0; is < ns; is++){
						if (DriftSpecies[is])
							rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-Ly/2)/delta)*cosh((grid->getYN(i,j,k)-Ly/2)/delta))))/FourPI;
						else
							rhons[is][i][j][k] = rhoINIT[is]/FourPI;
					}
					// electric field
					Ex[i][j][k] =  0.0;
					Ey[i][j][k] =  0.0;
					Ez[i][j][k] =  0.0;
					// Magnetic field
					Bxn[i][j][k] = B0x*tanh((grid->getYN(i,j,k) - Ly/2)/delta);	
					// add the initial GEM perturbation
					// Bxn[i][j][k] += (B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXN(i,j,k)/Lx)*sin(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly  );
					Byn[i][j][k] = B0y; // - (B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXN(i,j,k)/Lx)*cos(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly); 
					// add the initial X perturbation
					xpert = grid->getXN(i,j,k)- Lx/2;
					ypert = grid->getYN(i,j,k)- Ly/2;
					exp_pert = exp(-(xpert/delta)*(xpert/delta)-(ypert/delta)*(ypert/delta));
					Bxn[i][j][k] +=(B0x*pertX)*exp_pert*(-cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*ypert/delta
														 -cos(M_PI*xpert/10.0/delta)*sin(M_PI*ypert/10.0/delta)*M_PI/10.0);
					Byn[i][j][k] +=(B0x*pertX)*exp_pert*(cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*xpert/delta
														 +sin(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*M_PI/10.0);
					// guide field
					Bzn[i][j][k] = B0z;
				}
					// initialize B on centers
					for (int i=0; i < nxc; i++)
						for (int j=0; j < nyc; j++)
							for (int k=0; k < nzc; k++){
								// Magnetic field
								Bxc[i][j][k] = B0x*tanh((grid->getYC(i,j,k) - Ly/2)/delta);	
								// add the initial GEM perturbation
								//Bxc[i][j][k] += (B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXC(i,j,k)/Lx)*sin(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly  );
								Byc[i][j][k] = B0y; //  - (B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXC(i,j,k)/Lx)*cos(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly); 
								// add the initial X perturbation
								xpert = grid->getXC(i,j,k)- Lx/2;
								ypert = grid->getYC(i,j,k)- Ly/2;
								exp_pert = exp(-(xpert/delta)*(xpert/delta)-(ypert/delta)*(ypert/delta));
								Bxc[i][j][k] +=(B0x*pertX)*exp_pert*(-cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*ypert/delta
																	 -cos(M_PI*xpert/10.0/delta)*sin(M_PI*ypert/10.0/delta)*M_PI/10.0);
								Byc[i][j][k] +=(B0x*pertX)*exp_pert*(cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*xpert/delta
																	 +sin(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*M_PI/10.0);
								// guide field
								Bzc[i][j][k] = B0z;
								
							}
								for (int is=0 ; is<ns; is++)
									grid->interpN2C(rhocs,is,rhons);
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}

/** initialize GEM challenge with no Perturbation */
inline void EMfields3D::initGEMnoPert(VirtualTopology3D * vct, Grid * grid) {
  if (restart1 == 0) {

    // initialize
    if (vct->getCartesian_rank() == 0) {
      cout << "----------------------------------------------" << endl;
      cout << "Initialize GEM Challenge without Perturbation" << endl;
      cout << "----------------------------------------------" << endl;
      cout << "B0x                              = " << B0x << endl;
      cout << "B0y                              = " << B0y << endl;
      cout << "B0z                              = " << B0z << endl;
      cout << "Delta (current sheet thickness) = " << delta << endl;
      for (int i = 0; i < ns; i++) {
        cout << "rho species " << i << " = " << rhoINIT[i];
        if (DriftSpecies[i])
          cout << " DRIFTING " << endl;
        else
          cout << " BACKGROUND " << endl;
      }
      cout << "-------------------------" << endl;
    }
    for (int i = 0; i < nxn; i++)
      for (int j = 0; j < nyn; j++)
        for (int k = 0; k < nzn; k++) {
          // initialize the density for species
          for (int is = 0; is < ns; is++) {
            if (DriftSpecies[is])
              rhons[is][i][j][k] = ((rhoINIT[is] / (cosh((grid->getYN(i, j, k) - Ly / 2) / delta) * cosh((grid->getYN(i, j, k) - Ly / 2) / delta)))) / FourPI;
            else
              rhons[is][i][j][k] = rhoINIT[is] / FourPI;
          }
          // electric field
          Ex[i][j][k] = 0.0;
          Ey[i][j][k] = 0.0;
          Ez[i][j][k] = 0.0;
          // Magnetic field
          Bxn[i][j][k] = B0x * tanh((grid->getYN(i, j, k) - Ly / 2) / delta);
          Byn[i][j][k] = B0y;
          // guide field
          Bzn[i][j][k] = B0z;
        }
    // initialize B on centers
    for (int i = 0; i < nxc; i++)
      for (int j = 0; j < nyc; j++)
        for (int k = 0; k < nzc; k++) {
          // Magnetic field
          Bxc[i][j][k] = B0x * tanh((grid->getYC(i, j, k) - Ly / 2) / delta);
          Byc[i][j][k] = B0y;
          // guide field
          Bzc[i][j][k] = B0z;

        }
    for (int is = 0; is < ns; is++)
      grid->interpN2C(rhocs, is, rhons);
  }
  else {
    init(vct, grid);            // use the fields from restart file
  }
}



/**  initiliaze EM for GEM challange */
inline void EMfields3D::initGEMdoublePeriodicMHD(VirtualTopology3D *vct, Grid *grid){

    
//	if (restart1 ==0){
//		// initialize
//		if (vct->getCartesian_rank() ==0){
//			cout << "------------------------------------------" << endl;
//			cout << "Initialize GEM Challenge with Pertubation" << endl;
//			cout << "------------------------------------------" << endl;
//			cout << "B0x                              = " << B0x << endl;
//			cout << "B0y                              = " << B0y << endl;
//			cout << "B0z                              = " << B0z << endl;
//			cout << "Delta (current sheet thickness) = " << delta << endl;
//			for (int i=0; i < ns; i++){
//				cout << "rho species " << i <<" = " << rhoINIT[i];
//				if (DriftSpecies[i])
//					cout << " DRIFTING " << endl;
//				else
//					cout << " BACKGROUND " << endl;
//			}
//			cout << "-------------------------" << endl;
//		}
//		for (int i=0; i < nxn; i++)
//			for (int j=0; j < nyn; j++)
//				for (int k=0; k < nzn; k++){
//					// initialize the density for species
//					for (int is=0; is < ns; is++){
//                        //						if (DriftSpecies[is])
//                        //							rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-Ly/2)/delta)*cosh((grid->getYN(i,j,k)-Ly/2)/delta))))/FourPI;
//                        //						else
//                        //							rhons[is][i][j][k] = rhoINIT[is]/FourPI;
//                        
//                        // for a regual asymmetric CS
//                        if ((is==0) || (is==1))
//                            rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-Ly/4.0)/delta)*cosh((grid->getYN(i,j,k)-Ly/4.0)/delta))))/FourPI;
//                        else if ((is==2) || (is==3))
//                            rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-3.0*Ly/4.0)/delta)*cosh((grid->getYN(i,j,k)-3.0*Ly/4.0)/delta))))/FourPI;
//                        else
//                            rhons[is][i][j][k] = (rhoINIT[is] - (R_parameter*R_parameter)-(2.0*R_parameter*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))) )/FourPI;
//                        
//                        // asymmetric CS + oxygen
//                        //                        if ((is==0) || (is==1))
//                        //                            rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-Ly/4.0)/delta)*cosh((grid->getYN(i,j,k)-Ly/4.0)/delta))))/FourPI;
//                        //                        else if ((is==2) || (is==3))
//                        //                            rhons[is][i][j][k] = ((rhoINIT[is]/(cosh((grid->getYN(i,j,k)-3.0*Ly/4.0)/delta)*cosh((grid->getYN(i,j,k)-3.0*Ly/4.0)/delta))))/FourPI;
//                        //                        else if ((is==4) || (is==5))
//                        //                            rhons[is][i][j][k] = 0.5*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta) - tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))*(rhoINIT[is] - (R_parameter*R_parameter)-(2.0*R_parameter*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))) )/FourPI;
//                        //			            else
//                        //                            rhons[is][i][j][k] = (1.0-(0.5*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta) - tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))))*(rhoINIT[is] - (R_parameter*R_parameter)-(2.0*R_parameter*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta)  - 1.0   -  tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))) )/FourPI;
//                        
//					}
//					// electric field
//					Ex[i][j][k] =  0.0;
//					Ey[i][j][k] =  0.0;
//					Ez[i][j][k] =  0.0;
//					// Magnetic field
//					Bxn[i][j][k] = (B0x*R_parameter) + B0x*1.0*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta));
//					// add the initial GEM perturbation
//					// Bxn[i][j][k] += (B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXN(i,j,k)/Lx)*sin(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly  );
//					Byn[i][j][k] = B0y; // - (B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXN(i,j,k)/Lx)*cos(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly);
//					// add the initial X perturbation
//					xpert = grid->getXN(i,j,k)- Lx/2.0;
//					ypert = grid->getYN(i,j,k)- Ly/4.0;
//					exp_pert = exp(-(xpert/delta)*(xpert/delta)-(ypert/delta)*(ypert/delta));
//					Bxn[i][j][k] +=(B0x*pertX)*exp_pert;// *(-cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*ypert/delta
//                    // -cos(M_PI*xpert/10.0/delta)*sin(M_PI*ypert/10.0/delta)*M_PI/10.0);
//					Byn[i][j][k] +=(B0x*pertX)*exp_pert;// *(cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*xpert/delta
//                    // +sin(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*M_PI/10.0);
//					// guide field
//					Bzn[i][j][k] = B0z;
//				}
//        // initialize B on centers
//        for (int i=0; i < nxc; i++)
//            for (int j=0; j < nyc; j++)
//                for (int k=0; k < nzc; k++){
//                    // Magnetic field
//                    Bxc[i][j][k] = (B0x*R_parameter) + B0x*1.0*(tanh((grid->getYC(i,j,k) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYC(i,j,k) - (3.0*Ly/4.0))/delta));
//                    // add the initial GEM perturbation
//                    //Bxc[i][j][k] += (B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXC(i,j,k)/Lx)*sin(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly  );
//                    Byc[i][j][k] = B0y; //  - (B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXC(i,j,k)/Lx)*cos(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly);
//                    // add the initial X perturbation
//                    xpert = grid->getXC(i,j,k)- Lx/2.0;
//                    ypert = grid->getYC(i,j,k)- Ly/4.0;
//                    exp_pert = exp(-(xpert/delta)*(xpert/delta)-(ypert/delta)*(ypert/delta));
//                    Bxc[i][j][k] +=(B0x*pertX)*exp_pert;// *(-cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*ypert/delta
//                    // -cos(M_PI*xpert/10.0/delta)*sin(M_PI*ypert/10.0/delta)*M_PI/10.0);
//                    Byc[i][j][k] +=(B0x*pertX)*exp_pert;// *(cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*xpert/delta
//                    // +sin(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*M_PI/10.0);
//                    // guide field
//                    Bzc[i][j][k] = B0z;
//                    
//                }
//        for (int is=0 ; is<ns; is++)
//            grid->interpN2C(rhocs,is,rhons);
//	} else {
//		init(vct,grid);  // use the fields from restart file
//	}

    
    
    
    
    
    
    
    
    
    
    
    
// Erase everything above.
    
    
    
    
	// perturbation localized in X
	double pertX ; //was 0.4
	double xpert, ypert, exp_pert;
    double R_parameter = 0.0;


    if (Case=="SWEET-PARKER")
        pertX=0.4;
    else pertX=0.0;
    
    int x, y, z;
    int order=0;
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------------------" << endl;
			cout << "Initialize Fields and density for two current sheets" << endl;
            cout << "Perturbation only on the top current sheet" << endl;
			cout << "----------------------------------------------------" << endl;
            cout << "Case: " << Case << endl;
			cout << "B0x: " << B0x << endl;
			cout << "B0y: " << B0y << endl;
			cout << "B0z: " << B0z << endl;
            cout << "----------------------------------------------------" << endl;
		//	for (int i=0; i < ns; i++){
		//		cout << "rho species " << i <<" = " << rhoINIT[i];
		//	}
        }
        
         
		for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
                for (int jj=0; jj <nzc; jj++){
                    Bxc[i][j][jj] = (B0x*R_parameter) + B0x*1.0*(tanh((grid->getYC(i,j,jj) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYC(i,j,jj) - (3.0*Ly/4.0))/delta));
                    Byc[i][j][jj] = B0y;
                    // guide field
                    Bzc[i][j][jj] = B0z;


                    //  A simple uniform density initial condition.
                    //  rhoc[i][j][jj] =  rhoINIT[0];
                    
                    rhoc[i][j][jj] =  rhoINIT[1] + ((rhoINIT[0]/(cosh((grid->getYN(i,j,jj)-Ly/4.0)/delta)*cosh((grid->getYN(i,j,jj)-Ly/4.0)/delta))))+((rhoINIT[0]/(cosh((grid->getYN(i,j,jj)-3.0*Ly/4.0)/delta)*cosh((grid->getYN(i,j,jj)-3.0*Ly/4.0)/delta))));

                    
                    Uc[i][j][jj]=(((1.0  -    ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5     )  )/(C_g-1)) + ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5;


                    xpert = grid->getXC(i,j,jj)- Lx/2.0;
                    ypert = grid->getYC(i,j,jj)- 3.0*Ly/4.0;
                    exp_pert = 1.0;//exp(-(xpert/delta)*(xpert/delta)-(ypert/delta)*(ypert/delta));


//                    Bxc[i][j][jj] +=(B0x*pertX)*exp_pert*(-cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*ypert/delta
//                                                          -cos(M_PI*xpert/10.0/delta)*sin(M_PI*ypert/10.0/delta)*M_PI/10.0);
//                    Byc[i][j][jj] +=(B0x*pertX)*exp_pert*(cos(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*2.0*xpert/delta
//                                                         +sin(M_PI*xpert/10.0/delta)*cos(M_PI*ypert/10.0/delta)*M_PI/10.0);
  
// ! Attention. Need GEM-like perturbation to start the Sweet-Parker like simulation.

                    Bxc[i][j][jj] +=(B0x*pertX)*((M_PI*pertX*cos((2.0*M_PI*xpert)/Lx)*sin((M_PI*ypert)/Ly))/Ly);//*0.5*(sgn_MHD(ypert+(Ly/4.0))  + sgn_MHD((Ly/4.0)-ypert)    );
                    Byc[i][j][jj] +=(B0x*pertX)*(-(2.0*M_PI*pertX*cos((M_PI*ypert)/Ly)*sin((2.0*M_PI*xpert)/Lx))/Lx);//*0.5*(sgn_MHD(ypert+(Ly/4.0))  + sgn_MHD((Ly/4.0)-ypert));


                    // Need the next line for asymmetric configurations (to compensate for the magnetic field pressure).
                    //                        rhons[is][i][j][k] = (rhoINIT[is] - (R_parameter*R_parameter)-(2.0*R_parameter*(tanh((grid->getYN(i,j,k) - (Ly/4.0))/delta)  - 1.0   -    tanh((grid->getYN(i,j,k) - (3.0*Ly/4.0))/delta))) )/FourPI;
        
     
                    
                    pxc[i][j][jj] = 0.0;
                    pyc[i][j][jj] = 0.0;
                    pzc[i][j][jj] = 0.0;
                    
                    // Assume initial temperature at t=0
                    // Uc[i][j][jj]=(((rhoc[i][j][jj]))/(C_g-1)) + ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5;
                }
        
        
        Boundarize(nxc,nyc,nzc,rhoc,pxc,pyc,pzc,Bxc,Byc,Bzc,Uc,vct);
        
        
        
        eq(BxStored,Bxc,nxc,nyc,nzc);
        eq(ByStored,Byc,nxc,nyc,nzc);
        eq(BzStored,Bzc,nxc,nyc,nzc);
        eq(pxStored,pxc,nxc,nyc,nzc);
        eq(pyStored,pyc,nxc,nyc,nzc);
        eq(pzStored,pzc,nxc,nyc,nzc);
        eq(UStored,Uc,nxc,nyc,nzc);
        eq(roStored,rhoc,nxc,nyc,nzc);
        
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        //cout<<"Test communication start "<<endl;
        
        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
        
        //cout<<"Test communication done "<<endl;
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        eq(Bxc0,Bxc,nxc,nyc,nzc);
        eq(Byc0,Byc,nxc,nyc,nzc);
        eq(Bzc0,Bzc,nxc,nyc,nzc);
        eq(pxc0,pxc,nxc,nyc,nzc);
        eq(pyc0,pyc,nxc,nyc,nzc);
        eq(pzc0,pzc,nxc,nyc,nzc);
        eq(Uc0,Uc,nxc,nyc,nzc);
        eq(rhoc0,rhoc,nxc,nyc,nzc);
        
        eqValue(0.0, Q__rhoc, nxc, nyc, nzc);
        eqValue(0.0, Q__pxc, nxc, nyc, nzc);
        eqValue(0.0, Q__pyc, nxc, nyc, nzc);
        eqValue(0.0, Q__pzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bxc, nxc, nyc, nzc);
        eqValue(0.0, Q__Byc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Uc, nxc, nyc, nzc);
        
        
	} else {
		init(vct,grid);  // use the fields from restart file
	}

    
        
    
}








/**
*
*  Init Force Free (JxB=0)
*/
inline void EMfields3D::initForceFree(VirtualTopology3D *vct, Grid *grid){
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "Initialize Force Free with Perturbation" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			cout << "Delta (current sheet thickness) = " << delta << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
		}
		for (int i=0; i < nxn; i++)
		for (int j=0; j < nyn; j++)
		 for (int k=0; k < nzn; k++){
		   // initialize the density for species
		   for (int is=0; is < ns; is++){
						rhons[is][i][j][k] = rhoINIT[is]/FourPI;
				}
				// electric field
				Ex[i][j][k] =  0.0;
				Ey[i][j][k] =  0.0;
				Ez[i][j][k] =  0.0;
				// Magnetic field
				Bxn[i][j][k] = B0x*tanh((grid->getYN(i,j,k) - Ly/2)/delta);
				// add the initial GEM perturbation
				Bxn[i][j][k] +=(B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXN(i,j,k)/Lx)*sin(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly  );
				Byn[i][j][k] = B0y -(B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXN(i,j,k)/Lx)*cos(M_PI*(grid->getYN(i,j,k)- Ly/2)/Ly);
				// guide field
				Bzn[i][j][k] = B0z/cosh((grid->getYN(i,j,k) - Ly/2)/delta);
			}
		   for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
			  for (int k=0; k <nzc; k++) {
				Bxc[i][j][k] = B0x*tanh((grid->getYC(i,j,k) - Ly/2)/delta);
				// add the perturbation
				Bxc[i][j][k] +=(B0x/10.0)*(M_PI/Ly)*cos(2*M_PI*grid->getXC(i,j,k)/Lx)*sin(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly  );
				Byc[i][j][k] = B0y -(B0x/10.0)*(2*M_PI/Lx)*sin(2*M_PI*grid->getXC(i,j,k)/Lx)*cos(M_PI*(grid->getYC(i,j,k)- Ly/2)/Ly);
				// guide field
				Bzc[i][j][k] = B0z/cosh((grid->getYC(i,j,k) - Ly/2)/delta);
			}
	
		for (int is=0 ; is<ns; is++)
			grid->interpN2C(rhocs,is,rhons);
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}
/**
*
*  Init 3D null point(s) following Galsgaard & Nordlund (1997), JGR 102, p. 231
*
*  (c) V. Olshevsky
*/
inline void EMfields3D::initNullGalsgaard(VirtualTopology3D *vct, Grid *grid){
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "       Initialize 3D null point(s) Galsgaard & Nordlund variation" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
		}		              
        
        for (int i=0; i < nxn; i++)
		for (int j=0; j < nyn; j++)
		 for (int k=0; k < nzn; k++){
		   // initialize the density for species
		   for (int is=0; is < ns; is++){
					rhons[is][i][j][k] = rhoINIT[is]/FourPI;                    
				}
				// electric field
				Ex[i][j][k] =  0.0;
				Ey[i][j][k] =  0.0;
				Ez[i][j][k] =  0.0;
				// Magnetic field
				Bxn[i][j][k] = B0x*cos(2.*M_PI*grid->getYN(i,j,k)/Ly) - sin(2.*M_PI*grid->getZN(i,j,k)/Lz);
				Byn[i][j][k] = B0x*cos(2.*M_PI*grid->getZN(i,j,k)/Lz) - sin(2.*M_PI*grid->getXN(i,j,k)/Lx);
				Bzn[i][j][k] = B0x*cos(2.*M_PI*grid->getXN(i,j,k)/Lx) - sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
			}
		   for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
			  for (int k=0; k <nzc; k++) {
				Bxc[i][j][k] = B0x*(cos(2.*M_PI*grid->getYC(i,j,k)/Ly) - sin(2.*M_PI*grid->getZC(i,j,k)/Lz));
				Byc[i][j][k] = B0x*(cos(2.*M_PI*grid->getZC(i,j,k)/Lz) - sin(2.*M_PI*grid->getXC(i,j,k)/Lx));
				Bzc[i][j][k] = B0x*(cos(2.*M_PI*grid->getXC(i,j,k)/Lx) - sin(2.*M_PI*grid->getYC(i,j,k)/Ly));
			}
		for (int is=0 ; is<ns; is++)
			grid->interpN2C(rhocs,is,rhons);
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}
/**
*
*  Init 3D null point(s) 
*
*  (c) V. Olshevsky
*/
inline void EMfields3D::initNullPoint(VirtualTopology3D *vct, Grid *grid){
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "       Initialize 3D null point(s)" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
		}		              
        
        for (int i=0; i < nxn; i++)
		for (int j=0; j < nyn; j++)
		 for (int k=0; k < nzn; k++){
		   // initialize the density for species
		   for (int is=0; is < ns; is++){
					rhons[is][i][j][k] = rhoINIT[is]/FourPI;
                    //rhons[is][i][j][k] = 1./FourPI*(0.0004 - 0.5*(Bxn[i][j][k]*Bxn[i][j][k] + Byn[i][j][k]*Byn[i][j][k] + Bzn[i][j][k]*Bzn[i][j][k]))/((uth[0]*uth[0])/abs(qom[0]) + (uth[1]*uth[1]));
				}
				// electric field
				Ex[i][j][k] =  0.0;
				Ey[i][j][k] =  0.0;
				Ez[i][j][k] =  0.0;
				// Magnetic field
				Bxn[i][j][k] = -B0x*cos(2.*M_PI*grid->getXN(i,j,k)/Lx)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
				Byn[i][j][k] = B0x*cos(2.*M_PI*grid->getYN(i,j,k)/Ly)*(-2.*sin(2.*M_PI*grid->getZN(i,j,k)/Lz) + sin(2.*M_PI*grid->getXN(i,j,k)/Lx));
				Bzn[i][j][k] = 2.*B0x*cos(2.*M_PI*grid->getZN(i,j,k)/Lz)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
			}
		   for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
			  for (int k=0; k <nzc; k++) {
				Bxc[i][j][k] = -B0x*cos(2.*M_PI*grid->getXC(i,j,k)/Lx)*sin(2.*M_PI*grid->getYC(i,j,k)/Ly);
				Byc[i][j][k] = B0x*cos(2.*M_PI*grid->getYC(i,j,k)/Ly)*(-2.*sin(2.*M_PI*grid->getZC(i,j,k)/Lz) + sin(2.*M_PI*grid->getXC(i,j,k)/Lx));
				Bzc[i][j][k] = 2.*B0x*cos(2.*M_PI*grid->getZC(i,j,k)/Lz)*sin(2.*M_PI*grid->getYC(i,j,k)/Ly);
			}
/*
           // currents are used to calculate in the Maxwell's solver
           // The ion current is equal to 0 (all current is on electrons)
           for (int i=0; i < nxn; i++)
              for (int j=0; j < nyn; j++)
                for (int k=0; k < nzn; k++){
                    Jxs[1][i][j][k] = 0.0; // ion species is species 1
                    Jys[1][i][j][k] = 0.0; // ion species is species 1
                    Jzs[1][i][j][k] = 0.0; // ion species is species 1
                }
             // calculate the electron current from
             eqValue(0.0,tempXN,nxn,nyn,nzn);
             eqValue(0.0,tempYN,nxn,nyn,nzn);
             eqValue(0.0,tempZN,nxn,nyn,nzn);
             grid->curlC2N(tempXN,tempYN,tempZN,Bxc,Byc,Bzc); // here you calculate curl(B)
             // all current is on electrons, calculated from Ampere's law
            for (int i=0; i < nxn; i++)
              for (int j=0; j < nyn; j++)
                for (int k=0; k < nzn; k++){  // electrons are species 0
                    Jxs[0][i][j][k] = c*tempXN[i][j][k]/FourPI; // ion species is species 1
                    Jys[0][i][j][k] = c*tempYN[i][j][k]/FourPI; // ion species is species 1
                    Jzs[0][i][j][k] = c*tempZN[i][j][k]/FourPI; // ion species is species 1
                }
*/	
		for (int is=0 ; is<ns; is++)
			grid->interpN2C(rhocs,is,rhons);
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}


/** Initialize the EM field with constants values or from restart*/
inline void EMfields3D::initBEAM(VirtualTopology3D * vct, Grid * grid, double x_center, double y_center, double z_center, double radius) {
  double distance;
  // initialize E and rhos on nodes
  if (restart1 == 0) {
    for (int i = 0; i < nxn; i++)
      for (int j = 0; j < nyn; j++)
        for (int k = 0; k < nzn; k++) {
          Ex[i][j][k] = 0.0;
          Ey[i][j][k] = 0.0;
          Ez[i][j][k] = 0.0;
          Bxn[i][j][k] = 0.0;
          Byn[i][j][k] = 0.0;
          Bzn[i][j][k] = 0.0;
          distance = (grid->getXN(i, j, k) - x_center) * (grid->getXN(i, j, k) - x_center) / (radius * radius) + (grid->getYN(i, j, k) - y_center) * (grid->getYN(i, j, k) - y_center) / (radius * radius) + (grid->getZN(i, j, k) - z_center) * (grid->getZN(i, j, k) - z_center) / (4 * radius * radius);
          // plasma
          rhons[0][i][j][k] = rhoINIT[0] / FourPI;  // initialize with constant density
          // electrons
          rhons[1][i][j][k] = rhoINIT[1] / FourPI;
          // beam
          if (distance < 1.0)
            rhons[2][i][j][k] = rhoINIT[2] / FourPI;
          else
            rhons[2][i][j][k] = 0.0;
        }
    // initialize B on centers
    for (int i = 0; i < nxc; i++)
      for (int j = 0; j < nyc; j++)
        for (int k = 0; k < nzc; k++) {
          // Magnetic field
          Bxc[i][j][k] = 0.0;
          Byc[i][j][k] = 0.0;
          Bzc[i][j][k] = 0.0;


        }
    for (int is = 0; is < ns; is++)
      grid->interpN2C(rhocs, is, rhons);
  }
  else {                        // EM initialization from RESTART
    init(vct, grid);            // use the fields from restart file
  }

}


/** Calculate the susceptibility on the boundary left*/
inline void EMfields3D::sustensorLeftY(double **susxy, double **susyy, double **suszy) {
  double beta, omcx, omcy, omcz, denom;
  for (int i = 0; i < nxn; i++)
    for (int k = 0; k < nzn; k++) {
      susxy[i][k] = 0.0;
      susyy[i][k] = 1.0;
      suszy[i][k] = 0.0;
    }
  for (int is = 0; is < ns; is++) {
    beta = .5 * qom[is] * dt / c;
    for (int i = 0; i < nxn; i++)
      for (int k = 0; k < nzn; k++) {
        omcx = beta * Bxn[i][1][k];
        omcy = beta * Byn[i][1][k];
        omcz = beta * Bzn[i][1][k];
        denom = FourPI / 2 * delt * dt / c * qom[is] * rhons[is][i][1][k] / (1.0 + omcx * omcx + omcy * omcy + omcz * omcz);
        susxy[i][k] += (omcz + omcx * omcy) * denom;
        susyy[i][k] += (1.0 + omcy * omcy) * denom;
        suszy[i][k] += (omcy * omcz - omcx) * denom;
      }
  }

}
/** Calculate the susceptibility on the boundary right*/
inline void EMfields3D::sustensorRightY(double **susxy, double **susyy, double **suszy) {
  double beta, omcx, omcy, omcz, denom;
  for (int i = 0; i < nxn; i++)
    for (int k = 0; k < nzn; k++) {
      susxy[i][k] = 0.0;
      susyy[i][k] = 1.0;
      suszy[i][k] = 0.0;
    }
  for (int is = 0; is < ns; is++) {
    beta = .5 * qom[is] * dt / c;
    for (int i = 0; i < nxn; i++)
      for (int k = 0; k < nzn; k++) {
        omcx = beta * Bxn[i][nyn - 2][k];
        omcy = beta * Byn[i][nyn - 2][k];
        omcz = beta * Bzn[i][nyn - 2][k];
        denom = FourPI / 2 * delt * dt / c * qom[is] * rhons[is][i][nyn - 2][k] / (1.0 + omcx * omcx + omcy * omcy + omcz * omcz);
        susxy[i][k] += (omcz + omcx * omcy) * denom;
        susyy[i][k] += (1.0 + omcy * omcy) * denom;
        suszy[i][k] += (omcy * omcz - omcx) * denom;
      }
  }
}

/** Perfect conductor boundary conditions: LEFT wall */
inline void EMfields3D::perfectConductorLeft(double ***imageX, double ***imageY, double ***imageZ, double ***vectorX, double ***vectorY, double ***vectorZ, int dir, Grid * grid) {
  double **susxy;
  double **susyy;
  double **suszy;
  switch (dir) {
    case 0:                    // boundary condition on X-DIRECTION 
      for (int i = 1; i < nyn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          imageX[1][i][j] = vectorX[1][i][j];
          imageY[1][i][j] = vectorY[1][i][j];
          imageZ[1][i][j] = vectorZ[1][i][j];
        }
      break;
    case 1:                    // boundary condition on Y-DIRECTION
      susxy = newArr(double, nxn, nzn);
      susyy = newArr(double, nxn, nzn);
      suszy = newArr(double, nxn, nzn);
      sustensorLeftY(susxy, susyy, suszy);
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          imageX[i][1][j] = vectorX[i][1][j];
          imageY[i][1][j] = vectorY[i][1][j] - (Ey[i][1][j] - susxy[i][j] * vectorX[i][1][j] - suszy[i][j] * vectorZ[i][1][j] - Jyh[i][1][j] * dt * th * FourPI) / susyy[i][j];
          imageZ[i][1][j] = vectorZ[i][1][j];
        }
      break;
    case 2:                    // boundary condition on Y-DIRECTION
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nyn - 1; j++) {
          imageX[i][j][1] = vectorX[i][j][1];
          imageY[i][j][1] = vectorX[i][j][1];
          imageZ[i][j][1] = vectorZ[i][j][1];
        }
      break;
  }
  delArr(susxy, nxn);
  delArr(susyy, nxn);
  delArr(suszy, nxn);
}
/** Perfect conductor boundary conditions: RIGHT wall */
inline void EMfields3D::perfectConductorRight(double ***imageX, double ***imageY, double ***imageZ, double ***vectorX, double ***vectorY, double ***vectorZ, int dir, Grid * grid) {
  double beta, omcx, omcy, omcz, denom;
  double **susxy;
  double **susyy;
  double **suszy;
  switch (dir) {
    case 0:                    // boundary condition on X-DIRECTION RIGHT

      for (int i = 1; i < nyn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          imageX[nxn - 2][i][j] = vectorX[nxn - 2][i][j];
          imageY[nxn - 2][i][j] = vectorY[nxn - 2][i][j];
          imageZ[nxn - 2][i][j] = vectorZ[nxn - 2][i][j];
        }
      break;
    case 1:                    // boundary condition on Y-DIRECTION RIGHT
      susxy = newArr(double, nxn, nzn);
      susyy = newArr(double, nxn, nzn);
      suszy = newArr(double, nxn, nzn);
      sustensorRightY(susxy, susyy, suszy);
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          imageX[i][nyn - 2][j] = vectorX[i][nyn - 2][j];
          imageY[i][nyn - 2][j] = vectorY[i][nyn - 2][j] - (Ey[i][nyn - 2][j] - susxy[i][j] * vectorX[i][nyn - 2][j] - suszy[i][j] * vectorZ[i][nyn - 2][j] - Jyh[i][nyn - 2][j] * dt * th * FourPI) / susyy[i][j];
          imageZ[i][nyn - 2][j] = vectorZ[i][nyn - 2][j];
        }
      break;
    case 2:                    // boundary condition on Z-DIRECTION RIGHT

      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nyn - 1; j++) {
          imageX[i][j][nzn - 2] = vectorX[i][j][nzn - 2];
          imageY[i][j][nzn - 2] = vectorY[i][j][nzn - 2];
          imageZ[i][j][nzn - 2] = vectorZ[i][j][nzn - 2];
        }
      break;
  }
  delArr(susxy, nxn);
  delArr(susyy, nxn);
  delArr(suszy, nxn);
}
/** Perfect conductor boundary conditions for source: LEFT WALL*/
inline void EMfields3D::perfectConductorLeftS(double ***vectorX, double ***vectorY, double ***vectorZ, int dir) {
  switch (dir) {
    case 0:                    // boundary condition on X-DIRECTION LEFT
      for (int i = 1; i < nyn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          vectorX[1][i][j] = 0.0;
          vectorY[1][i][j] = 0.0;
          vectorZ[1][i][j] = 0.0;
        }
      break;
    case 1:                    // boundary condition on Y-DIRECTION LEFT
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          vectorX[i][1][j] = 0.0;
          vectorY[i][1][j] = 0.0;
          vectorZ[i][1][j] = 0.0;
        }
      break;
    case 2:                    // boundary condition on Z-DIRECTION LEFT
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nyn - 1; j++) {
          vectorX[i][j][1] = 0.0;
          vectorY[i][j][1] = 0.0;
          vectorZ[i][j][1] = 0.0;
        }
      break;
  }
}

/** Perfect conductor boundary conditions for source: RIGHT WALL*/
inline void EMfields3D::perfectConductorRightS(double ***vectorX, double ***vectorY, double ***vectorZ, int dir) {
  switch (dir) {
    case 0:                    // boundary condition on X-DIRECTION RIGHT
      for (int i = 1; i < nyn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          vectorX[nxn - 2][i][j] = 0.0;
          vectorY[nxn - 2][i][j] = 0.0;
          vectorZ[nxn - 2][i][j] = 0.0;
        }
      break;
    case 1:                    // boundary condition on Y-DIRECTION RIGHT
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nzn - 1; j++) {
          vectorX[i][nyn - 2][j] = 0.0;
          vectorY[i][nyn - 2][j] = 0.0;
          vectorZ[i][nyn - 2][j] = 0.0;
        }
      break;
    case 2:
      for (int i = 1; i < nxn - 1; i++)
        for (int j = 1; j < nyn - 1; j++) {
          vectorX[i][j][nzn - 2] = 0.0;
          vectorY[i][j][nzn - 2] = 0.0;
          vectorZ[i][j][nzn - 2] = 0.0;
        }
      break;
  }

}
/** get Potential array ***/
inline double ***EMfields3D::getPHI() {
  return (PHI);
}
/** get Ex(X,Y,Z)  */
inline double &EMfields3D::getEx(int indexX, int indexY, int indexZ) const {
  return (Ex[indexX][indexY][indexZ]);
}
/** get Electric field  component X array*/
inline double ***EMfields3D::getEx() {
  return (Ex);
}
/** get Ey(X,Y,Z)  */
inline double &EMfields3D::getEy(int indexX, int indexY, int indexZ) const {
  return (Ey[indexX][indexY][indexZ]);
}
/** get Electric field  component Y array*/
inline double ***EMfields3D::getEy() {
  return (Ey);
}
/** get Ez(X,Y,Z)  */
inline double &EMfields3D::getEz(int indexX, int indexY, int indexZ) const {
  return (Ez[indexX][indexY][indexZ]);
}
/** get Electric field  component Z array*/
inline double ***EMfields3D::getEz() {
  return (Ez);
}
/** get Bx(X,Y,Z)  */
inline double &EMfields3D::getBx(int indexX, int indexY, int indexZ) const {
  return (Bxn[indexX][indexY][indexZ]);
}
/** get Magnetic Field component X array*/
inline double ***EMfields3D::getBx() {
  return (Bxn);
}
/**  get By(X,Y,Z) */
inline double &EMfields3D::getBy(int indexX, int indexY, int indexZ) const {
  return (Byn[indexX][indexY][indexZ]);
}
/** get Magnetic Field component Y array*/
inline double ***EMfields3D::getBy() {
  return (Byn);
}
/**  get Bz(X,Y,Z) */
inline double &EMfields3D::getBz(int indexX, int indexY, int indexZ) const {
  return (Bzn[indexX][indexY][indexZ]);
}
/** get Magnetic Field component Z array*/
inline double ***EMfields3D::getBz() {
  return (Bzn);
}
/** get rhoc(X,Y,Z) */
inline double &EMfields3D::getRHOc(int indexX, int indexY, int indexZ) const {
  return (rhoc[indexX][indexY][indexZ]);
}
inline double ***EMfields3D::getRHOc() {
  return (rhoc);
}
/** get density on node(indexX,indexY,indexZ)  */
inline double &EMfields3D::getRHOn(int indexX, int indexY, int indexZ) const {
  return (rhon[indexX][indexY][indexZ]);
}
/** get density array defined on nodes*/
inline double ***EMfields3D::getRHOn() {
  return (rhon);
}
/** get rhos(X,Y,Z) : density for species*/
inline double &EMfields3D::getRHOns(int indexX, int indexY, int indexZ, int is) const {
  return (rhons[is][indexX][indexY][indexZ]);
}
/** SPECIES: get density array defined on center cells  */
inline double &EMfields3D::getRHOcs(int indexX, int indexY, int indexZ, int is) const {
  return (rhocs[is][indexX][indexY][indexZ]);
}
/** get density array defined on nodes*/
inline double ****EMfields3D::getRHOns() {
  return (rhons);
}
/** SPECIES: get pressure tensor component XX defined on nodes */
inline double ****EMfields3D::getpXXsn() {
  return (pXXsn);
}
/** SPECIES: get pressure tensor component XY defined on nodes */
inline double ****EMfields3D::getpXYsn() {
  return (pXYsn);
}
/** SPECIES: get pressure tensor component XZ defined on nodes */
inline double ****EMfields3D::getpXZsn() {
  return (pXZsn);
}
/** SPECIES: get pressure tensor component YY defined on nodes */
inline double ****EMfields3D::getpYYsn() {
  return (pYYsn);
}
/** SPECIES: get pressure tensor component YZ defined on nodes */
inline double ****EMfields3D::getpYZsn() {
  return (pYZsn);
}
/** SPECIES: get pressure tensor component ZZ defined on nodes */
inline double ****EMfields3D::getpZZsn() {
  return (pZZsn);
}
/** get current -Direction X */
inline double &EMfields3D::getJx(int indexX, int indexY, int indexZ) const {
  return (Jx[indexX][indexY][indexZ]);
}
/** get current array X component **/
inline double ***EMfields3D::getJx() {
  return (Jx);
}
/** get current -Direction Y */
inline double &EMfields3D::getJy(int indexX, int indexY, int indexZ) const {
  return (Jy[indexX][indexY][indexZ]);
}
/** get current array Y component **/
inline double ***EMfields3D::getJy() {
  return (Jy);
}
/** get current -Direction Z */
inline double &EMfields3D::getJz(int indexX, int indexY, int indexZ) const {
  return (Jz[indexX][indexY][indexZ]);
}
/** get current array Z component **/
inline double ***EMfields3D::getJz() {
  return (Jz);
}
/**SPECIES: get current array X component */
inline double ****EMfields3D::getJxs() {
  return (Jxs);
}
/** get Jxs(X,Y,Z,is) : density for species*/
inline double &EMfields3D::getJxs(int indexX, int indexY, int indexZ, int is) const {
  return (Jxs[is][indexX][indexY][indexZ]);
}
/**SPECIES: get current array Y component */
inline double ****EMfields3D::getJys() {
  return (Jys);
}
/** get Jxs(X,Y,Z,is) : density for species*/
inline double &EMfields3D::getJys(int indexX, int indexY, int indexZ, int is) const {
  return (Jys[is][indexX][indexY][indexZ]);
}
/**SPECIES: get current array Z component */
inline double ****EMfields3D::getJzs() {
  return (Jzs);
}
/** get Jxs(X,Y,Z,is) : density for species*/
inline double &EMfields3D::getJzs(int indexX, int indexY, int indexZ, int is) const {
  return (Jzs[is][indexX][indexY][indexZ]);
}


/** get the electric field energy*/
inline double EMfields3D::getEenergy(void) {
  double localEenergy = 0.0;
  double totalEenergy = 0.0;
  for (int i = 1; i < nxn - 2; i++)
    for (int j = 1; j < nyn - 2; j++)
      for (int k = 1; k < nzn - 2; k++)
        localEenergy += .5 * dx * dy * dz * (Ex[i][j][k] * Ex[i][j][k] + Ey[i][j][k] * Ey[i][j][k] + Ez[i][j][k] * Ez[i][j][k]) / (FourPI);

  //MPI_Allreduce(&localEenergy, &totalEenergy, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  totalEenergy = localEenergy;
  return (totalEenergy);

}

/** get the magnetic field energy*/
inline double EMfields3D::getBenergy(void) {
  double localBenergy = 0.0;
  double totalBenergy = 0.0;
  for (int i = 1; i < nxn - 2; i++)
    for (int j = 1; j < nyn - 2; j++)
      for (int k = 1; k < nzn - 2; k++)
        localBenergy += .5 * dx * dy * dz * (Bxn[i][j][k] * Bxn[i][j][k] + Byn[i][j][k] * Byn[i][j][k] + Bzn[i][j][k] * Bzn[i][j][k]) / (FourPI);
    totalBenergy = localBenergy;
  //MPI_Allreduce(&localBenergy, &totalBenergy, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  return (totalBenergy);
}

/** get bulk kinetic energy*/
inline double EMfields3D::getBulkEnergy(int is) {
  double localBenergy = 0.0;
  double totalBenergy = 0.0;
  for (int i = 1; i < nxn - 2; i++)
    for (int j = 1; j < nyn - 2; j++)
      for (int k = 1; k < nzn - 2; k++)
        localBenergy += .5 * dx * dy * dz * (Jxs[is][i][j][k] * Jxs[is][i][j][k] + Jys[is][i][j][k] * Jys[is][i][j][k] + Jzs[is][i][j][k] * Jzs[is][i][j][k]) / (rhons[is][i][j][k]);

  //MPI_Allreduce(&localBenergy, &totalBenergy, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  totalBenergy = localBenergy;
  return (totalBenergy);
}




inline int EMfields3D::getStopFlag(void){
    return(stopFlag);
}


/** get the magnetic field energy*/
inline double EMfields3D::getBenergyMHD(void) {
    double localBenergy = 0.0;
    double totalBenergy = 0.0;
    for (int i = 2; i < nxc-2; i++)
        for (int j = 2; j < nyc-2; j++)
            for (int k = 2; k < nzc-2; k++)
                localBenergy += .5 * dx * dy * dz * (Bxc[i][j][k] * Bxc[i][j][k] + Byc[i][j][k] * Byc[i][j][k] + Bzc[i][j][k] * Bzc[i][j][k]);
    
    //MPI_Allreduce(&localBenergy, &totalBenergy, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    totalBenergy = localBenergy;
    return (totalBenergy);
}

/** get the total energy*/
inline double EMfields3D::getTotenergyMHD(void) {
    double localU = 0.0;
    double totalU = 0.0;
    for (int i = 2; i < nxc-2; i++)
        for (int j = 2; j < nyc-2; j++)
            for (int k = 2; k < nzc-2; k++)
                localU += dx * dy * dz * Uc[i][j][k];
    totalU = localU;
    //MPI_Allreduce(&localU, &totalU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    return (totalU);
}

/** get the total energy*/
inline double EMfields3D::getKeMHD(void) {
    double localU = 0.0;
    double totalU = 0.0;
    for (int i = 2; i < nxc-2; i++)
        for (int j = 2; j < nyc-2; j++)
            for (int k = 2; k < nzc-2; k++)
                localU += .5 * dx * dy * dz * (pxc[i][j][k] * pxc[i][j][k] + pyc[i][j][k] * pyc[i][j][k] + pzc[i][j][k] * pzc[i][j][k]) / rhoc[i][j][k];
    
    totalU = localU;
    //MPI_Allreduce(&localU, &totalU, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    return (totalU);
}





/** Print info about electromagnetic field */
inline void EMfields3D::print(void) const {


}
/** destructor: deallocate arrays*/
inline EMfields3D::~EMfields3D() {
  // nodes
  delArr3(Ex, nxn, nyn);
  delArr3(Ey, nxn, nyn);
  delArr3(Ez, nxn, nyn);
  delArr3(Exth, nxn, nyn);
  delArr3(Eyth, nxn, nyn);
  delArr3(Ezth, nxn, nyn);
  delArr3(Bxn, nxn, nyn);
  delArr3(Byn, nxn, nyn);
  delArr3(Bzn, nxn, nyn);
  delArr3(rhon, nxn, nyn);
  delArr3(Jx, nxn, nyn);
  delArr3(Jy, nxn, nyn);
  delArr3(Jz, nxn, nyn);
  delArr3(Jxh, nxn, nyn);
  delArr3(Jyh, nxn, nyn);
  delArr3(Jzh, nxn, nyn);
  // nodes and species
  delArr4(rhons, ns, nxn, nyn);
  delArr4(Jxs, ns, nxn, nyn);
  delArr4(Jys, ns, nxn, nyn);
  delArr4(Jzs, ns, nxn, nyn);
  delArr4(pXXsn, ns, nxn, nyn);
  delArr4(pXYsn, ns, nxn, nyn);
  delArr4(pXZsn, ns, nxn, nyn);
  delArr4(pYYsn, ns, nxn, nyn);
  delArr4(pYZsn, ns, nxn, nyn);
  delArr4(pZZsn, ns, nxn, nyn);
  // central points
  delArr3(PHI, nxc, nyc);
//  delArr3(Bxc, nxc, nyc);
//  delArr3(Byc, nxc, nyc);
//  delArr3(Bzc, nxc, nyc);
//  delArr3(rhoc, nxc, nyc);
  delArr3(rhoh, nxc, nyc);
  // various stuff needs to be deallocated too
  delArr3(tempXC, nxc, nyc);
  delArr3(tempYC, nxc, nyc);
  delArr3(tempZC, nxc, nyc);
  delArr3(tempXN, nxn, nyn);
  delArr3(tempYN, nxn, nyn);
  delArr3(tempZN, nxn, nyn);
  delArr3(tempC, nxc, nyc);
  delArr3(tempX, nxn, nyn);
  delArr3(tempY, nxn, nyn);
  delArr3(tempZ, nxn, nyn);
  delArr3(temp2X, nxn, nyn);
  delArr3(temp2Y, nxn, nyn);
  delArr3(temp2Z, nxn, nyn);
  delArr3(imageX, nxn, nyn);
  delArr3(imageY, nxn, nyn);
  delArr3(imageZ, nxn, nyn);
  delArr3(Dx, nxn, nyn);
  delArr3(Dy, nxn, nyn);
  delArr3(Dz, nxn, nyn);
  delArr3(vectX, nxn, nyn);
  delArr3(vectY, nxn, nyn);
  delArr3(vectZ, nxn, nyn);
  delArr3(divC, nxc, nyc);
    
    
  
  // MHD_update
    
    delArr3(Bxc,nxc,nyc);
    delArr3(Byc,nxc,nyc);
    delArr3(Bzc,nxc,nyc);
    
    delArr3(Uc,nxc,nyc);
    
    delArr3(pxc,nxc,nyc);
    delArr3(pyc,nxc,nyc);
    delArr3(pzc,nxc,nyc);
    
    delArr3(rhoc,nxc,nyc);
    
    delArr3(grid_eta,nxc,nyc);
    
    
    // Flux limiters
    delArr3(dBxc,nxc,nyc);
    delArr3(dByc,nxc,nyc);
    delArr3(dBzc,nxc,nyc);
    
    delArr3(dUc,nxc,nyc);
    
    delArr3(dpxc,nxc,nyc);
    delArr3(dpyc,nxc,nyc);
    delArr3(dpzc,nxc,nyc);
    
    delArr3(drhoc,nxc,nyc);
    
    
    // Sources
    delArr3(Q__Bxc,nxc,nyc);
    delArr3(Q__Byc,nxc,nyc);
    delArr3(Q__Bzc,nxc,nyc);
    
    delArr3(Q__Uc,nxc,nyc);
    
    delArr3(Q__pxc,nxc,nyc);
    delArr3(Q__pyc,nxc,nyc);
    delArr3(Q__pzc,nxc,nyc);
    
    delArr3(Q__rhoc,nxc,nyc);
    
    //Keep the copy
    delArr3(Bxc1,nxc,nyc);
    delArr3(Byc1,nxc,nyc);
    delArr3(Bzc1,nxc,nyc);
    
    delArr3(Uc1,nxc,nyc);
    
    delArr3(pxc1,nxc,nyc);
    delArr3(pyc1,nxc,nyc);
    delArr3(pzc1,nxc,nyc);
    
    delArr3(rhoc1,nxc,nyc);
    
    
    // ANOTHER 2 COPIES!!!
    
    delArr3(Bxc2,nxc,nyc);
    delArr3(Byc2,nxc,nyc);
    delArr3(Bzc2,nxc,nyc);
    
    delArr3(Uc2,nxc,nyc);
    
    delArr3(pxc2,nxc,nyc);
    delArr3(pyc2,nxc,nyc);
    delArr3(pzc2,nxc,nyc);
    
    delArr3(rhoc2,nxc,nyc);
    
    delArr3(Bxc0,nxc,nyc);
    delArr3(Byc0,nxc,nyc);
    delArr3(Bzc0,nxc,nyc);
    
    delArr3(Uc0,nxc,nyc);
    
    delArr3(pxc0,nxc,nyc);
    delArr3(pyc0,nxc,nyc);
    delArr3(pzc0,nxc,nyc);
    
    delArr3(rhoc0,nxc,nyc);
    
    //COPY
    delArr3(roStored,nxc,nyc);
    delArr3(pxStored,nxc,nyc);
    delArr3(pyStored,nxc,nyc);
    delArr3(pzStored,nxc,nyc);
    delArr3(BxStored,nxc,nyc);
    delArr3(ByStored,nxc,nyc);
    delArr3(BzStored,nxc,nyc);
    delArr3(UStored,nxc,nyc);
    
    
}



//used for usual Hesse-Birn boundary simulations.
inline void EMfields3D::Boundarize_fi(int nx, int ny, int nz, double ***mfi, VirtualTopology3D *vct){
    // !!!! This is 'Z' boundary. For 25D Cartesian Nz=1
    // Make it 'Simply Open' (similar to TVDLF2010 code)
    
    int x,y,z;
    
    BEGIN_z_full
    BEGIN_y_full
    for (int jj=0;jj<2;jj++){
        //mfi[x][y][jj]=mfi[x][y][2];
        mfi[jj][y][z]=0.0;//mfi[2][y][z];
    }
    END_y_full
    END_z_full

    BEGIN_z_full
    BEGIN_y_full
    for (int jj=nx-2;jj<nx;jj++){
        //mfi[x][y][jj]=mfi[x][y][2];
        mfi[jj][y][z]=0.0;//mfi[nx-3][y][z];
    }
    END_y_full
    END_z_full



    /*
     //z=0;
     BEGIN_x_full
     BEGIN_y_full
     for (int jj=0;jj<2;jj++){
     //mfi[x][y][jj]=mfi[x][y][2];
     mfi[x][y][jj]=0.0;
     }
     END_y_full
     END_x_full
     
     
     //z=SIZE_Z-1; - ok
     BEGIN_x_full
     BEGIN_y_full
     for (int jj=nz-2;jj<nz;jj++){
     //mfi[x][y][jj]=mfi[x][y][nz-3];
     mfi[x][y][jj]=0.0;
     }
     END_y_full
     END_x_full
     */
    
    // Note: Periodic BCs are set by communicator!
/*
    if (vct->getZleft_neighbor() ==MPI_PROC_NULL){
    //z=0;
    BEGIN_x_full
    BEGIN_y_full
    for (int jj=0;jj<2;jj++){
          mfi[x][y][jj]=mfi[x][y][2];
        //mfi[x][y][jj]=mfi[x][y][nz-4+jj];
    }
    END_y_full
    END_x_full
    }
    
    if (vct->getZright_neighbor() ==MPI_PROC_NULL){
    //z=SIZE_Z-1; - ok
    BEGIN_x_full
    BEGIN_y_full
    for (int jj=nz-2;jj<nz;jj++){
          mfi[x][y][jj]=mfi[x][y][nz-3];
        //mfi[x][y][jj]=mfi[x][y][jj-nz+4];
    }
    END_y_full
    END_x_full
    }
    */
    
    
    
}


//
//
//inline void EMfields3D::calculateFieldMHD(Grid *grid, VirtualTopology3D *vct, int step_cycle){
//    
//    int temp_int;
//    /*
//     if(MHDtime<0.1) relaxation_wrapper(grid, vct, step_cycle);
//     else
//     {
//     
//     if(COPY_STATE)
//     {
//     copy_state(nxc,nyc,nzc,vct);
//     COPY_STATE=false;
//     cout<<"COPY STATE DONE"<<endl;
//     }
//     */
//    
//    //cout<< "EMfields::calculateFieldMHD" << endl;
//    
//    //cout<<"dt before calc_dt: "<< dt<<endl;
//   
//   //  if ((step_cycle%10)==0) Poisson_correction_MHD(grid,vct);
///*
//
//    // MPI_Barrier(MPI_COMM_WORLD);
//    calc_dt(grid,vct);
//    //cout<<"dt after calc_dt: "<< dt<<endl;
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Lx(1,grid,vct);
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Ly(1,grid,vct);
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Lz(1,grid,vct);
//    
//    MHDtime=MHDtime+dt;
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Lz(1,grid,vct);
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Ly(1,grid,vct);
//    
//    // MPI_Barrier(MPI_COMM_WORLD);
//    temp_int=calc_Lx(1,grid,vct);
// 
//    MHDtime=MHDtime+dt;
//    */
//    
//	if (vct->getCartesian_rank() ==0){
//        cout << "time cycle"<< step_cycle << " simulation: "<<MHDtime<< endl;
//		cout << temp_int << endl;
//  		cout << "calc_Lx calculated" << endl;
//    }
// 
//
//    /*
//     }
//     */
//}
//



inline void EMfields3D::calculateFieldMHD(Grid *grid, VirtualTopology3D *vct, int step_cycle){
    
    int temp_int;
    /*
     if(MHDtime<0.1) relaxation_wrapper(grid, vct, step_cycle);
     else
     {
     
     if(COPY_STATE)
     {
     copy_state(nxc,nyc,nzc,vct);
     COPY_STATE=false;
     cout<<"COPY STATE DONE"<<endl;
     }
     */
    
    //cout<< "EMfields::calculateFieldMHD" << endl;
    
    //cout<<"dt before calc_dt: "<< dt<<endl;
    
    calc_dt(grid,vct);
    //cout<<"dt after calc_dt: "<< dt<<endl;
    temp_int=calc_Lx(1,grid,vct);
    temp_int=calc_Ly(1,grid,vct);
//    temp_int=calc_Lz(1,grid,vct);
    MHDtime=MHDtime+dt;
//    temp_int=calc_Lz(1,grid,vct);
    temp_int=calc_Ly(1,grid,vct);
    temp_int=calc_Lx(1,grid,vct);
    MHDtime=MHDtime+dt;
    
    if ((step_cycle%10)==0) Poisson_correction_MHD(grid,vct);
    
	if (vct->getCartesian_rank() == 0){
        cout << "Computational cycle: "<< step_cycle << endl;
        cout << "Simulation time: "<< MHDtime << endl;
	//	cout << temp_int << endl;
  	//	cout << "calc_Lx calculated" << endl;
    }
    /*
     }
     */
}

inline void EMfields3D::initRhoTest2(VirtualTopology3D *vct, Grid *grid){
    double xpert,ypert,zpert, rho_size,exp_pert;
    int order,x,y,z;
    order = 5;
    rho_size = 0.5;
    stopFlag=1;
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "       Initialize 3D null point(s)" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
        }
        //variables used by the MHD solver: Bxc, Byc, Bzc, pxc, pyc, pzc, Uc,
        for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
                for (int jj=0; jj <nzc; jj++) {
                    Bxc[i][j][jj] = 0.0;
                    Byc[i][j][jj] = 0.0;
                    Bzc[i][j][jj] = 0.0;
                    
                    xpert = grid->getXC(i,j,jj)- Lx/3.0;
                    ypert = grid->getYC(i,j,jj)- Ly/3.0;
                    zpert = grid->getZC(i,j,jj)- Lz/4.0;
                    
                    
                    exp_pert = exp(-(xpert/rho_size)*(xpert/rho_size)-(ypert/rho_size)*(ypert/rho_size)-(zpert/rho_size)*(zpert/rho_size));
                    
                    
                    rhoc[i][j][jj] = 1.0 + 1.0*exp_pert;
                    pxc[i][j][jj] = 0.0;
                    pyc[i][j][jj] = 0.0;
                    pzc[i][j][jj] = 0.0;
                    Uc[i][j][jj]=(((rhoc[i][j][jj])*1.0)/(C_g-1)) + ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5;
                }
    
    
    
    Boundarize(nxc,nyc,nzc,rhoc,pxc,pyc,pzc,Bxc,Byc,Bzc,Uc,vct);
    
    
    
    eq(BxStored,Bxc,nxc,nyc,nzc);
    eq(ByStored,Byc,nxc,nyc,nzc);
    eq(BzStored,Bzc,nxc,nyc,nzc);
    eq(pxStored,pxc,nxc,nyc,nzc);
    eq(pyStored,pyc,nxc,nyc,nzc);
    eq(pzStored,pzc,nxc,nyc,nzc);
    eq(UStored,Uc,nxc,nyc,nzc);
    eq(roStored,rhoc,nxc,nyc,nzc);
    
    
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    cout<<"Test communication start "<<endl;
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
    communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
    communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
    
    cout<<"Test communication done "<<endl;
    
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    eq(Bxc0,Bxc,nxc,nyc,nzc);
    eq(Byc0,Byc,nxc,nyc,nzc);
    eq(Bzc0,Bzc,nxc,nyc,nzc);
    eq(pxc0,pxc,nxc,nyc,nzc);
    eq(pyc0,pyc,nxc,nyc,nzc);
    eq(pzc0,pzc,nxc,nyc,nzc);
    eq(Uc0,Uc,nxc,nyc,nzc);
    eq(rhoc0,rhoc,nxc,nyc,nzc);
    
    eqValue(0.0, Q__rhoc, nxc, nyc, nzc);
    eqValue(0.0, Q__pxc, nxc, nyc, nzc);
    eqValue(0.0, Q__pyc, nxc, nyc, nzc);
    eqValue(0.0, Q__pzc, nxc, nyc, nzc);
    eqValue(0.0, Q__Bxc, nxc, nyc, nzc);
    eqValue(0.0, Q__Byc, nxc, nyc, nzc);
    eqValue(0.0, Q__Bzc, nxc, nyc, nzc);
    eqValue(0.0, Q__Uc, nxc, nyc, nzc);
    
    
} else {
    init(vct,grid);  // use the fields from restart file
}
}



inline void EMfields3D::initRhoTest(VirtualTopology3D *vct, Grid *grid){
    int x, y, z;
    int order=0;
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "       Initialize 3D null point(s)" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
		}
        
        for (int i=0; i < nxn; i++)
            for (int j=0; j < nyn; j++)
                for (int k=0; k < nzn; k++){
                    // initialize the density for species
                    for (int is=0; is < ns; is++){
                        //    rhons[is][i][j][k] = rhoINIT[is]/FourPI;
                        //rhons[is][i][j][k] = 1./FourPI*(0.0004 - 0.5*(Bxn[i][j][k]*Bxn[i][j][k] + Byn[i][j][k]*Byn[i][j][k] + Bzn[i][j][k]*Bzn[i][j][k]))/((uth[0]*uth[0])/abs(qom[0]) + (uth[1]*uth[1]));
                    }
                    // electric field
                    //  Ex[i][j][k] =  0.0;
                    //  Ey[i][j][k] =  0.0;
                    //  Ez[i][j][k] =  0.0;
                    // Magnetic field
                    // Bxn[i][j][k] = -B0x*cos(2.*M_PI*grid->getXN(i,j,k)/Lx)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
                    // Byn[i][j][k] = B0x*cos(2.*M_PI*grid->getYN(i,j,k)/Ly)*(-2.*sin(2.*M_PI*grid->getZN(i,j,k)/Lz) + sin(2.*M_PI*grid->getXN(i,j,k)/Lx));
                    // Bzn[i][j][k] = 2.*B0x*cos(2.*M_PI*grid->getZN(i,j,k)/Lz)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
                }
        
        //      communicateNode(nxn,nyn,Ex,vct);
        //		communicateNode(nxn,nyn,Ey,vct);
        //		communicateNode(nxn,nyn,Ez,vct);
        
        
        /*
         // currents are used to calculate in the Maxwell's solver
         // The ion current is equal to 0 (all current is on electrons)
         for (int i=0; i < nxn; i++)
         for (int j=0; j < nyn; j++)
         for (int k=0; k < nzn; k++){
         Jxs[1][i][j][k] = 0.0; // ion species is species 1
         Jys[1][i][j][k] = 0.0; // ion species is species 1
         Jzs[1][i][j][k] = 0.0; // ion species is species 1
         }
         // calculate the electron current from
         eqValue(0.0,tempXN,nxn,nyn,nzn);
         eqValue(0.0,tempYN,nxn,nyn,nzn);
         eqValue(0.0,tempZN,nxn,nyn,nzn);
         grid->curlC2N(tempXN,tempYN,tempZN,Bxc,Byc,Bzc); // here you calculate curl(B)
         // all current is on electrons, calculated from Ampere's law
         for (int i=0; i < nxn; i++)
         for (int j=0; j < nyn; j++)
         for (int k=0; k < nzn; k++){  // electrons are species 0
         Jxs[0][i][j][k] = c*tempXN[i][j][k]/FourPI; // ion species is species 1
         Jys[0][i][j][k] = c*tempYN[i][j][k]/FourPI; // ion species is species 1
         Jzs[0][i][j][k] = c*tempZN[i][j][k]/FourPI; // ion species is species 1
         }
         */
        //		for (int is=0 ; is<ns; is++)
        //			grid->interpN2C(rhocs,is,rhons);
        //        variables used by the MHD solver: Bxc, Byc, Bzc, pxc, pyc, pzc, Uc, /
        
        
		for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
                for (int jj=0; jj <nzc; jj++){
                    Bxc[i][j][jj] = 0.0;
                    Byc[i][j][jj] = 1.0;
                    
                    
                    if (vct->getCartesian_rank()==0)
                        Bzc[i][j][jj] = 1.0;
                    else
                        Bzc[i][j][jj] = 2.0;
                    
                    rhoc[i][j][jj] = rhoINIT[0];
                    pxc[i][j][jj] = 0.0;
                    pyc[i][j][jj] = 0.0;
                    pzc[i][j][jj] = 0.0;
                    
                    
                    Uc[i][j][jj]=(((rhoc[i][j][jj])*0.6)/(C_g-1)) + ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5;
                }
        
        
        Boundarize(nxc,nyc,nzc,rhoc,pxc,pyc,pzc,Bxc,Byc,Bzc,Uc,vct);
        
        
        
        eq(BxStored,Bxc,nxc,nyc,nzc);
        eq(ByStored,Byc,nxc,nyc,nzc);
        eq(BzStored,Bzc,nxc,nyc,nzc);
        eq(pxStored,pxc,nxc,nyc,nzc);
        eq(pyStored,pyc,nxc,nyc,nzc);
        eq(pzStored,pzc,nxc,nyc,nzc);
        eq(UStored,Uc,nxc,nyc,nzc);
        eq(roStored,rhoc,nxc,nyc,nzc);
        
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        cout<<"Test communication start "<<endl;
        
        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
        
        cout<<"Test communication done "<<endl;
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        eq(Bxc0,Bxc,nxc,nyc,nzc);
        eq(Byc0,Byc,nxc,nyc,nzc);
        eq(Bzc0,Bzc,nxc,nyc,nzc);
        eq(pxc0,pxc,nxc,nyc,nzc);
        eq(pyc0,pyc,nxc,nyc,nzc);
        eq(pzc0,pzc,nxc,nyc,nzc);
        eq(Uc0,Uc,nxc,nyc,nzc);
        eq(rhoc0,rhoc,nxc,nyc,nzc);
        
        eqValue(0.0, Q__rhoc, nxc, nyc, nzc);
        eqValue(0.0, Q__pxc, nxc, nyc, nzc);
        eqValue(0.0, Q__pyc, nxc, nyc, nzc);
        eqValue(0.0, Q__pzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bxc, nxc, nyc, nzc);
        eqValue(0.0, Q__Byc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Uc, nxc, nyc, nzc);
//        
//        
//        cout<<"Test communication start "<<endl;
//        
//        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
//        
//        cout<<"Test communication done "<<endl;
//        cout<<"Test communication start "<<endl;
//        
//        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
//        
//        cout<<"Test communication done "<<endl;
//        cout<<"Test communication start "<<endl;
//        
//        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
//        
//        cout<<"Test communication done "<<endl;
//        
//        
//        cout<<"Test communication start "<<endl;
//        
//        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
//        
//        cout<<"Test communication done "<<endl;
//        
//        cout<<"Test communication start "<<endl;
//        
//        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
//        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
//        
//        cout<<"Test communication done "<<endl;
//        
//        BEGIN_full
//        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
//        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
//        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
//        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
//        
//        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
//        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
//        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
//        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
//        END_full
        
        
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}



inline void EMfields3D::initNullPoint_MHD(VirtualTopology3D *vct, Grid *grid){
    int x, y, z;
    int order=0;
	if (restart1 ==0){
		
		// initialize
		if (vct->getCartesian_rank() ==0){
			cout << "----------------------------------------" << endl;
			cout << "       Initialize 3D null point(s)" << endl;
			cout << "----------------------------------------" << endl;
			cout << "B0x                              = " << B0x << endl;
			cout << "B0y                              = " << B0y << endl;
			cout << "B0z                              = " << B0z << endl;
			for (int i=0; i < ns; i++){
				cout << "rho species " << i <<" = " << rhoINIT[i];
			}
			cout << "Smoothing Factor = " << Smooth << endl;
			cout << "-------------------------" << endl;
		}
        
        for (int i=0; i < nxn; i++)
            for (int j=0; j < nyn; j++)
                for (int k=0; k < nzn; k++){
                    // initialize the density for species
                    for (int is=0; is < ns; is++){
                        //    rhons[is][i][j][k] = rhoINIT[is]/FourPI;
                        //rhons[is][i][j][k] = 1./FourPI*(0.0004 - 0.5*(Bxn[i][j][k]*Bxn[i][j][k] + Byn[i][j][k]*Byn[i][j][k] + Bzn[i][j][k]*Bzn[i][j][k]))/((uth[0]*uth[0])/abs(qom[0]) + (uth[1]*uth[1]));
                    }
                    // electric field
                    //  Ex[i][j][k] =  0.0;
                    //  Ey[i][j][k] =  0.0;
                    //  Ez[i][j][k] =  0.0;
                    // Magnetic field
                    // Bxn[i][j][k] = -B0x*cos(2.*M_PI*grid->getXN(i,j,k)/Lx)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
                    // Byn[i][j][k] = B0x*cos(2.*M_PI*grid->getYN(i,j,k)/Ly)*(-2.*sin(2.*M_PI*grid->getZN(i,j,k)/Lz) + sin(2.*M_PI*grid->getXN(i,j,k)/Lx));
                    // Bzn[i][j][k] = 2.*B0x*cos(2.*M_PI*grid->getZN(i,j,k)/Lz)*sin(2.*M_PI*grid->getYN(i,j,k)/Ly);
                }
        
        //      communicateNode(nxn,nyn,Ex,vct);
        //		communicateNode(nxn,nyn,Ey,vct);
        //		communicateNode(nxn,nyn,Ez,vct);
        
        
        /*
         // currents are used to calculate in the Maxwell's solver
         // The ion current is equal to 0 (all current is on electrons)
         for (int i=0; i < nxn; i++)
         for (int j=0; j < nyn; j++)
         for (int k=0; k < nzn; k++){
         Jxs[1][i][j][k] = 0.0; // ion species is species 1
         Jys[1][i][j][k] = 0.0; // ion species is species 1
         Jzs[1][i][j][k] = 0.0; // ion species is species 1
         }
         // calculate the electron current from
         eqValue(0.0,tempXN,nxn,nyn,nzn);
         eqValue(0.0,tempYN,nxn,nyn,nzn);
         eqValue(0.0,tempZN,nxn,nyn,nzn);
         grid->curlC2N(tempXN,tempYN,tempZN,Bxc,Byc,Bzc); // here you calculate curl(B)
         // all current is on electrons, calculated from Ampere's law
         for (int i=0; i < nxn; i++)
         for (int j=0; j < nyn; j++)
         for (int k=0; k < nzn; k++){  // electrons are species 0
         Jxs[0][i][j][k] = c*tempXN[i][j][k]/FourPI; // ion species is species 1
         Jys[0][i][j][k] = c*tempYN[i][j][k]/FourPI; // ion species is species 1
         Jzs[0][i][j][k] = c*tempZN[i][j][k]/FourPI; // ion species is species 1
         }
         */
        //		for (int is=0 ; is<ns; is++)
        //			grid->interpN2C(rhocs,is,rhons);
        //        variables used by the MHD solver: Bxc, Byc, Bzc, pxc, pyc, pzc, Uc, /
        
        
		for (int i=0; i <nxc; i++)
			for (int j=0; j <nyc; j++)
                for (int jj=0; jj <nzc; jj++){
                    Bxc[i][j][jj] = -B0x*cos(2.*M_PI*grid->getXC(i,j,jj)/Lx)*sin(2.*M_PI*grid->getYC(i,j,jj)/Ly);
                    Byc[i][j][jj] = B0x*cos(2.*M_PI*grid->getYC(i,j,jj)/Ly)*(-2.*sin(2.*M_PI*grid->getZC(i,j,jj)/Lz) + sin(2.*M_PI*grid->getXC(i,j,jj)/Lx));
                    Bzc[i][j][jj] = 2.*B0x*cos(2.*M_PI*grid->getZC(i,j,jj)/Lz)*sin(2.*M_PI*grid->getYC(i,j,jj)/Ly);
                    rhoc[i][j][jj] = rhoINIT[0];
                    pxc[i][j][jj] = 0.0;
                    pyc[i][j][jj] = 0.0;
                    pzc[i][j][jj] = 0.0;
                    
                    
                    Uc[i][j][jj]=(((rhoc[i][j][jj])*0.06)/(C_g-1)) + ((Bxc[i][j][jj]*Bxc[i][j][jj]) + (Byc[i][j][jj]*Byc[i][j][jj]) + (Bzc[i][j][jj]*Bzc[i][j][jj]))*0.5;
                }
        
        
        Boundarize(nxc,nyc,nzc,rhoc,pxc,pyc,pzc,Bxc,Byc,Bzc,Uc,vct);
        
        
        
        eq(BxStored,Bxc,nxc,nyc,nzc);
        eq(ByStored,Byc,nxc,nyc,nzc);
        eq(BzStored,Bzc,nxc,nyc,nzc);
        eq(pxStored,pxc,nxc,nyc,nzc);
        eq(pyStored,pyc,nxc,nyc,nzc);
        eq(pzStored,pzc,nxc,nyc,nzc);
        eq(UStored,Uc,nxc,nyc,nzc);
        eq(roStored,rhoc,nxc,nyc,nzc);
        
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        cout<<"Test communication start "<<endl;
        
        communicateCenterMHD(nxc,nyc,nzc,Bxc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Byc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Bzc,vct);
        communicateCenterMHD(nxc,nyc,nzc,rhoc,vct);
        communicateCenterMHD(nxc,nyc,nzc,Uc,vct);
        
        cout<<"Test communication done "<<endl;
        
        BEGIN_full
        if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "init 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
        if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "init 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "init 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "init 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
        
        if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "init 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "init 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "init 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "init 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
        END_full
        
        
        eq(Bxc0,Bxc,nxc,nyc,nzc);
        eq(Byc0,Byc,nxc,nyc,nzc);
        eq(Bzc0,Bzc,nxc,nyc,nzc);
        eq(pxc0,pxc,nxc,nyc,nzc);
        eq(pyc0,pyc,nxc,nyc,nzc);
        eq(pzc0,pzc,nxc,nyc,nzc);
        eq(Uc0,Uc,nxc,nyc,nzc);
        eq(rhoc0,rhoc,nxc,nyc,nzc);
        
        eqValue(0.0, Q__rhoc, nxc, nyc, nzc);
        eqValue(0.0, Q__pxc, nxc, nyc, nzc);
        eqValue(0.0, Q__pyc, nxc, nyc, nzc);
        eqValue(0.0, Q__pzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bxc, nxc, nyc, nzc);
        eqValue(0.0, Q__Byc, nxc, nyc, nzc);
        eqValue(0.0, Q__Bzc, nxc, nyc, nzc);
        eqValue(0.0, Q__Uc, nxc, nyc, nzc);
        
        
	} else {
		init(vct,grid);  // use the fields from restart file
	}
}



inline void EMfields3D::Boundarize(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, VirtualTopology3D *vct){
    
    
    // !!!! This is 'Z' boundary. For 25D Cartesian Nz=1
    // Make it 'Simply Open' (similar to TVDLF2010 code)
    
    int x,y,z;
    double pressure_temp1;


    //x=nx; - ok
    BEGIN_z_full
    BEGIN_y_full
    for (int jj=nx-2;jj<nx;jj++){
        
        //              (*mro)[x][y][SIZE_Z-1]=(*mro)[x][y][SIZE_Z-3];(*mro)[x][y][SIZE_Z-2]=(*mro)[x][y][SIZE_Z-3];
        //mro[x][y][nz-1]=roStored[x][y][nz-2];
        mro[jj][y][z]=mro[nx-3][y][z];
        
        mpx[jj][y][z]=mpx[nx-3][y][z];
        mpy[jj][y][z]=mpy[nx-3][y][z];
        mpz[jj][y][z]=mpz[nx-3][y][z];
        //mU[x][y][jj]=mU[x][y][nz-3];
        //              (*mBx)[x][y][SIZE_Z-1]=(*mBx)[x][y][SIZE_Z-3];(*mBx)[x][y][SIZE_Z-2]=(*mBx)[x][y][SIZE_Z-3];
        
        //mBx[x][y][nz-1]=BxStored[x][y][nz-2];
        mBx[jj][y][z]=mBx[nx-3][y][z];
        mBy[jj][y][z]=mBy[nx-3][y][z];
        mBz[jj][y][z]=mBz[nx-3][y][z];
        
        
        mU[jj][y][z]=mU[nx-3][y][z];
        
    }
    END_y_full
    END_z_full
    
    
    
    //x=0; - ok
    BEGIN_z_full
    BEGIN_y_full
    for (int jj=0;jj<2;jj++){
        
        //              (*mro)[x][y][SIZE_Z-1]=(*mro)[x][y][SIZE_Z-3];(*mro)[x][y][SIZE_Z-2]=(*mro)[x][y][SIZE_Z-3];
        //mro[x][y][nz-1]=roStored[x][y][nz-2];
        mro[jj][y][z]=mro[2][y][z];
        
        mpx[jj][y][z]=mpx[2][y][z];
        mpy[jj][y][z]=mpy[2][y][z];
        mpz[jj][y][z]=mpz[2][y][z];
        //mU[x][y][jj]=mU[x][y][nz-3];
        //              (*mBx)[x][y][SIZE_Z-1]=(*mBx)[x][y][SIZE_Z-3];(*mBx)[x][y][SIZE_Z-2]=(*mBx)[x][y][SIZE_Z-3];
        
        //mBx[x][y][nz-1]=BxStored[x][y][nz-2];
        mBx[jj][y][z]=mBx[2][y][z];
        mBy[jj][y][z]=mBy[2][y][z];
        mBz[jj][y][z]=mBz[2][y][z];
        
        
        mU[jj][y][z]=mU[2][y][z];
        
    }
    END_y_full
    END_z_full
    
    
    /*
     //z=0;
     BEGIN_x_full
     BEGIN_y_full
     for (int jj=0;jj<2;jj++){
     //              (*mro)[x][y][1]=(*mro)[x][y][2];(*mro)[x][y][0]=(*mro)[x][y][2];
     //              mro[x][y][jj]=roStored[x][y][2];
     mro[x][y][jj]=mro[x][y][2];
     
     mpx[x][y][jj]=mpx[x][y][2];
     mpy[x][y][jj]=mpy[x][y][2];
     mpz[x][y][jj]=mpz[x][y][2];
     mU[x][y][jj]=mU[x][y][2];
     //              (*mBx)[x][y][1]=(*mBx)[x][y][2];(*mBx)[x][y][0]=(*mBx)[x][y][2];
     //              mBx[x][y][jj]=BxStored[x][y][2];
     mBx[x][y][jj]=mBx[x][y][2];
     mBy[x][y][jj]=mBy[x][y][2];
     mBz[x][y][jj]=mBz[x][y][2];
     
     //                mU[x][y][jj]=mU[x][y][2];
     
     }
     END_y_full
     END_x_full
     */
    
    // PERIODIC BOUNDARY in Z
    //z=0;
    /*
    BEGIN_x_full
    BEGIN_y_full
    for (int jj=0;jj<2;jj++){
        mro[x][y][jj]=mro[x][y][nz-4+jj];
        
        mpx[x][y][jj]=mpx[x][y][nz-4+jj];
        mpy[x][y][jj]=mpy[x][y][nz-4+jj];
        mpz[x][y][jj]=mpz[x][y][nz-4+jj];
        mU[x][y][jj]=mU[x][y][nz-4+jj];
        
        mBx[x][y][jj]=mBx[x][y][nz-4+jj];
        mBy[x][y][jj]=mBy[x][y][nz-4+jj];
        mBz[x][y][jj]=mBz[x][y][nz-4+jj];
        
        
    }
    END_y_full
    END_x_full
    */
    //Do nothing: consider fully periodic box.

    /*
     //z=SIZE_Z-1; - ok
     BEGIN_x_full
     BEGIN_y_full
     for (int jj=nz-2;jj<nz;jj++){
     
     //              (*mro)[x][y][SIZE_Z-1]=(*mro)[x][y][SIZE_Z-3];(*mro)[x][y][SIZE_Z-2]=(*mro)[x][y][SIZE_Z-3];
     //mro[x][y][nz-1]=roStored[x][y][nz-2];
     mro[x][y][jj]=mro[x][y][nz-3];
     
     mpx[x][y][jj]=mpx[x][y][nz-3];
     mpy[x][y][jj]=mpy[x][y][nz-3];
     mpz[x][y][jj]=mpz[x][y][nz-3];
     //mU[x][y][jj]=mU[x][y][nz-3];
     //              (*mBx)[x][y][SIZE_Z-1]=(*mBx)[x][y][SIZE_Z-3];(*mBx)[x][y][SIZE_Z-2]=(*mBx)[x][y][SIZE_Z-3];
     
     //mBx[x][y][nz-1]=BxStored[x][y][nz-2];
     mBx[x][y][jj]=mBx[x][y][nz-3];
     mBy[x][y][jj]=mBy[x][y][nz-3];
     mBz[x][y][jj]=mBz[x][y][nz-3];
     
     
     mU[x][y][jj]=mU[x][y][nz-3];
     
     }
     END_y_full
     END_x_full
     */
    
    // PERIODIC BOUNDARY IN Z
    /*
    BEGIN_x_full
    BEGIN_y_full
    for (int jj=nz-2;jj<nz;jj++){
        
        //              (*mro)[x][y][SIZE_Z-1]=(*mro)[x][y][SIZE_Z-3];(*mro)[x][y][SIZE_Z-2]=(*mro)[x][y][SIZE_Z-3];
        //mro[x][y][nz-1]=roStored[x][y][nz-2];
        mro[x][y][jj]=mro[x][y][jj-nz+4];
        
        mpx[x][y][jj]=mpx[x][y][jj-nz+4];
        mpy[x][y][jj]=mpy[x][y][jj-nz+4];
        mpz[x][y][jj]=mpz[x][y][jj-nz+4];
        //mU[x][y][jj]=mU[x][y][jj-nz+4];
        //              (*mBx)[x][y][SIZE_Z-1]=(*mBx)[x][y][SIZE_Z-3];(*mBx)[x][y][SIZE_Z-2]=(*mBx)[x][y][SIZE_Z-3];
        
        //mBx[x][y][nz-1]=BxStored[x][y][nz-2];
        mBx[x][y][jj]=mBx[x][y][jj-nz+4];
        mBy[x][y][jj]=mBy[x][y][jj-nz+4];
        mBz[x][y][jj]=mBz[x][y][jj-nz+4];
        
        
        mU[x][y][jj]=mU[x][y][jj-nz+4];
    }
    END_y_full
    END_x_full
    */
    //Do nothing: consider fully periodic box.
    
}

inline double &EMfields3D::getBxC(int indexX, int indexY, int indexZ) const {
    return (Bxc0[indexX][indexY][indexZ]);
}


inline double &EMfields3D::getByC(int indexX, int indexY, int indexZ) const{
    return (Byc0[indexX][indexY][indexZ]);
}


inline double &EMfields3D::getBzC(int indexX, int indexY, int indexZ) const{
 return (Bzc0[indexX][indexY][indexZ]) ;
}


inline double &EMfields3D::getpxC(int indexX, int indexY, int indexZ) const{
 return (pxc0[indexX][indexY][indexZ]) ;
}


inline double &EMfields3D::getpyC(int indexX, int indexY, int indexZ) const{
    return (pyc0[indexX][indexY][indexZ]);
}
inline double &EMfields3D::getpzC(int indexX, int indexY, int indexZ) const{
    return (pzc0[indexX][indexY][indexZ]);
    
}

inline double &EMfields3D::getRHO0(int indexX, int indexY, int indexZ) const{
        return (rhoc0[indexX][indexY][indexZ]);

}
inline double &EMfields3D::getgrid_eta(int indexX, int indexY, int indexZ) const{
 return (grid_eta[indexX][indexY][indexZ]);
}



inline double*** EMfields3D::getRHO0() {return(rhoc0);}
inline double*** EMfields3D::getgrid_eta() {return(grid_eta);}
inline double*** EMfields3D::getBxC() {return(Bxc0);}
inline double*** EMfields3D::getByC() {return(Byc0);}
inline double*** EMfields3D::getBzC() {return(Bzc0);}
inline double*** EMfields3D::getpxC() {return(pxc0);}
inline double*** EMfields3D::getpyC() {return(pyc0);}
inline double*** EMfields3D::getpzC() {return(pzc0);}
inline double*** EMfields3D::getUC() {return(Uc0);}
inline double EMfields3D::getMHDtime() {return(MHDtime);}

void EMfields3D::put_Sources_fluxes(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid)
{
    
	SET_DELTA
    int x,y,z;
	double ExBxP,ExBxC,ExBxM,ExByP,ExByC,ExByM,ExBzP,ExBzC,ExBzM;
	double ***Ex_temp = newArr3(double,nx,ny,nz);
	double ***Ey_temp = newArr3(double,nx,ny,nz);
	double ***Ez_temp = newArr3(double,nx,ny,nz);
    //        cout << "Ex_temp: " << Ex_temp[4][7][9];
	put_resistivity(nx,ny,nz,mro,mpx,mpy,mpz,mBx,mBy,mBz,mU,grid);
    //        cout << "grid_eta: " << grid_eta[4][7][9];
	BEGIN_small
    
    ASSIGN_DELTA
    Ex_temp[x][y][z]=((grid_eta[x][y][z])*(nrotxB));//+(((C_h)/((*ro)[x][y]))*(  (rotyB)*((*Bz)[x][y]) - (rotzB)*((*By)[x][y]    )             ));
    Ey_temp[x][y][z]=((grid_eta[x][y][z])*(nrotyB));//+(((C_h)/((*ro)[x][y]))*(  (rotzB)*((*Bx)[x][y]) - (rotxB)*((*Bz)[x][y]    )             ));
    Ez_temp[x][y][z]=((grid_eta[x][y][z])*(nrotzB));//+(((C_h)/((*ro)[x][y]))*(  (rotxB)*((*By)[x][y]) - (rotyB)*((*Bx)[x][y]    )             ));
    END_small
    
    
    
    //  cout << "grid_eta: " << Ex_temp[4][7][9];
    //	BEGIN_small
    BEGIN_x_small2
    BEGIN_y_small2
    BEGIN_z_small2
    
    
    Q__rhoc[x][y][z]=(double)0.0;
    Q__pxc[x][y][z]=(double)0.0;
    Q__pyc[x][y][z]=(double)0.0;
    Q__pzc[x][y][z]=(double)0.0;
    
    ASSIGN_DELTA
    
    ExBxP=(((mBy[x+1][y][z])*(Ez_temp[x+1][y][z]) )- ((mBz[x+1][y][z])*(Ey_temp[x+1][y][z]) ));
    ExBxC=(((mBy[x][y][z])*(Ez_temp[x][y][z]) )- ((mBz[x][y][z])*(Ey_temp[x][y][z]) ));
    ExBxM=(((mBy[x-1][y][z])*(Ez_temp[x-1][y][z]) )- ((mBz[x-1][y][z])*(Ey_temp[x-1][y][z]) ));
    
    ExByP=(  ((mBz[x][y+1][z])*(Ex_temp[x][y+1][z]) ) - ((mBx[x][y+1][z])*(Ez_temp[x][y+1][z]) ));
    ExByC=(  ((mBz[x][y][z])*(Ex_temp[x][y][z]) ) - ((mBx[x][y][z])*(Ez_temp[x][y][z]) ));
    ExByM=(  ((mBz[x][y-1][z])*(Ex_temp[x][y-1][z]) ) - ((mBx[x][y-1][z])*(Ez_temp[x][y-1][z]) ));
    
    ExBzP=(  ((mBx[x][y][z+1])*(Ey_temp[x][y][z+1]) ) - ((mBy[x][y][z+1])*(Ex_temp[x][y][z+1]) ));
    ExBzC=(  ((mBx[x][y][z])*(Ey_temp[x][y][z]) ) - ((mBy[x][y][z])*(Ex_temp[x][y][z]) ));
    ExBzM=(  ((mBx[x][y][z-1])*(Ey_temp[x][y][z-1]) ) - ((mBy[x][y][z-1])*(Ex_temp[x][y][z-1]) ));
    
    Q__Uc[x][y][z]=(nFD(ExBxP,ExBxC,ExBxM,dxplus,dxminus))+(nFD(ExByP,ExByC,ExByM,dyplus,dyminus))+(nFD(ExBzP,ExBzC,ExBzM,dzplus,dzminus));
    Q__Bxc[x][y][z]=( ((nFDz(Ey_temp,dzplus,dzminus)))-((nFDy(Ez_temp,dyplus,dyminus))) );
    Q__Byc[x][y][z]=(nFDx(Ez_temp,dxplus,dxminus))-((nFDz(Ex_temp,dzplus,dzminus)));
    Q__Bzc[x][y][z]=((nFDy(Ex_temp,dyplus,dyminus))-(nFDx(Ey_temp,dxplus,dxminus)));
    
    END_z_small
    END_y_small
    END_x_small
    
    
    //END_small
    
    
	delArr3(Ex_temp,nx,ny);
	delArr3(Ey_temp,nx,ny);
	delArr3(Ez_temp,nx,ny);
    
}


inline void EMfields3D::put_resistivity(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid)
{
    if (Case=="PETSCHEK"){
    int x,y,z;
    double grid_eta_factor;
    //       cout << "getXN" << grid->getXN(2,4,5);       cout << "getYN" << grid->getYN(2,4,5);        cout << "getZN" << grid->getZN(2,4,5);
	BEGIN_small
    grid_eta[x][y][z]=0.0;
    if(MHDtime>0.0){
        grid_eta_factor=- (((grid->getXC(x,y,z)) - (0.5*Lx))*((grid->getXC(x,y,z)) - (0.5*Lx))/0.5)- (((grid->getYC(x,y,z)) - (3.0*Ly/4.0))*((grid->getYC(x,y,z)) - (3.0*Ly/4.0))/0.5);
        if (grid_eta_factor<-12.0)
            grid_eta[x][y][z]=0.0;
        else
            grid_eta[x][y][z]+=mod_MHD(0.02*exp(grid_eta_factor));
    }
    else  grid_eta[x][y][z]=0.0;

    END_small
    } else if (Case=="SWEET-PARKER"){
      
        
        int x,y,z;
        double grid_eta_factor;
        //       cout << "getXN" << grid->getXN(2,4,5);       cout << "getYN" << grid->getYN(2,4,5);        cout << "getZN" << grid->getZN(2,4,5);
        BEGIN_small
        grid_eta[x][y][z]=0.0;
        if(MHDtime>0.0){

            if ((grid->getYC(x,y,z))>(Ly/2.0)) grid_eta[x][y][z]=0.01;
            else grid_eta[x][y][z]=0.0;


        }
        else  grid_eta[x][y][z]=0.0;
        
        END_small
        
    } else if (Case=="MA-SWEET-PARKER"){
        
        
        int x,y,z;
        double grid_eta_factor;
        //       cout << "getXN" << grid->getXN(2,4,5);       cout << "getYN" << grid->getYN(2,4,5);        cout << "getZN" << grid->getZN(2,4,5);
        BEGIN_small
        grid_eta[x][y][z]=0.0;
        if(MHDtime>0.0){
            
            if ((grid->getYC(x,y,z))>(Ly/2.0)) grid_eta[x][y][z]=0.02;
            else grid_eta[x][y][z]=0.0;
            
            
        }
        else  grid_eta[x][y][z]=0.0;
        
        END_small
        
    }


 else if (Case=="PETSCHEKNS"){
    
    
    int x,y,z;
    double grid_eta_factor;
    //       cout << "getXN" << grid->getXN(2,4,5);       cout << "getYN" << grid->getYN(2,4,5);        cout << "getZN" << grid->getZN(2,4,5);
	BEGIN_small
    grid_eta[x][y][z]=0.0;
    if(MHDtime<20.0){
        grid_eta_factor=- (((grid->getXC(x,y,z)) - (0.5*Lx))*((grid->getXC(x,y,z)) - (0.5*Lx))/0.5)- (((grid->getYC(x,y,z)) - (3.0*Ly/4.0))*((grid->getYC(x,y,z)) - (3.0*Ly/4.0))/0.5);
        if (grid_eta_factor<-12.0)
            grid_eta[x][y][z]=0.0;
            else
                grid_eta[x][y][z]+=mod_MHD(0.04*exp(grid_eta_factor));
                }
    else  grid_eta[x][y][z]=0.0;
        
        END_small
        
        }





}


/*
 inline void EMfields3D::put_resistivity(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid)
 {
 int x,y,z;
 BEGIN_full
 grid_eta[x][y][z]=0.01;
 END_full
 }
 */



inline double EMfields3D::F_cq(double Bxd,double Byd,double Bzd,double Bqd,double rod,double gd,double Pd,double vqd)
{
	double ca2,caq2,cs2,cfq,ao,b2;
    if (!(Pd==Pd)) cout<< "in Pd NaN" <<endl;
    
	Pd=cut_MHD(Pd,0.00000001);
    
	//modification.
	ca2=(Bxd*Bxd+Byd*Byd+Bzd*Bzd)*(1.0/mu)/rod;
    // zameneny Valfven B2/(4piro)
    //	ca2=(Bx*Bx+By*By+Bz*Bz)/ro;
    
    
    
	caq2=Bqd*Bqd*(1.0/mu)/rod;
    //	caq2=Bq*Bq/ro;
	cs2=(gd*Pd)/rod;
    
	ao=cs2+ca2;
	b2=4.0*cs2*caq2;
    
	cfq=0.707106781*sqrt(ao+sqrt(ao*ao-b2));
    /*        if ((ao*ao-b2)<0) cout<<"woops... Catched (ao*ao-b2)<0 in calc_dt" <<endl;
     
     if (!(cfq==cfq)) cout<<"woops... Catched (ao*ao-b2)<0 in calc_dt" <<endl;
     
     
     if (!(Bxd==Bxd)) cout<< "Bxd NaN" <<endl;
     if (!(Byd==Byd)) cout<< "Byd NaN" <<endl;
     if (!(Bzd==Bzd)) cout<< "Bzd NaN" <<endl;
     if (!(Bqd==Bqd)) cout<< "Bqd NaN" <<endl;
     
     if (!(rod==rod)) cout<< "rod NaN" <<endl;
     if (!(gd==gd)) cout<< "gd NaN" <<endl;
     if (!(Pd==Pd)) cout<< "Pd NaN" <<endl;
     
     */
    
	return ((double)(mod_MHD(vqd)+cfq));
}


inline void EMfields3D::calc_dt(Grid *grid, VirtualTopology3D *vct)
{
	double cx,cy,cz,mcx,mcy,mcz,dt1,mdt;
    int x,y,z;
    
	mcx=(double)0.0;
	mcy=(double)0.0;
	mcz=(double)0.0;
	dt1=(double)0.0;
	mdt=(double)0.0;
    
    for (int x=2;x<3;x++) {
        for (int y=2;y<3;y++) {
            for (int z=2;z<3;z++) {
                //	    BEGIN_small
                cx=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Bxc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pxc0[x][y][z]) / rhoc0[x][y][z] )  );
                cy=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Byc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pyc0[x][y][z]) / rhoc0[x][y][z] )  );
                cz=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Bzc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pzc0[x][y][z]) / rhoc0[x][y][z] )  );
                
                dt=(min3_MHD(CN*dx/cx,CN*dy/cy,CN*dz/cz));
                //		END_small
            }
        }
 	}
    
    //    BEGIN_full
    //    BEGIN_small
    //      END_small
    //      END_full
    
    BEGIN_x_small2
    BEGIN_y_small2
    BEGIN_z_small2
    
    cx=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Bxc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pxc0[x][y][z]) / rhoc0[x][y][z] )  );
    cy=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Byc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pyc0[x][y][z]) / rhoc0[x][y][z] )  );
    cz=F_cq(Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Bzc0[x][y][z],rhoc0[x][y][z],C_g,(ppp_MHD(rhoc0[x][y][z],pxc0[x][y][z],pyc0[x][y][z],pzc0[x][y][z],Bxc0[x][y][z],Byc0[x][y][z],Bzc0[x][y][z],Uc0[x][y][z],C_g)), ((pzc0[x][y][z]) / rhoc0[x][y][z] )  );
    mdt=min3_MHD(CN*dx/cx,CN*dy/cy,CN*dz/cz);
    if (mdt<dt) dt=mdt;
    
    END_z_small
    END_y_small
    END_x_small
    
    
    //	dt=(min(CN/dx/mcx,CN/dy/mcy));
	mdt=dt;
    //	cout << "dt hyperbolic: " << (double)dt << endl;
    
    BEGIN_x_small2
    BEGIN_y_small2
    BEGIN_z_small2
    
    //	BEGIN_small
    //	if ((grid_eta[x][y][z])<0.0)
    //	{
    //		dt1=(min3_MHD(dt,min(-1.0*dx*dx/(grid_eta[x][y][z]),-1.0*dz*dz/(grid_eta[x][y][z])),-1.0*dy*dy/(grid_eta[x][y][z])));
    //		if (dt1<mdt) mdt=dt1;
    //	}
    if (grid_eta[x][y][z]>0.000001)
    {
        dt1=(min3_MHD(dt,min(1.0*dx*dx/(grid_eta[x][y][z]),1.0*dz*dz/(grid_eta[x][y][z])),1.0*dy*dy/(grid_eta[x][y][z])));
        if (dt1<mdt) mdt=dt1;
    }
    //    END_small
    END_z_small
    END_y_small
    END_x_small
    
    
    //        cout << "dt hyp+parabolic: " << (double)mdt << endl;
 	dt=min(dt,mdt);
    // MPI_Barrier(MPI_COMM_WORLD);
    //MPI_Allreduce(&dt,&mdt,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
    mdt = dt;
    dt=mdt;
    //        cout << "min(dt) over all processors: " << (double)dt << endl;
}

void EMfields3D::Fix_P(int nx, int ny, int nz, double ***mro,double ***mpx,double ***mpy,double ***mpz,double ***mBx,double ***mBy,double ***mBz,double ***mU, Grid *grid)
{
 /*
    double pressure_temp;
    int x,y,z,order;
    
    order = 0;
    BEGIN_full
    if (!(mro[x][y][z]==mro[x][y][z])) { cout << "Fix_P begin 5 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(mpx[x][y][z]==mpx[x][y][z]))   { cout << "Fix_P begin 5 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mpy[x][y][z]==mpy[x][y][z]))   { cout << "Fix_P begin 5 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(mpz[x][y][z]==mpz[x][y][z]))   { cout << "Fix_P begin 5 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(mBx[x][y][z]==mBx[x][y][z]))  {cout << "Fix_P begin 5 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mBy[x][y][z]==mBy[x][y][z]))  {cout << "Fix_P begin 5 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mBz[x][y][z]==mBz[x][y][z]))  {cout << "Fix_P begin 5 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mU[x][y][z]==mU[x][y][z]))  { cout << "Fix_P begin 5 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    
    
    BEGIN_full
    //   BEGIN_x_small
    //    BEGIN_y_small
    //     BEGIN_z_small
    pressure_temp=ppp_MHD((mro[x][y][z]),(mpx[x][y][z]),(mpy[x][y][z]),(mpz[x][y][z]),(mBx[x][y][z]),(mBy[x][y][z]), (mBz[x][y][z]),(mU[x][y][z]),C_g);
    if (!(pressure_temp==pressure_temp))  { cout << "Fix_P pressure catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    
    
    //((mro[x][y][z]),(mpx[x][y][z]),(mpy[x][y][z]),(mpz[x][y][z]),(mBx[x][y][z]),(mBy[x][y][z]), (mBz[x][y][z]),(mU[x][y][z]),C_g)
	if (pressure_temp<0.0001) pressure_temp=.0001;
	(mU[x][y][z])=(pressure_temp/(C_g-1.0)) \
    +(0.5)*((((mpx[x][y][z])*(mpx[x][y][z]))+((mpy[x][y][z])*(mpy[x][y][z]))+((mpz[x][y][z])*(mpz[x][y][z])))/(mro[x][y][z])) \
    +(0.5)*(((mBx[x][y][z])*(mBx[x][y][z]))+((mBy[x][y][z])*(mBy[x][y][z]))+((mBz[x][y][z])*(mBz[x][y][z])));
    
    if (!(mU[x][y][z]==mU[x][y][z]))  { cout << "mU  catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1 <<" order: "<<order<<"  \n";stopFlag=-1;   }
    
    //     END_z_small
    //    END_y_small
    //  END_x_small
    
    END_full
    
    BEGIN_full
    if (!(mro[x][y][z]==mro[x][y][z])) { cout << "Fix_P begin 6 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(mpx[x][y][z]==mpx[x][y][z]))   { cout << "Fix_P begin 6 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mpy[x][y][z]==mpy[x][y][z]))   { cout << "Fix_P begin 6 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(mpz[x][y][z]==mpz[x][y][z]))   { cout << "Fix_P begin 6 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(mBx[x][y][z]==mBx[x][y][z]))  {cout << "Fix_P begin 6 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mBy[x][y][z]==mBy[x][y][z]))  {cout << "Fix_P begin 6 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mBz[x][y][z]==mBz[x][y][z]))  {cout << "Fix_P begin 6 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(mU[x][y][z]==mU[x][y][z]))  { cout<< mU[x][y][z] << "Fix_P begin 6 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<1<<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    */
    
    
}





int EMfields3D::calc_Lx(int order,Grid *grid, VirtualTopology3D *vct)
{
    
    
	double roL1,pxL1,pyL1,pzL1,UL1,BxL1,ByL1,BzL1,vxL1,vyL1,vzL1,jxL1,jyL1,jzL1,PL1,gPxL1,gPyL1,ExL1,EyL1,EzL1;
	double roR1,pxR1,pyR1,pzR1,UR1,BxR1,ByR1,BzR1,vxR1,vyR1,vzR1,jxR1,jyR1,jzR1,PR1,gPxR1,gPyR1,ExR1,EyR1,EzR1;
	double roL2,pxL2,pyL2,pzL2,UL2,BxL2,ByL2,BzL2,vxL2,vyL2,vzL2,jxL2,jyL2,jzL2,PL2,gPxL2,gPyL2,ExL2,EyL2,EzL2;
	double roR2,pxR2,pyR2,pzR2,UR2,BxR2,ByR2,BzR2,vxR2,vyR2,vzR2,jxR2,jyR2,jzR2,PR2,gPxR2,gPyR2,ExR2,EyR2,EzR2;
    
	double GroL1,GpxL1,GpyL1,GpzL1,GUL1,GBxL1,GByL1,GBzL1;
	double GroR1,GpxR1,GpyR1,GpzR1,GUR1,GBxR1,GByR1,GBzR1;
	double GroL2,GpxL2,GpyL2,GpzL2,GUL2,GBxL2,GByL2,GBzL2;
	double GroR2,GpxR2,GpyR2,GpzR2,GUR2,GBxR2,GByR2,GBzR2;
    
	double c1,c2;
	double renormL,renormR;
    int x,y,z;
	int result;
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "calc _Lx" << endl;
    
    
    
    //        cout << "test1" << endl;
    
    /*
     cout << "Calc_Lx: Bx at all y, x=5,z=7";
     BEGIN_y_full
     cout << Bxc[5][y][7] << "   ";
     END_y_full
     cout << endl;
     */
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc0[x][y][z]==rhoc0[x][y][z])) { cout << "Calc_Lx begin 0 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc0[x][y][z]==pxc0[x][y][z]))   { cout << "Calc_Lx begin 0 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc0[x][y][z]==pyc0[x][y][z]))   { cout << "Calc_Lx begin 0 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc0[x][y][z]==pzc0[x][y][z]))   { cout << "Calc_Lx begin 0 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc0[x][y][z]==Bxc0[x][y][z]))  {cout << "Calc_Lx begin 0 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc0[x][y][z]==Byc0[x][y][z]))  {cout << "Calc_Lx begin 0 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc0[x][y][z]==Bzc0[x][y][z]))  {cout << "Calc_Lx begin 0 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc0[x][y][z]==Uc0[x][y][z]))  { cout << "Calc_Lx begin 0 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    eq(rhoc,rhoc0,nxc,nyc,nzc);
    eq(pxc,pxc0,nxc,nyc,nzc);
    eq(pyc,pyc0,nxc,nyc,nzc);
    eq(pzc,pzc0,nxc,nyc,nzc);
    eq(Uc,Uc0,nxc,nyc,nzc);
    eq(Bxc,Bxc0,nxc,nyc,nzc);
    eq(Byc,Byc0,nxc,nyc,nzc);
    eq(Bzc,Bzc0,nxc,nyc,nzc);
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lx begin 1 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lx begin 1 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lx begin 1 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lx begin 1 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lx begin 1 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lx begin 1 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lx begin 1 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lx begin 1 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    
    
 	put_Sources_fluxes(nxc,nyc,nzc,rhoc0,pxc0,pyc0,pzc0,Bxc0,Byc0,Bzc0,Uc0,grid);
    
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Lx begin  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Lx begin  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Lx begin  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
    
    
    //        cout << "test3" << endl;
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "Sources fluxes Lx done. " << endl;
    
    //        cout << "test4" << endl;
    
    eqValue(0.0, dUc, nxc, nyc, nzc);
    eqValue(0.0, drhoc, nxc, nyc, nzc);
    eqValue(0.0, dpxc, nxc, nyc, nzc);
    eqValue(0.0, dpyc, nxc, nyc, nzc);
    eqValue(0.0, dpzc, nxc, nyc, nzc);
    eqValue(0.0, dBxc, nxc, nyc, nzc);
    eqValue(0.0, dByc, nxc, nyc, nzc);
    eqValue(0.0, dBzc, nxc, nyc, nzc);
    //        cout << "test5" << endl;
    
	BEGIN_small
    
	
    /*		(*dU)[x][y]=F_dWq((*U)[x-1][y],(*U)[x][y],(*U)[x+1][y]);
     (*dro)[x][y]=F_dWq((*ro)[x-1][y],(*ro)[x][y],(*ro)[x+1][y]);
     (*dpx)[x][y]=F_dWq((*px)[x-1][y],(*px)[x][y],(*px)[x+1][y]);
     (*dpy)[x][y]=F_dWq((*py)[x-1][y],(*py)[x][y],(*py)[x+1][y]);
     (*dpz)[x][y]=F_dWq((*pz)[x-1][y],(*pz)[x][y],(*pz)[x+1][y]);
     (*dBx)[x][y]=F_dWq((*Bx)[x-1][y],(*Bx)[x][y],(*Bx)[x+1][y]);
     (*dBy)[x][y]=F_dWq((*By)[x-1][y],(*By)[x][y],(*By)[x+1][y]);
     (*dBz)[x][y]=F_dWq((*Bz)[x-1][y],(*Bz)[x][y],(*Bz)[x+1][y]);
     */
    //renormL=(two)*( ((*grid_dx)[x-1][y][z])/( ((*grid_dx)[x-1][y][z])+((*grid_dx)[x][y][z]))  );
    //renormR=(two)*( ((*grid_dx)[x+1][y][z])/( ((*grid_dx)[x+1][y][z])+((*grid_dx)[x][y][z]))  );
    renormL=(double)1.0;
    renormR=(double)1.0;
    
    
    dUc[x][y][z]=F_dWq(((Uc[x][y][z]-Uc[x-1][y][z])*(renormL)),(( Uc[x+1][y][z]-Uc[x][y][z] )*(renormR)));
    drhoc[x][y][z]=F_dWq(((rhoc[x][y][z]-rhoc[x-1][y][z])*(renormL)),(( rhoc[x+1][y][z]-rhoc[x][y][z] )*(renormR)));
    dpxc[x][y][z]=F_dWq(((pxc[x][y][z]-pxc[x-1][y][z])*(renormL)),(( pxc[x+1][y][z]-pxc[x][y][z] )*(renormR)));
    dpyc[x][y][z]=F_dWq(((pyc[x][y][z]-pyc[x-1][y][z])*(renormL)),(( pyc[x+1][y][z]-pyc[x][y][z] )*(renormR)));
    dpzc[x][y][z]=F_dWq(((pzc[x][y][z]-pzc[x-1][y][z])*(renormL)),(( pzc[x+1][y][z]-pzc[x][y][z] )*(renormR)));
    dBxc[x][y][z]=F_dWq(((Bxc[x][y][z]-Bxc[x-1][y][z])*(renormL)),(( Bxc[x+1][y][z]-Bxc[x][y][z] )*(renormR)));
    dByc[x][y][z]=F_dWq(((Byc[x][y][z]-Byc[x-1][y][z])*(renormL)),(( Byc[x+1][y][z]-Byc[x][y][z] )*(renormR)));
    dBzc[x][y][z]=F_dWq(((Bzc[x][y][z]-Bzc[x-1][y][z])*(renormL)),(( Bzc[x+1][y][z]-Bzc[x][y][z] )*(renormR)));
    
    END_small
    
    //        cout << "test6" << endl;
    
	BEGIN_small
    
    roR1=rhoc[x][y][z]-0.5*drhoc[x][y][z];
    roL2=rhoc[x][y][z]+0.5*drhoc[x][y][z];
    
    pxR1=pxc[x][y][z]-0.5*dpxc[x][y][z];
    pxL2=pxc[x][y][z]+0.5*dpxc[x][y][z];
    pyR1=pyc[x][y][z]-0.5*dpyc[x][y][z];
    pyL2=pyc[x][y][z]+0.5*dpyc[x][y][z];
    pzR1=pzc[x][y][z]-0.5*dpzc[x][y][z];
    pzL2=pzc[x][y][z]+0.5*dpzc[x][y][z];
    UR1=Uc[x][y][z]-0.5*dUc[x][y][z];
    UL2=Uc[x][y][z]+0.5*dUc[x][y][z];
    BxR1=Bxc[x][y][z]-0.5*dBxc[x][y][z];
    BxL2=Bxc[x][y][z]+0.5*dBxc[x][y][z];
    ByR1=Byc[x][y][z]-0.5*dByc[x][y][z];
    ByL2=Byc[x][y][z]+0.5*dByc[x][y][z];
    BzR1=Bzc[x][y][z]-0.5*dBzc[x][y][z];
    BzL2=Bzc[x][y][z]+0.5*dBzc[x][y][z];
    
    GroR1=Gmx_ro(pxR1);
    GroL2=Gmx_ro(pxL2);
    GpxR1=Gmx_px(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpxL2=Gmx_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    
    GpyR1=Gmx_py(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpyL2=Gmx_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR1=Gmx_pz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpzL2=Gmx_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR1=Gxm_U(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GUL2=Gxm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR1=Gmx_Bx(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBxL2=Gmx_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR1=Gmx_By(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GByL2=Gmx_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR1=Gmx_Bz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBzL2=Gmx_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    
    // If Sources==2...
    rhoc1[x][y][z]=F_W1(rhoc[x][y][z],dt,invdx,GroR1,GroL2,Q__rhoc[x][y][z]);
    pxc1[x][y][z]=F_W1(pxc[x][y][z],dt,invdx,GpxR1,GpxL2,Q__pxc[x][y][z]);
    pyc1[x][y][z]=F_W1(pyc[x][y][z],dt,invdx,GpyR1,GpyL2,Q__pyc[x][y][z]);
    pzc1[x][y][z]=F_W1(pzc[x][y][z],dt,invdx,GpzR1,GpzL2,Q__pzc[x][y][z]);
    Uc1[x][y][z]=F_W1(Uc[x][y][z],dt,invdx,GUR1,GUL2,Q__Uc[x][y][z]);
    Bxc1[x][y][z]=F_W1(Bxc[x][y][z],dt,invdx,GBxR1,GBxL2,Q__Bxc[x][y][z]);
    Byc1[x][y][z]=F_W1(Byc[x][y][z],dt,invdx,GByR1,GByL2,Q__Byc[x][y][z]);
    Bzc1[x][y][z]=F_W1(Bzc[x][y][z],dt,invdx,GBzR1,GBzL2,Q__Bzc[x][y][z]);
    //endif
    
    END_small
    
    // pridelyvaem boundary k gotovomu resheniyu.
    
    
    // MPI_Barrier(MPI_COMM_WORLD);
    
    communicateCenterMHD(nxc,nyc,nzc,rhoc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Uc1,vct);
    // MPI_Barrier(MPI_COMM_WORLD);
    
    Fix_P(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    Boundarize(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,vct);
    
    eq(rhoc,rhoc1,nxc,nyc,nzc);
    eq(pxc,pxc1,nxc,nyc,nzc);
    eq(pyc,pyc1,nxc,nyc,nzc);
    eq(pzc,pzc1,nxc,nyc,nzc);
    eq(Uc,Uc1,nxc,nyc,nzc);
    eq(Bxc,Bxc1,nxc,nyc,nzc);
    eq(Byc,Byc1,nxc,nyc,nzc);
    eq(Bzc,Bzc1,nxc,nyc,nzc);
    
    //	ro=ro1;
    //	px=px1;
    //	py=py1;
    //	pz=pz1;
    //	U =U1;
    //	Bx=Bx1;
    //	By=By1;
    //	Bz=Bz1;
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lx mid  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lx mid  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lx mid  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lx mid  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lx mid  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lx mid  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lx mid  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lx mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    
	put_Sources_fluxes(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Lx mid  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Lx mid  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Lx mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
    
    
	///////////////
    
	BEGIN_z_small
	BEGIN_y_small
    
    
    x=0;
    
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x+1][y][z]-0.5*drhoc[x+1][y][z];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x+1][y][z]-0.5*dpxc[x+1][y][z];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x+1][y][z]-0.5*dpyc[x+1][y][z];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x+1][y][z]-0.5*dpzc[x+1][y][z];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x+1][y][z]-0.5*dUc[x+1][y][z];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x+1][y][z]-0.5*dBxc[x+1][y][z];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x+1][y][z]-0.5*dByc[x+1][y][z];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x+1][y][z]-0.5*dBzc[x+1][y][z];
    GroL2=Gmx_ro(pxL2);
    GroR2=Gmx_ro(pxR2);
    GpxL2=Gmx_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpxR2=Gmx_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmx_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR2=Gmx_py(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpzL2=Gmx_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR2=Gmx_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2     );
    GUL2=Gxm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gxm_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2 );
    GBxL2=Gmx_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmx_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2	);
    GByL2=Gmx_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmx_By( roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2  );
    GBzL2=Gmx_Bz( roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2   );
    GBzR2=Gmx_Bz( roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2  );
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(BxL2,BxR2),avr_MHD(roL2,roR2),C_g,avr_MHD(   (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g))  ,   (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    ),avr_MHD((pxL2/roL2),(pxR2/roR2 )));
    
    BEGIN_x_small
    
    roL1=roL2;
    roR1=roR2;
    pxL1=pxL2;
    pxR1=pxR2;
    pyL1=pyL2;
    pyR1=pyR2;
    pzL1=pzL2;
    pzR1=pzR2;
    UL1=UL2;
    UR1=UR2;
    BxL1=BxL2;
    BxR1=BxR2;
    ByL1=ByL2;
    ByR1=ByR2;
    BzL1=BzL2;
    BzR1=BzR2;
    GroL1=GroL2;
    GroR1=GroR2;
    GpxL1=GpxL2;
    GpxR1=GpxR2;
    GpyL1=GpyL2;
    GpyR1=GpyR2;
    GpzL1=GpzL2;
    GpzR1=GpzR2;
    GUL1=GUL2;
    GUR1=GUR2;
    GBxL1=GBxL2;
    GBxR1=GBxR2;
    GByL1=GByL2;
    GByR1=GByR2;
    GBzL1=GBzL2;
    GBzR1=GBzR2;
    
    c1=c2;
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x+1][y][z]-0.5*drhoc[x+1][y][z];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x+1][y][z]-0.5*dpxc[x+1][y][z];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x+1][y][z]-0.5*dpyc[x+1][y][z];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x+1][y][z]-0.5*dpzc[x+1][y][z];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x+1][y][z]-0.5*dUc[x+1][y][z];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x+1][y][z]-0.5*dBxc[x+1][y][z];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x+1][y][z]-0.5*dByc[x+1][y][z];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x+1][y][z]-0.5*dBzc[x+1][y][z];
    
    GroL2=Gmx_ro(pxL2);
    GroR2=Gmx_ro(pxR2);
    GpxL2=Gmx_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpxR2=Gmx_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmx_py( roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2 );
    GpyR2=Gmx_py( roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2  );
    GpzL2=Gmx_pz( roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2  );
    GpzR2=Gmx_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2     );
    GUL2=Gxm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gxm_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2 );
    GBxL2=Gmx_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmx_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GByL2=Gmx_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmx_By( roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2  );
    GBzL2=Gmx_Bz( roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2   );
    GBzR2=Gmx_Bz( roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2  );
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(BxL2,BxR2),avr_MHD(roL2,roR2),C_g,avr_MHD(  (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g)) ,  (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    )    ,avr_MHD((pxL2/roL2),(pxR2/roR2)));
    
    //#if (SOURCES==2)
    rhoc2[x][y][z]=F_W2(rhoc0[x][y][z],dt,invdx,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,Q__rhoc[x][y][z]);
    pxc2[x][y][z]=F_W2(pxc0[x][y][z],dt,invdx,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,Q__pxc[x][y][z]);
    pyc2[x][y][z]=F_W2(pyc0[x][y][z],dt,invdx,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,Q__pyc[x][y][z]);
    pzc2[x][y][z]=F_W2(pzc0[x][y][z],dt,invdx,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,Q__pzc[x][y][z]);
    Uc2[x][y][z]=F_W2(Uc0[x][y][z],dt,invdx,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,Q__Uc[x][y][z]);
    Bxc2[x][y][z]=F_W2(Bxc0[x][y][z],dt,invdx,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,Q__Bxc[x][y][z]);
    Byc2[x][y][z]=F_W2(Byc0[x][y][z],dt,invdx,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,Q__Byc[x][y][z]);
    Bzc2[x][y][z]=F_W2(Bzc0[x][y][z],dt,invdx,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,Q__Bzc[x][y][z]);
    
    
    //#else
    /*			(*ro2)[x][y][z]=F_W2((*ro0)[x][y][z],dt,dx,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,0.0);
     (*px2)[x][y][z]=F_W2((*px0)[x][y][z],dt,dx,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,0.0);
     (*py2)[x][y][z]=F_W2((*py0)[x][y][z],dt,dx,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,0.0);
     (*pz2)[x][y][z]=F_W2((*pz0)[x][y][z],dt,dx,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,0.0);
     (*U2)[x][y][z]=F_W2((*U0)[x][y][z],dt,dx,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,0.0);
     (*Bx2)[x][y][z]=F_W2((*Bx0)[x][y][z],dt,dx,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,0.0);
     (*By2)[x][y][z]=F_W2((*By0)[x][y][z],dt,dx,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,0.0);
     (*Bz2)[x][y][z]=F_W2((*Bz0)[x][y][z],dt,dx,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,0.0);
     */
    //#endif
    
    END_x_small
	END_y_small
    END_z_small
    ///boundaries
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    // MPI_Barrier(MPI_COMM_WORLD);
    
    communicateCenterMHD(nxc,nyc,nzc,rhoc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc2,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc2,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Uc2,vct);
    // MPI_Barrier(MPI_COMM_WORLD);
    Fix_P(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,grid);
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    // ponyatno...
    /*
     BEGIN_full
     (*ro0)[x][y][z]=(*ro2)[x][y][z];
     (*px0)[x][y][z]=(*px2)[x][y][z];
     (*py0)[x][y][z]=(*py2)[x][y][z];
     (*pz0)[x][y][z]=(*pz2)[x][y][z];
     (*U0)[x][y][z]=(*U2)[x][y][z];
     (*Bx0)[x][y][z]=(*Bx2)[x][y][z];
     (*By0)[x][y][z]=(*By2)[x][y][z];
     (*Bz0)[x][y][z]=(*Bz2)[x][y][z];
     END_full
     
     ro=ro0;
     px=px0;
     py=py0;
     pz=pz0;
     U =U0;
     Bx=Bx0;
     By=By0;
     Bz=Bz0;
     */
    
    eq(rhoc,rhoc2,nxc,nyc,nzc);
    eq(pxc,pxc2,nxc,nyc,nzc);
    eq(pyc,pyc2,nxc,nyc,nzc);
    eq(pzc,pzc2,nxc,nyc,nzc);
    eq(Uc,Uc2,nxc,nyc,nzc);
    eq(Bxc,Bxc2,nxc,nyc,nzc);
    eq(Byc,Byc2,nxc,nyc,nzc);
    eq(Bzc,Bzc2,nxc,nyc,nzc);
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lx end  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lx end  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lx end  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lx end  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lx end  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lx end  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lx end  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lx end  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    eq(rhoc0,rhoc2,nxc,nyc,nzc);
    eq(pxc0,pxc2,nxc,nyc,nzc);
    eq(pyc0,pyc2,nxc,nyc,nzc);
    eq(pzc0,pzc2,nxc,nyc,nzc);
    eq(Uc0,Uc2,nxc,nyc,nzc);
    eq(Bxc0,Bxc2,nxc,nyc,nzc);
    eq(Byc0,Byc2,nxc,nyc,nzc);
    eq(Bzc0,Bzc2,nxc,nyc,nzc);
    
    
    
    /*
     if (CONSTRAIN==3)
     constrain();
     else
     if (CONSTRAIN==2&&order%3==0)
     constrain();
     else
     if (CONSTRAIN==1&&order==6)
     constrain();
     
     result=customProceed1(order,2);
     if (result)
     return result;
     
     return 0;
     
     */
    
    return 0;
    
}




int EMfields3D::calc_Ly(int order,Grid *grid, VirtualTopology3D *vct)
{
    //printf("calc _Ly");
	double roL1,pxL1,pyL1,pzL1,UL1,BxL1,ByL1,BzL1,vxL1,vyL1,vzL1,jxL1,jyL1,jzL1,PL1,gPxL1,gPyL1,ExL1,EyL1,EzL1;
	double roR1,pxR1,pyR1,pzR1,UR1,BxR1,ByR1,BzR1,vxR1,vyR1,vzR1,jxR1,jyR1,jzR1,PR1,gPxR1,gPyR1,ExR1,EyR1,EzR1;
	double roL2,pxL2,pyL2,pzL2,UL2,BxL2,ByL2,BzL2,vxL2,vyL2,vzL2,jxL2,jyL2,jzL2,PL2,gPxL2,gPyL2,ExL2,EyL2,EzL2;
	double roR2,pxR2,pyR2,pzR2,UR2,BxR2,ByR2,BzR2,vxR2,vyR2,vzR2,jxR2,jyR2,jzR2,PR2,gPxR2,gPyR2,ExR2,EyR2,EzR2;
    
	double GroL1,GpxL1,GpyL1,GpzL1,GUL1,GBxL1,GByL1,GBzL1;
	double GroR1,GpxR1,GpyR1,GpzR1,GUR1,GBxR1,GByR1,GBzR1;
	double GroL2,GpxL2,GpyL2,GpzL2,GUL2,GBxL2,GByL2,GBzL2;
	double GroR2,GpxR2,GpyR2,GpzR2,GUR2,GBxR2,GByR2,GBzR2;
    
	double c1,c2;
	double renormL,renormR;
    int x,y,z;
    
	int result;
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "calc _Ly" << endl;
    
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc0[x][y][z]==rhoc0[x][y][z])) { cout << "Calc_Ly begin 0 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc0[x][y][z]==pxc0[x][y][z]))   { cout << "Calc_Ly begin 0 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc0[x][y][z]==pyc0[x][y][z]))   { cout << "Calc_Ly begin 0 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc0[x][y][z]==pzc0[x][y][z]))   { cout << "Calc_Ly begin 0 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc0[x][y][z]==Bxc0[x][y][z]))  {cout << "Calc_Ly begin 0 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc0[x][y][z]==Byc0[x][y][z]))  {cout << "Calc_Ly begin 0 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc0[x][y][z]==Bzc0[x][y][z]))  {cout << "Calc_Ly begin 0 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc0[x][y][z]==Uc0[x][y][z]))  { cout << "Calc_Ly begin 0 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    eq(rhoc,rhoc0,nxc,nyc,nzc);
    eq(pxc,pxc0,nxc,nyc,nzc);
    eq(pyc,pyc0,nxc,nyc,nzc);
    eq(pzc,pzc0,nxc,nyc,nzc);
    eq(Uc,Uc0,nxc,nyc,nzc);
    eq(Bxc,Bxc0,nxc,nyc,nzc);
    eq(Byc,Byc0,nxc,nyc,nzc);
    eq(Bzc,Bzc0,nxc,nyc,nzc);
    
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Ly begin 1 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Ly begin 1 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Ly begin 1 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Ly begin 1 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Ly begin 1 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Ly begin 1 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Ly begin 1 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Ly begin 1 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    
    
    
 	put_Sources_fluxes(nxc,nyc,nzc,rhoc0,pxc0,pyc0,pzc0,Bxc0,Byc0,Bzc0,Uc0,grid);
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Ly begin  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Ly begin  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Ly begin  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "Sources fluxes Ly done. " << endl;
    
    eqValue(0.0, dUc, nxc, nyc, nzc);
    eqValue(0.0, drhoc, nxc, nyc, nzc);
    eqValue(0.0, dpxc, nxc, nyc, nzc);
    eqValue(0.0, dpyc, nxc, nyc, nzc);
    eqValue(0.0, dpzc, nxc, nyc, nzc);
    eqValue(0.0, dBxc, nxc, nyc, nzc);
    eqValue(0.0, dByc, nxc, nyc, nzc);
    eqValue(0.0, dBzc, nxc, nyc, nzc);
    
    
	BEGIN_small
    
    /*		(*dU)[x][y]=F_dWq((*U)[x][y-1],(*U)[x][y],(*U0)[x][y+1]);
     (*dro)[x][y]=F_dWq((*ro)[x][y-1],(*ro)[x][y],(*ro)[x][y+1]);
     (*dpx)[x][y]=F_dWq((*px)[x][y-1],(*px)[x][y],(*px)[x][y+1]);
     (*dpy)[x][y]=F_dWq((*py)[x][y-1],(*py)[x][y],(*py)[x][y+1]);
     (*dpz)[x][y]=F_dWq((*pz)[x][y-1],(*pz)[x][y],(*pz)[x][y+1]);
     (*dBy)[x][y]=F_dWq((*By)[x][y-1],(*By)[x][y],(*By)[x][y+1]);
     (*dBx)[x][y]=F_dWq((*Bx)[x][y-1],(*Bx)[x][y],(*Bx)[x][y+1]);
     (*dBz)[x][y]=F_dWq((*Bz)[x][y-1],(*Bz)[x][y],(*Bz)[x][y+1]);
     */
    //renormL=(two)*( ((*grid_dy)[x][y-1][z])/( ((*grid_dy)[x][y-1][z])+((*grid_dy)[x][y][z]))  );
    //renormR=(two)*( ((*grid_dy)[x][y+1][z])/( ((*grid_dy)[x][y+1][z])+((*grid_dy)[x][y][z]))  );
    renormL=(double)1.0;
    renormR=(double)1.0;
    
    
    dUc[x][y][z]=F_dWq(((Uc[x][y][z]-Uc[x][y-1][z])*renormL),(( Uc[x][y+1][z]-Uc[x][y][z] )*renormR));
    drhoc[x][y][z]=F_dWq(((rhoc[x][y][z]-rhoc[x][y-1][z])*renormL),(( rhoc[x][y+1][z]-rhoc[x][y][z] )*renormR));
    dpxc[x][y][z]=F_dWq(((pxc[x][y][z]-pxc[x][y-1][z])*renormL),(( pxc[x][y+1][z]-pxc[x][y][z] )*renormR));
    dpyc[x][y][z]=F_dWq(((pyc[x][y][z]-pyc[x][y-1][z])*renormL),(( pyc[x][y+1][z]-pyc[x][y][z] )*renormR));
    dpzc[x][y][z]=F_dWq(((pzc[x][y][z]-pzc[x][y-1][z])*renormL),(( pzc[x][y+1][z]-pzc[x][y][z] )*renormR));
    dBxc[x][y][z]=F_dWq(((Bxc[x][y][z]-Bxc[x][y-1][z])*renormL),(( Bxc[x][y+1][z]-Bxc[x][y][z] )*renormR));
    dByc[x][y][z]=F_dWq(((Byc[x][y][z]-Byc[x][y-1][z])*renormL),(( Byc[x][y+1][z]-Byc[x][y][z] )*renormR));
    dBzc[x][y][z]=F_dWq(((Bzc[x][y][z]-Bzc[x][y-1][z])*renormL),(( Bzc[x][y+1][z]-Bzc[x][y][z] )*renormR));
    
	END_small
    
    
	BEGIN_small
    roR1=rhoc[x][y][z]-0.5*drhoc[x][y][z];
    roL2=rhoc[x][y][z]+0.5*drhoc[x][y][z];
    pxR1=pxc[x][y][z]-0.5*dpxc[x][y][z];
    pxL2=pxc[x][y][z]+0.5*dpxc[x][y][z];
    pyR1=pyc[x][y][z]-0.5*dpyc[x][y][z];
    pyL2=pyc[x][y][z]+0.5*dpyc[x][y][z];
    pzR1=pzc[x][y][z]-0.5*dpzc[x][y][z];
    pzL2=pzc[x][y][z]+0.5*dpzc[x][y][z];
    UR1=Uc[x][y][z]-0.5*dUc[x][y][z];
    UL2=Uc[x][y][z]+0.5*dUc[x][y][z];
    BxR1=Bxc[x][y][z]-0.5*dBxc[x][y][z];
    BxL2=Bxc[x][y][z]+0.5*dBxc[x][y][z];
    ByR1=Byc[x][y][z]-0.5*dByc[x][y][z];
    ByL2=Byc[x][y][z]+0.5*dByc[x][y][z];
    BzR1=Bzc[x][y][z]-0.5*dBzc[x][y][z];
    BzL2=Bzc[x][y][z]+0.5*dBzc[x][y][z];
    GroR1=Gmy_ro(pyR1);
    GroL2=Gmy_ro(pyL2);
    GpxR1=Gmy_px(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpxL2=Gmy_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR1=Gmy_py(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpyL2=Gmy_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR1=Gmy_pz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpzL2=Gmy_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR1=Gym_U(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GUL2=Gym_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR1=Gmy_Bx(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBxL2=Gmy_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR1=Gmy_By(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GByL2=Gmy_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR1=Gmy_Bz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBzL2=Gmy_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    //#if (SOURCES==2)
    rhoc1[x][y][z]=F_W1(rhoc[x][y][z],dt,invdy,GroR1,GroL2,Q__rhoc[x][y][z]);
    pxc1[x][y][z]=F_W1(pxc[x][y][z],dt,invdy,GpxR1,GpxL2,Q__pxc[x][y][z]);
    pyc1[x][y][z]=F_W1(pyc[x][y][z],dt,invdy,GpyR1,GpyL2,Q__pyc[x][y][z]);
    pzc1[x][y][z]=F_W1(pzc[x][y][z],dt,invdy,GpzR1,GpzL2,Q__pzc[x][y][z]);
    Uc1[x][y][z]=F_W1(Uc[x][y][z],dt,invdy,GUR1,GUL2,Q__Uc[x][y][z]);
    Bxc1[x][y][z]=F_W1(Bxc[x][y][z],dt,invdy,GBxR1,GBxL2,Q__Bxc[x][y][z]);
    Byc1[x][y][z]=F_W1(Byc[x][y][z],dt,invdy,GByR1,GByL2,Q__Byc[x][y][z]);
    Bzc1[x][y][z]=F_W1(Bzc[x][y][z],dt,invdy,GBzR1,GBzL2,Q__Bzc[x][y][z]);
    //#else
    //#endif
    
    
    END_small
    // MPI_Barrier(MPI_COMM_WORLD);
    
    communicateCenterMHD(nxc,nyc,nzc,rhoc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Uc1,vct);
    // MPI_Barrier(MPI_COMM_WORLD);
    
    Fix_P(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    Boundarize(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,vct);
    
    
    
    
    eq(rhoc,rhoc1,nxc,nyc,nzc);
    eq(pxc,pxc1,nxc,nyc,nzc);
    eq(pyc,pyc1,nxc,nyc,nzc);
    eq(pzc,pzc1,nxc,nyc,nzc);
    eq(Uc,Uc1,nxc,nyc,nzc);
    eq(Bxc,Bxc1,nxc,nyc,nzc);
    eq(Byc,Byc1,nxc,nyc,nzc);
    eq(Bzc,Bzc1,nxc,nyc,nzc);
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Ly mid  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Ly mid  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Ly mid  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Ly mid  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Ly mid  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Ly mid  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Ly mid  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Ly mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    
    
	put_Sources_fluxes(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Ly mid  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Ly mid  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Ly mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
    
	BEGIN_z_small
	BEGIN_x_small
    y=0;
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x][y+1][z]-0.5*drhoc[x][y+1][z];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x][y+1][z]-0.5*dpxc[x][y+1][z];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x][y+1][z]-0.5*dpyc[x][y+1][z];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x][y+1][z]-0.5*dpzc[x][y+1][z];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x][y+1][z]-0.5*dUc[x][y+1][z];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x][y+1][z]-0.5*dBxc[x][y+1][z];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x][y+1][z]-0.5*dByc[x][y+1][z];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x][y+1][z]-0.5*dBzc[x][y+1][z];
    GroL2=Gmy_ro(pyL2);
    GroR2=Gmy_ro(pyR2);
    GpxL2=Gmy_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2 );
    GpxR2=Gmy_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmy_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR2=Gmy_py(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpzL2=Gmy_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR2=Gmy_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GUL2=Gym_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gym_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2 );
    GBxL2=Gmy_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmy_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GByL2=Gmy_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmy_By(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBzL2=Gmy_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR2=Gmy_Bz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(ByL2,ByR2),avr_MHD(roL2,roR2),C_g,avr_MHD(  (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g)) ,  (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    ) ,avr_MHD((pyL2/roL2),(pyR2/roR2)));
    
    BEGIN_y_small
    
    roL1=roL2;
    roR1=roR2;
    pxL1=pxL2;
    pxR1=pxR2;
    pyL1=pyL2;
    pyR1=pyR2;
    pzL1=pzL2;
    pzR1=pzR2;
    UL1=UL2;
    UR1=UR2;
    BxL1=BxL2;
    BxR1=BxR2;
    ByL1=ByL2;
    ByR1=ByR2;
    BzL1=BzL2;
    BzR1=BzR2;
    GroL1=GroL2;
    GroR1=GroR2;
    GpxL1=GpxL2;
    GpxR1=GpxR2;
    GpyL1=GpyL2;
    GpyR1=GpyR2;
    GpzL1=GpzL2;
    GpzR1=GpzR2;
    GUL1=GUL2;
    GUR1=GUR2;
    GBxL1=GBxL2;
    GBxR1=GBxR2;
    GByL1=GByL2;
    GByR1=GByR2;
    GBzL1=GBzL2;
    GBzR1=GBzR2;
    
    c1=c2;
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x][y+1][z]-0.5*drhoc[x][y+1][z];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x][y+1][z]-0.5*dpxc[x][y+1][z];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x][y+1][z]-0.5*dpyc[x][y+1][z];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x][y+1][z]-0.5*dpzc[x][y+1][z];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x][y+1][z]-0.5*dUc[x][y+1][z];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x][y+1][z]-0.5*dBxc[x][y+1][z];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x][y+1][z]-0.5*dByc[x][y+1][z];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x][y+1][z]-0.5*dBzc[x][y+1][z];
    GroL2=Gmy_ro(pyL2);
    GroR2=Gmy_ro(pyR2);
    GpxL2=Gmy_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpxR2=Gmy_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmy_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR2=Gmy_py(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpzL2=Gmy_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR2=Gmy_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GUL2=Gym_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gym_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBxL2=Gmy_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmy_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GByL2=Gmy_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmy_By(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBzL2=Gmy_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR2=Gmy_Bz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    
    
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(ByL2,ByR2),avr_MHD(roL2,roR2),C_g,avr_MHD(  (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g)) ,  (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    ) ,avr_MHD((pyL2/roL2),(pyR2/roR2)));
    //#if (SOURCES==)
    rhoc2[x][y][z]=F_W2(rhoc0[x][y][z],dt,invdy,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,Q__rhoc[x][y][z]);
    pxc2[x][y][z]=F_W2(pxc0[x][y][z],dt,invdy,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,Q__pxc[x][y][z]);
    pyc2[x][y][z]=F_W2(pyc0[x][y][z],dt,invdy,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,Q__pyc[x][y][z]);
    pzc2[x][y][z]=F_W2(pzc0[x][y][z],dt,invdy,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,Q__pzc[x][y][z]);
    Uc2[x][y][z]=F_W2(Uc0[x][y][z],dt,invdy,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,Q__Uc[x][y][z]);
    Bxc2[x][y][z]=F_W2(Bxc0[x][y][z],dt,invdy,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,Q__Bxc[x][y][z]);
    Byc2[x][y][z]=F_W2(Byc0[x][y][z],dt,invdy,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,Q__Byc[x][y][z]);
    Bzc2[x][y][z]=F_W2(Bzc0[x][y][z],dt,invdy,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,Q__Bzc[x][y][z]);
    //#else
    /*			(*ro2)[x][y][z]=F_W2((*ro0)[x][y][z],dt,dy,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,0.0);
     (*px2)[x][y][z]=F_W2((*px0)[x][y][z],dt,dy,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,0.0);
     (*py2)[x][y][z]=F_W2((*py0)[x][y][z],dt,dy,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,0.0);
     (*pz2)[x][y][z]=F_W2((*pz0)[x][y][z],dt,dy,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,0.0);
     (*U2)[x][y][z]=F_W2((*U0)[x][y][z],dt,dy,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,0.0);
     (*Bx2)[x][y][z]=F_W2((*Bx0)[x][y][z],dt,dy,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,0.0);
     (*By2)[x][y][z]=F_W2((*By0)[x][y][z],dt,dy,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,0.0);
     (*Bz2)[x][y][z]=F_W2((*Bz0)[x][y][z],dt,dy,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,0.0);
     */
    //#endif
    
    END_y_small
	END_x_small
    END_z_small
    //boundaries
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    // MPI_Barrier(MPI_COMM_WORLD);
    communicateCenterMHD(nxc,nyc,nzc,rhoc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc2,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc2,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Uc2,vct);
    //        cout << "test13" << endl;
    
    // MPI_Barrier(MPI_COMM_WORLD);
    
    Fix_P(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,grid);
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    eq(rhoc,rhoc2,nxc,nyc,nzc);
    eq(pxc,pxc2,nxc,nyc,nzc);
    eq(pyc,pyc2,nxc,nyc,nzc);
    eq(pzc,pzc2,nxc,nyc,nzc);
    eq(Uc,Uc2,nxc,nyc,nzc);
    eq(Bxc,Bxc2,nxc,nyc,nzc);
    eq(Byc,Byc2,nxc,nyc,nzc);
    eq(Bzc,Bzc2,nxc,nyc,nzc);
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Ly end  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Ly end  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Ly end  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Ly end  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Ly end  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Ly end  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Ly end  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Ly end  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    eq(rhoc0,rhoc2,nxc,nyc,nzc);
    eq(pxc0,pxc2,nxc,nyc,nzc);
    eq(pyc0,pyc2,nxc,nyc,nzc);
    eq(pzc0,pzc2,nxc,nyc,nzc);
    eq(Uc0,Uc2,nxc,nyc,nzc);
    eq(Bxc0,Bxc2,nxc,nyc,nzc);
    eq(Byc0,Byc2,nxc,nyc,nzc);
    eq(Bzc0,Bzc2,nxc,nyc,nzc);
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc0[x][y][z]==rhoc0[x][y][z])) { cout << "Calc_Ly end 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc0[x][y][z]==pxc0[x][y][z]))   { cout << "Calc_Ly end 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc0[x][y][z]==pyc0[x][y][z]))   { cout << "Calc_Ly end 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc0[x][y][z]==pzc0[x][y][z]))   { cout << "Calc_Ly end 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc0[x][y][z]==Bxc0[x][y][z]))  {cout << "Calc_Ly end 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc0[x][y][z]==Byc0[x][y][z]))  {cout << "Calc_Ly end 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc0[x][y][z]==Bzc0[x][y][z]))  {cout << "Calc_Ly end 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc0[x][y][z]==Uc0[x][y][z]))  { cout << "Calc_Ly end 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    /*
     
     if (CONSTRAIN==3)
     constrain();
     else
     if (CONSTRAIN==2&&order%3==0)
     constrain();
     else
     if (CONSTRAIN==1&&order==6)
     constrain();
     
     //	saveNameNumber("Byc",(order*1000)+k,By2);
     
     result=customProceed1(order,1);
     if (result)
     return result;
     
     return 0;
     */
    return 0;
}


int EMfields3D::calc_Lz(int order,Grid *grid, VirtualTopology3D *vct)
{
    
	double roL1,pxL1,pyL1,pzL1,UL1,BxL1,ByL1,BzL1,vxL1,vyL1,vzL1,jxL1,jyL1,jzL1,PL1,gPxL1,gPyL1,ExL1,EyL1,EzL1;
	double roR1,pxR1,pyR1,pzR1,UR1,BxR1,ByR1,BzR1,vxR1,vyR1,vzR1,jxR1,jyR1,jzR1,PR1,gPxR1,gPyR1,ExR1,EyR1,EzR1;
	double roL2,pxL2,pyL2,pzL2,UL2,BxL2,ByL2,BzL2,vxL2,vyL2,vzL2,jxL2,jyL2,jzL2,PL2,gPxL2,gPyL2,ExL2,EyL2,EzL2;
	double roR2,pxR2,pyR2,pzR2,UR2,BxR2,ByR2,BzR2,vxR2,vyR2,vzR2,jxR2,jyR2,jzR2,PR2,gPxR2,gPyR2,ExR2,EyR2,EzR2;
    
	double GroL1,GpxL1,GpyL1,GpzL1,GUL1,GBxL1,GByL1,GBzL1;
	double GroR1,GpxR1,GpyR1,GpzR1,GUR1,GBxR1,GByR1,GBzR1;
	double GroL2,GpxL2,GpyL2,GpzL2,GUL2,GBxL2,GByL2,GBzL2;
	double GroR2,GpxR2,GpyR2,GpzR2,GUR2,GBxR2,GByR2,GBzR2;
    
	double c1,c2;
    
	double renormL,renormR;
    int x,y,z;
	int result;
    
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "calc _Lz" << endl;
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc0[x][y][z]==rhoc0[x][y][z])) { cout << "Calc_Lz begin 0 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc0[x][y][z]==pxc0[x][y][z]))   { cout << "Calc_Lz begin 0 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc0[x][y][z]==pyc0[x][y][z]))   { cout << "Calc_Lz begin 0 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc0[x][y][z]==pzc0[x][y][z]))   { cout << "Calc_Lz begin 0 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc0[x][y][z]==Bxc0[x][y][z]))  {cout << "Calc_Lz begin 0 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc0[x][y][z]==Byc0[x][y][z]))  {cout << "Calc_Lz begin 0 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc0[x][y][z]==Bzc0[x][y][z]))  {cout << "Calc_Lz begin 0 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc0[x][y][z]==Uc0[x][y][z]))  { cout << "Calc_Lz begin 0 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    eq(rhoc,rhoc0,nxc,nyc,nzc);
    eq(pxc,pxc0,nxc,nyc,nzc);
    eq(pyc,pyc0,nxc,nyc,nzc);
    eq(pzc,pzc0,nxc,nyc,nzc);
    eq(Uc,Uc0,nxc,nyc,nzc);
    eq(Bxc,Bxc0,nxc,nyc,nzc);
    eq(Byc,Byc0,nxc,nyc,nzc);
    eq(Bzc,Bzc0,nxc,nyc,nzc);
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lz begin 1 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lz begin 1 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lz begin 1 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lz begin 1 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lz begin 1 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lz begin 1 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lz begin 1 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lz begin 1 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
        
 	put_Sources_fluxes(nxc,nyc,nzc,rhoc0,pxc0,pyc0,pzc0,Bxc0,Byc0,Bzc0,Uc0,grid);
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Lz begin  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Lz begin  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Lz begin  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
    //if (vct->getCartesian_rank() ==0)
    //	cout << "Sources fluxes Lz done. " << endl;
    
    eqValue(0.0, dUc, nxc, nyc, nzc);
    eqValue(0.0, drhoc, nxc, nyc, nzc);
    eqValue(0.0, dpxc, nxc, nyc, nzc);
    eqValue(0.0, dpyc, nxc, nyc, nzc);
    eqValue(0.0, dpzc, nxc, nyc, nzc);
    eqValue(0.0, dBxc, nxc, nyc, nzc);
    eqValue(0.0, dByc, nxc, nyc, nzc);
    eqValue(0.0, dBzc, nxc, nyc, nzc);
    
	BEGIN_small
    
    /*		(*dU)[x][y]=F_dWq((*U)[x][y-1],(*U)[x][y],(*U0)[x][y+1]);
     (*dro)[x][y]=F_dWq((*ro)[x][y-1],(*ro)[x][y],(*ro)[x][y+1]);
     (*dpx)[x][y]=F_dWq((*px)[x][y-1],(*px)[x][y],(*px)[x][y+1]);
     (*dpy)[x][y]=F_dWq((*py)[x][y-1],(*py)[x][y],(*py)[x][y+1]);
     (*dpz)[x][y]=F_dWq((*pz)[x][y-1],(*pz)[x][y],(*pz)[x][y+1]);
     (*dBy)[x][y]=F_dWq((*By)[x][y-1],(*By)[x][y],(*By)[x][y+1]);
     (*dBx)[x][y]=F_dWq((*Bx)[x][y-1],(*Bx)[x][y],(*Bx)[x][y+1]);
     (*dBz)[x][y]=F_dWq((*Bz)[x][y-1],(*Bz)[x][y],(*Bz)[x][y+1]);
     */
    //renormL=(two)*( ((*grid_dz)[x][y][z-1])/( ((*grid_dz)[x][y][z-1])+((*grid_dz)[x][y][z]))  );
    //renormR=(two)*( ((*grid_dz)[x][y][z+1])/( ((*grid_dz)[x][y][z+1])+((*grid_dz)[x][y][z]))  );
    renormL=(double)1.0;
    renormR=(double)1.0;
    
    
    dUc[x][y][z]=F_dWq(((Uc[x][y][z]-Uc[x][y][z-1])*renormL),(( Uc[x][y][z+1]-Uc[x][y][z] )*renormR));
    drhoc[x][y][z]=F_dWq(((rhoc[x][y][z]-rhoc[x][y][z-1])*renormL),(( rhoc[x][y][z+1]-rhoc[x][y][z] )*renormR));
    dpxc[x][y][z]=F_dWq(((pxc[x][y][z]-pxc[x][y][z-1])*renormL),(( pxc[x][y][z+1]-pxc[x][y][z] )*renormR));
    dpyc[x][y][z]=F_dWq(((pyc[x][y][z]-pyc[x][y][z-1])*renormL),(( pyc[x][y][z+1]-pyc[x][y][z] )*renormR));
    dpzc[x][y][z]=F_dWq(((pzc[x][y][z]-pzc[x][y][z-1])*renormL),(( pzc[x][y][z+1]-pzc[x][y][z] )*renormR));
    dBxc[x][y][z]=F_dWq(((Bxc[x][y][z]-Bxc[x][y][z-1])*renormL),(( Bxc[x][y][z+1]-Bxc[x][y][z] )*renormR));
    dByc[x][y][z]=F_dWq(((Byc[x][y][z]-Byc[x][y][z-1])*renormL),(( Byc[x][y][z+1]-Byc[x][y][z] )*renormR));
    dBzc[x][y][z]=F_dWq(((Bzc[x][y][z]-Bzc[x][y][z-1])*renormL),(( Bzc[x][y][z+1]-Bzc[x][y][z] )*renormR));
    
	END_small
    
    
    
	BEGIN_small
    roR1=rhoc[x][y][z]-0.5*drhoc[x][y][z];
    roL2=rhoc[x][y][z]+0.5*drhoc[x][y][z];
    pxR1=pxc[x][y][z]-0.5*dpxc[x][y][z];
    pxL2=pxc[x][y][z]+0.5*dpxc[x][y][z];
    pyR1=pyc[x][y][z]-0.5*dpyc[x][y][z];
    pyL2=pyc[x][y][z]+0.5*dpyc[x][y][z];
    pzR1=pzc[x][y][z]-0.5*dpzc[x][y][z];
    pzL2=pzc[x][y][z]+0.5*dpzc[x][y][z];
    UR1=Uc[x][y][z]-0.5*dUc[x][y][z];
    UL2=Uc[x][y][z]+0.5*dUc[x][y][z];
    BxR1=Bxc[x][y][z]-0.5*dBxc[x][y][z];
    BxL2=Bxc[x][y][z]+0.5*dBxc[x][y][z];
    ByR1=Byc[x][y][z]-0.5*dByc[x][y][z];
    ByL2=Byc[x][y][z]+0.5*dByc[x][y][z];
    BzR1=Bzc[x][y][z]-0.5*dBzc[x][y][z];
    BzL2=Bzc[x][y][z]+0.5*dBzc[x][y][z];
    GroR1=Gmz_ro(pzR1);
    GroL2=Gmz_ro(pzL2);
    GpxR1=Gmz_px(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpxL2=Gmz_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR1=Gmz_py(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpyL2=Gmz_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR1=Gmz_pz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GpzL2=Gmz_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR1=Gzm_U(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GUL2=Gzm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR1=Gmz_Bx(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBxL2=Gmz_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR1=Gmz_By(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GByL2=Gmz_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR1=Gmz_Bz(roR1,pxR1,pyR1,pzR1,C_g,BxR1,ByR1,BzR1,UR1);
    GBzL2=Gmz_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    
    //#if (SOURCES==2)
    rhoc1[x][y][z]=F_W1(rhoc[x][y][z],dt,invdz,GroR1,GroL2,Q__rhoc[x][y][z]);
    pxc1[x][y][z]=F_W1(pxc[x][y][z],dt,invdz,GpxR1,GpxL2,Q__pxc[x][y][z]);
    pyc1[x][y][z]=F_W1(pyc[x][y][z],dt,invdz,GpyR1,GpyL2,Q__pyc[x][y][z]);
    pzc1[x][y][z]=F_W1(pzc[x][y][z],dt,invdz,GpzR1,GpzL2,Q__pzc[x][y][z]);
    Uc1[x][y][z]=F_W1(Uc[x][y][z],dt,invdz,GUR1,GUL2,Q__Uc[x][y][z]);
    Bxc1[x][y][z]=F_W1(Bxc[x][y][z],dt,invdz,GBxR1,GBxL2,Q__Bxc[x][y][z]);
    Byc1[x][y][z]=F_W1(Byc[x][y][z],dt,invdz,GByR1,GByL2,Q__Byc[x][y][z]);
    Bzc1[x][y][z]=F_W1(Bzc[x][y][z],dt,invdz,GBzR1,GBzL2,Q__Bzc[x][y][z]);
    //#else
    
    
    END_small
    
    // MPI_Barrier(MPI_COMM_WORLD);
    
    communicateCenterMHD(nxc,nyc,nzc,rhoc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc1,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc1,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Uc1,vct);
    
    // MPI_Barrier(MPI_COMM_WORLD);
    Fix_P(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    Boundarize(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,vct);
    
    eq(rhoc,rhoc1,nxc,nyc,nzc);
    eq(pxc,pxc1,nxc,nyc,nzc);
    eq(pyc,pyc1,nxc,nyc,nzc);
    eq(pzc,pzc1,nxc,nyc,nzc);
    eq(Uc,Uc1,nxc,nyc,nzc);
    eq(Bxc,Bxc1,nxc,nyc,nzc);
    eq(Byc,Byc1,nxc,nyc,nzc);
    eq(Bzc,Bzc1,nxc,nyc,nzc);
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lz mid  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lz mid  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lz mid  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lz mid  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lz mid  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lz mid  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lz mid  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lz mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
	put_Sources_fluxes(nxc,nyc,nzc,rhoc1,pxc1,pyc1,pzc1,Bxc1,Byc1,Bzc1,Uc1,grid);
    
    // Check for isnan here.
    BEGIN_full
    if (!(Q__rhoc[x][y][z]==Q__rhoc[x][y][z])) { cout << "Calc_Lz mid  NaN Q__rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pxc[x][y][z]==Q__pxc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pyc[x][y][z]==Q__pyc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__pzc[x][y][z]==Q__pzc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    
    if (!(Q__Bxc[x][y][z]==Q__Bxc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Byc[x][y][z]==Q__Byc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Bzc[x][y][z]==Q__Bzc[x][y][z]))   { cout << "Calc_Lz mid  NaN Q__Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(Q__Uc[x][y][z]==Q__Uc[x][y][z]))     { cout << "Calc_Lz mid  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    END_full
    
	BEGIN_x_small
	BEGIN_y_small
    
    z=0;
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x][y][z+1]-0.5*drhoc[x][y][z+1];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x][y][z+1]-0.5*dpxc[x][y][z+1];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x][y][z+1]-0.5*dpyc[x][y][z+1];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x][y][z+1]-0.5*dpzc[x][y][z+1];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x][y][z+1]-0.5*dUc[x][y][z+1];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x][y][z+1]-0.5*dBxc[x][y][z+1];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x][y][z+1]-0.5*dByc[x][y][z+1];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x][y][z+1]-0.5*dBzc[x][y][z+1];
    GroL2=Gmz_ro(pzL2);
    GroR2=Gmz_ro(pzR2);
    GpxL2=Gmz_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2 );
    GpxR2=Gmz_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmz_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR2=Gmz_py(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpzL2=Gmz_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR2=Gmz_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GUL2=Gzm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gzm_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2 );
    GBxL2=Gmz_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmz_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GByL2=Gmz_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmz_By(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBzL2=Gmz_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR2=Gmz_Bz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(BzL2,BzR2),avr_MHD(roL2,roR2),C_g,avr_MHD(  (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g)) ,  (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    ) ,avr_MHD((pzL2/roL2),(pzR2/roR2)));
    
    BEGIN_z_small
    
    roL1=roL2;
    roR1=roR2;
    pxL1=pxL2;
    pxR1=pxR2;
    pyL1=pyL2;
    pyR1=pyR2;
    pzL1=pzL2;
    pzR1=pzR2;
    UL1=UL2;
    UR1=UR2;
    BxL1=BxL2;
    BxR1=BxR2;
    ByL1=ByL2;
    ByR1=ByR2;
    BzL1=BzL2;
    BzR1=BzR2;
    GroL1=GroL2;
    GroR1=GroR2;
    GpxL1=GpxL2;
    GpxR1=GpxR2;
    GpyL1=GpyL2;
    GpyR1=GpyR2;
    GpzL1=GpzL2;
    GpzR1=GpzR2;
    GUL1=GUL2;
    GUR1=GUR2;
    GBxL1=GBxL2;
    GBxR1=GBxR2;
    GByL1=GByL2;
    GByR1=GByR2;
    GBzL1=GBzL2;
    GBzR1=GBzR2;
    
    c1=c2;
    
    roL2=rhoc1[x][y][z]+0.5*drhoc[x][y][z];
    roR2=rhoc1[x][y][z+1]-0.5*drhoc[x][y][z+1];
    pxL2=pxc1[x][y][z]+0.5*dpxc[x][y][z];
    pxR2=pxc1[x][y][z+1]-0.5*dpxc[x][y][z+1];
    pyL2=pyc1[x][y][z]+0.5*dpyc[x][y][z];
    pyR2=pyc1[x][y][z+1]-0.5*dpyc[x][y][z+1];
    pzL2=pzc1[x][y][z]+0.5*dpzc[x][y][z];
    pzR2=pzc1[x][y][z+1]-0.5*dpzc[x][y][z+1];
    UL2=Uc1[x][y][z]+0.5*dUc[x][y][z];
    UR2=Uc1[x][y][z+1]-0.5*dUc[x][y][z+1];
    BxL2=Bxc1[x][y][z]+0.5*dBxc[x][y][z];
    BxR2=Bxc1[x][y][z+1]-0.5*dBxc[x][y][z+1];
    ByL2=Byc1[x][y][z]+0.5*dByc[x][y][z];
    ByR2=Byc1[x][y][z+1]-0.5*dByc[x][y][z+1];
    BzL2=Bzc1[x][y][z]+0.5*dBzc[x][y][z];
    BzR2=Bzc1[x][y][z+1]-0.5*dBzc[x][y][z+1];
    GroL2=Gmz_ro(pzL2);
    GroR2=Gmz_ro(pzR2);
    GpxL2=Gmz_px(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpxR2=Gmz_px(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpyL2=Gmz_py(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpyR2=Gmz_py(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GpzL2=Gmz_pz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GpzR2=Gmz_pz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GUL2=Gzm_U(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GUR2=Gzm_U(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBxL2=Gmz_Bx(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBxR2=Gmz_Bx(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GByL2=Gmz_By(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GByR2=Gmz_By(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    GBzL2=Gmz_Bz(roL2,pxL2,pyL2,pzL2,C_g,BxL2,ByL2,BzL2,UL2);
    GBzR2=Gmz_Bz(roR2,pxR2,pyR2,pzR2,C_g,BxR2,ByR2,BzR2,UR2);
    
    
    c2=F_cq(avr_MHD(BxL2,BxR2),avr_MHD(ByL2,ByR2),avr_MHD(BzL2,BzR2),avr_MHD(BzL2,BzR2),avr_MHD(roL2,roR2),C_g,avr_MHD(  (ppp_MHD(roL2,pxL2,pyL2,pzL2,BxL2,ByL2,BzL2,UL2,C_g)) ,  (ppp_MHD(roR2,pxR2,pyR2,pzR2,BxR2,ByR2,BzR2,UR2,C_g))    ) ,avr_MHD((pzL2/roL2),(pzR2/roR2)));
    
    //#if (SOURCES==2)
    rhoc2[x][y][z]=F_W2(rhoc0[x][y][z],dt,invdz,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,Q__rhoc[x][y][z]);
    pxc2[x][y][z]=F_W2(pxc0[x][y][z],dt,invdz,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,Q__pxc[x][y][z]);
    pyc2[x][y][z]=F_W2(pyc0[x][y][z],dt,invdz,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,Q__pyc[x][y][z]);
    pzc2[x][y][z]=F_W2(pzc0[x][y][z],dt,invdz,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,Q__pzc[x][y][z]);
    Uc2[x][y][z]=F_W2(Uc0[x][y][z],dt,invdz,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,Q__Uc[x][y][z]);
    Bxc2[x][y][z]=F_W2(Bxc0[x][y][z],dt,invdz,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,Q__Bxc[x][y][z]);
    Byc2[x][y][z]=F_W2(Byc0[x][y][z],dt,invdz,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,Q__Byc[x][y][z]);
    Bzc2[x][y][z]=F_W2(Bzc0[x][y][z],dt,invdz,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,Q__Bzc[x][y][z]);
    //#else
    /*			(*ro2)[x][y][z]=F_W2((*ro0)[x][y][z],dt,dy,GroL1,GroR1,GroL2,GroR2,roL1,roR1,roL2,roR2,c1,c2,0.0);
     (*px2)[x][y][z]=F_W2((*px0)[x][y][z],dt,dy,GpxL1,GpxR1,GpxL2,GpxR2,pxL1,pxR1,pxL2,pxR2,c1,c2,0.0);
     (*py2)[x][y][z]=F_W2((*py0)[x][y][z],dt,dy,GpyL1,GpyR1,GpyL2,GpyR2,pyL1,pyR1,pyL2,pyR2,c1,c2,0.0);
     (*pz2)[x][y][z]=F_W2((*pz0)[x][y][z],dt,dy,GpzL1,GpzR1,GpzL2,GpzR2,pzL1,pzR1,pzL2,pzR2,c1,c2,0.0);
     (*U2)[x][y][z]=F_W2((*U0)[x][y][z],dt,dy,GUL1,GUR1,GUL2,GUR2,UL1,UR1,UL2,UR2,c1,c2,0.0);
     (*Bx2)[x][y][z]=F_W2((*Bx0)[x][y][z],dt,dy,GBxL1,GBxR1,GBxL2,GBxR2,BxL1,BxR1,BxL2,BxR2,c1,c2,0.0);
     (*By2)[x][y][z]=F_W2((*By0)[x][y][z],dt,dy,GByL1,GByR1,GByL2,GByR2,ByL1,ByR1,ByL2,ByR2,c1,c2,0.0);
     (*Bz2)[x][y][z]=F_W2((*Bz0)[x][y][z],dt,dy,GBzL1,GBzR1,GBzL2,GBzR2,BzL1,BzR1,BzL2,BzR2,c1,c2,0.0);
     */
    //#endif
    
    
    //		(*P0)[x][y]=ppp_MHD((*ro2)[x][y],(*px2)[x][y],(*py2)[x][y],(*pz2)[x][y],(*Bx2)[x][y],(*By2)[x][y],(*Bz2)[x][y],(*U2)[x][y],C_g);
    END_z_small
	END_y_small
    END_x_small
    //boundaries
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    // MPI_Barrier(MPI_COMM_WORLD);
    communicateCenterMHD(nxc,nyc,nzc,rhoc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pyc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,pzc2,vct);
    
    communicateCenterMHD(nxc,nyc,nzc,Bxc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Byc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Bzc2,vct);
    communicateCenterMHD(nxc,nyc,nzc,Uc2,vct);
    
    
    // MPI_Barrier(MPI_COMM_WORLD);
    Fix_P(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,grid);
    Boundarize(nxc,nyc,nzc,rhoc2,pxc2,pyc2,pzc2,Bxc2,Byc2,Bzc2,Uc2,vct);
    
    // ponyatno...
    /*
     BEGIN_full
     (*ro0)[x][y][z]=(*ro2)[x][y][z];
     (*px0)[x][y][z]=(*px2)[x][y][z];
     (*py0)[x][y][z]=(*py2)[x][y][z];
     (*pz0)[x][y][z]=(*pz2)[x][y][z];
     (*U0)[x][y][z]=(*U2)[x][y][z];
     (*Bx0)[x][y][z]=(*Bx2)[x][y][z];
     (*By0)[x][y][z]=(*By2)[x][y][z];
     (*Bz0)[x][y][z]=(*Bz2)[x][y][z];
     END_full
     
     ro=ro0;
     px=px0;
     py=py0;
     pz=pz0;
     U =U0;
     Bx=Bx0;
     By=By0;
     Bz=Bz0;
     */
    
    eq(rhoc,rhoc2,nxc,nyc,nzc);
    eq(pxc,pxc2,nxc,nyc,nzc);
    eq(pyc,pyc2,nxc,nyc,nzc);
    eq(pzc,pzc2,nxc,nyc,nzc);
    eq(Uc,Uc2,nxc,nyc,nzc);
    eq(Bxc,Bxc2,nxc,nyc,nzc);
    eq(Byc,Byc2,nxc,nyc,nzc);
    eq(Bzc,Bzc2,nxc,nyc,nzc);
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc[x][y][z]==rhoc[x][y][z])) { cout << "Calc_Lz end  NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc[x][y][z]==pxc[x][y][z]))   { cout << "Calc_Lz end  NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc[x][y][z]==pyc[x][y][z]))   { cout << "Calc_Lz end  NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc[x][y][z]==pzc[x][y][z]))   { cout << "Calc_Lz end  NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc[x][y][z]==Bxc[x][y][z]))  {cout << "Calc_Lz end  NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc[x][y][z]==Byc[x][y][z]))  {cout << "Calc_Lz end  NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc[x][y][z]==Bzc[x][y][z]))  {cout << "Calc_Lz end  NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc[x][y][z]==Uc[x][y][z]))  { cout << "Calc_Lz end  NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    eq(rhoc0,rhoc2,nxc,nyc,nzc);
    eq(pxc0,pxc2,nxc,nyc,nzc);
    eq(pyc0,pyc2,nxc,nyc,nzc);
    eq(pzc0,pzc2,nxc,nyc,nzc);
    eq(Uc0,Uc2,nxc,nyc,nzc);
    eq(Bxc0,Bxc2,nxc,nyc,nzc);
    eq(Byc0,Byc2,nxc,nyc,nzc);
    eq(Bzc0,Bzc2,nxc,nyc,nzc);
    
    
    // Check for isnan here.
    BEGIN_full
    if (!(rhoc0[x][y][z]==rhoc0[x][y][z])) { cout << "Calc_Lz end 2 NaN rhoc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n"; stopFlag=-1;  }
    if (!(pxc0[x][y][z]==pxc0[x][y][z]))   { cout << "Calc_Lz end 2 NaN pxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(pyc0[x][y][z]==pyc0[x][y][z]))   { cout << "Calc_Lz end 2 NaN pyc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    if (!(pzc0[x][y][z]==pzc0[x][y][z]))   { cout << "Calc_Lz end 2 NaN pzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;  }
    
    if (!(Bxc0[x][y][z]==Bxc0[x][y][z]))  {cout << "Calc_Lz end 2 NaN Bxc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Byc0[x][y][z]==Byc0[x][y][z]))  {cout << "Calc_Lz end 2 NaN Byc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Bzc0[x][y][z]==Bzc0[x][y][z]))  {cout << "Calc_Lz end 2 NaN Bzc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    if (!(Uc0[x][y][z]==Uc0[x][y][z]))  { cout << "Calc_Lz end 2 NaN Uc catched: x "<<x<<" y: "<<y<<" z: "<<z<<" rank: "<<(vct->getCartesian_rank()) <<" order: "<<order<<"  \n";stopFlag=-1;   }
    END_full
    
    
    /*
     
     if (CONSTRAIN==3)
     constrain();
     else
     if (CONSTRAIN==2&&order%3==0)
     constrain();
     else
     if (CONSTRAIN==1&&order==6)
     constrain();
     
     //	saveNameNumber("Byc",(order*1000)+k,By2);
     
     result=customProceed1(order,1);
     if (result)
     return result;
     */
	return 0;
}










#endif
